#include "Constant.h"
#include "libconfig.h++"


namespace
{
	PetscScalar DeltaT_;
	PetscInt    TBeginEq_;
	PetscInt    TEndEq_;
	PetscInt    WriteEveryEq_;
	PetscInt    PrintEveryEq_;
	PetscInt    TBegin_  ;
	PetscInt    TEnd_  ;
	PetscInt    WriteEvery_  ;
	PetscInt    PrintEvery_;
	PetscInt    Nx_;
	PetscInt    Ny_;
	PetscInt    Nz_;
	PetscScalar Dx_;
	PetscScalar L_[3];
	PetscScalar HL_[3];
	PetscScalar Thickness_;
	PetscScalar Gap_;
	PetscInt    Sides_;
	PetscScalar Ec_;
	PetscScalar Damping_;
	PetscScalar Xi_;
	PetscScalar Re_;
	PetscScalar We_;
	PetscScalar Pe_;
// 	PetscScalar RhoRatio_;
// 	PetscScalar EtaRatio_;
	PetscScalar Volume_;
	PetscScalar GrowthCoef_;
	PetscScalar ContactAngle_;
	PetscScalar IsoPhi_;
	PetscScalar SmoothingCoef_;
	PetscScalar Gravity_[3];
	PetscScalar ThetaPhi_;
}


const PetscScalar& Constant::Time     ::DeltaT              = DeltaT_;
const PetscInt&    Constant::Time     ::TBeginEq            = TBeginEq_;
const PetscInt&    Constant::Time     ::TEndEq              = TEndEq_;
const PetscInt&    Constant::Time     ::WriteEveryEq        = WriteEveryEq_;
const PetscInt&    Constant::Time     ::PrintEveryEq        = PrintEveryEq_;
const PetscInt&    Constant::Time     ::TBegin              = TBegin_;
const PetscInt&    Constant::Time     ::TEnd                = TEnd_;
const PetscInt&    Constant::Time     ::WriteEvery          = WriteEvery_;
const PetscInt&    Constant::Time     ::PrintEvery          = PrintEvery_;
const PetscInt&    Constant::Grid     ::Nx                  = Nx_;
const PetscInt&    Constant::Grid     ::Ny                  = Ny_;
const PetscInt&    Constant::Grid     ::Nz                  = Nz_;
const PetscScalar& Constant::Grid     ::Dx                  = Dx_;
const PetscScalar* Constant::Grid     ::L                   = L_;
const PetscScalar* Constant::Grid     ::HL                  = HL_;
const PetscScalar& Constant::Template ::Thickness           = Thickness_;
const PetscScalar& Constant::Template ::Gap                 = Gap_;
const PetscInt&    Constant::Template ::Sides               = Sides_;
const PetscScalar& Constant::Hinge    ::Ec                  = Ec_;
const PetscScalar& Constant::Hinge    ::Damping             = Damping_;
const PetscScalar& Constant::Fluid    ::Xi                  = Xi_;
const PetscScalar& Constant::Fluid    ::Re                  = Re_;
const PetscScalar& Constant::Fluid    ::We                  = We_;
const PetscScalar& Constant::Fluid    ::Pe                  = Pe_;
// const PetscScalar& Constant::Fluid    ::RhoRatio            = RhoRatio_;
// const PetscScalar& Constant::Fluid    ::EtaRatio            = EtaRatio_;
const PetscScalar& Constant::Drop     ::Volume              = Volume_;
const PetscScalar& Constant::Drop     ::GrowthCoef          = GrowthCoef_;
const PetscScalar& Constant::Drop     ::ContactAngle        = ContactAngle_;
const PetscScalar& Constant::Drop     ::IsoPhi              = IsoPhi_;
const PetscScalar& Constant::Drop     ::SmoothingCoef       = SmoothingCoef_;
const PetscScalar* Constant::External ::Gravity             = Gravity_;
const PetscScalar& Constant::Numerical::ThetaPhi            = ThetaPhi_;


PetscErrorCode Constant::ReadInputFile
(
	const std::string &file
)
{
	std::string input = file;
	if (input == "")
	{
		char tmp[PETSC_MAX_PATH_LEN];
		PetscBool flg;
		PetscOptionsGetString(NULL,NULL,"-i",tmp,sizeof(tmp),&flg);
		if (!flg)
		{
			PetscPrintf(PETSC_COMM_WORLD,"  Error: Pass input file with the -i option\n");
			PetscEnd();
		}
		input = tmp;
	}

	PetscBool fileExists;
	PetscTestFile(input.c_str(),'\0',&fileExists);
	if (!fileExists)
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Error: %s not found\n",input.c_str());
		PetscEnd();
	}

	const PetscInt source = 0;
	PetscInt rank;
	MPI_Comm_rank(PETSC_COMM_WORLD,&rank);

	libconfig::Config cfg;
	cfg.setOptions(libconfig::Config::OptionAutoConvert);
	if (rank == source)
	{
		cfg.readFile(input);
	}

	//Read integer
	auto ReadInt = [&cfg,&rank,&source]
	(
		PetscInt &val,
		const std::string& str,
		bool print=true
	)
	{
		if (rank == source)
		{
			try
			{
				val = cfg.lookup(str);
				if (print)
				{
					printf("    %s = %i\n",str.substr(str.find_last_of(".") + 1).c_str(),val);
				}
			}
			catch (const libconfig::SettingNotFoundException &nfex)
			{
				printf("  Error: %s not found\n",str.c_str());
				PetscEnd();
			}
		}
		MPI_Bcast(&val,1,MPI_INT,source,PETSC_COMM_WORLD);
	};

	//Read integer
	auto ReadBool = [&cfg,&rank,&source]
	(
		PetscInt &val,
		const std::string& str,
		bool print=true
	)
	{
		if (rank == source)
		{
			try
			{
				std::string answer = cfg.lookup(str);
				val =
				(
					answer.compare("yes")  == 0 or
					answer.compare("y")    == 0 or
					answer.compare("True") == 0 or
					answer.compare("true") == 0
				) ? 1 : 0;
				if (print)
				{
					printf("    %s = %i\n",str.substr(str.find_last_of(".") + 1).c_str(),val);
				}
			}
			catch (const libconfig::SettingNotFoundException &nfex)
			{
				printf("  Error: %s not found\n",str.c_str());
				PetscEnd();
			}
		}
		MPI_Bcast(&val,1,MPI_INT,source,PETSC_COMM_WORLD);
	};

	//Read scalar
	auto ReadScalar = [&cfg,&rank,&source]
	(
		PetscScalar &val,
		const std::string& str,
		bool print=true
	)
	{
		if (rank == source)
		{
			try
			{
				val = cfg.lookup(str);
				if (print)
				{
					printf("    %s = %f\n",str.substr(str.find_last_of(".") + 1).c_str(),val);
				}
			}
			catch (const libconfig::SettingNotFoundException &nfex)
			{
				printf("  Error: %s not found\n",str.c_str());
				PetscEnd();
			}
		}
		MPI_Bcast(&val,1,MPIU_SCALAR,source,PETSC_COMM_WORLD);
	};

	//Read 3D scalar vector
	auto Read3DScalarArray = [&cfg,&rank,&source]
	(
		PetscScalar *val,
		const std::string& str,
		bool print=true
	)
	{
		if (rank == source)
		{
			try
			{
				const libconfig::Setting& setting = cfg.lookup(str);
				val[0] = setting[0];
				val[1] = setting[1];
				val[2] = setting[2];
				if (print)
				{
					printf("    %s = ( %f , %f , %f )\n",str.substr(str.find_last_of(".") + 1).c_str(),val[0],val[1],val[2]);
				}
			}
			catch (const libconfig::SettingNotFoundException &nfex)
			{
				printf("  Error: %s not found\n",str.c_str());
				PetscEnd();
			}
		}
		MPI_Bcast(val,3,MPIU_SCALAR,source,PETSC_COMM_WORLD);
	};

	bool dontPrint = false;

	//Time
	PetscPrintf(PETSC_COMM_WORLD,"  #Time\n");
	ReadScalar(DeltaT_,"Time.delta_t");

	//Equilibrium
	PetscPrintf(PETSC_COMM_WORLD,"  #Equilibrium\n");
	ReadInt(TBeginEq_,"Time.Equilibrium.restart",dontPrint);         if (TBeginEq_<0){TBeginEq_=-1;}
	ReadInt(TEndEq_,"Time.Equilibrium.n",dontPrint);                 if (TEndEq_<0){TEndEq_=0;} TEndEq_ += TBeginEq_;
	ReadInt(WriteEveryEq_,"Time.Equilibrium.write_every",dontPrint); if (WriteEveryEq_<1) {WriteEveryEq_=TEndEq_;}
	ReadInt(PrintEveryEq_,"Time.Equilibrium.print_every",dontPrint); if (PrintEveryEq_<1) {PrintEveryEq_=TEndEq_;}
	PetscPrintf(PETSC_COMM_WORLD,"    begin = %i\n",TBeginEq_);
	PetscPrintf(PETSC_COMM_WORLD,"    end = %i\n",TEndEq_);
	PetscPrintf(PETSC_COMM_WORLD,"    write_every = %i\n",WriteEveryEq_);
	PetscPrintf(PETSC_COMM_WORLD,"    print_every = %i\n",PrintEveryEq_);

	//Unsteady
	PetscPrintf(PETSC_COMM_WORLD,"  #Unsteady\n");
	ReadInt(TBegin_,"Time.Unsteady.restart",dontPrint);         if (TBegin_<0){TBegin_=-1;}
	ReadInt(TEnd_,"Time.Unsteady.n",dontPrint);                 if (TEnd_<0){TEnd_=0;} TEnd_ += TBegin_;
	ReadInt(WriteEvery_,"Time.Unsteady.write_every",dontPrint); if (WriteEvery_ <1) {WriteEvery_=TEnd_;}
	ReadInt(PrintEvery_,"Time.Unsteady.print_every",dontPrint); if (PrintEvery_ <1) {PrintEvery_=TEnd_;}
	PetscPrintf(PETSC_COMM_WORLD,"    begin = %i\n",TBegin_);
	PetscPrintf(PETSC_COMM_WORLD,"    end = %i\n",TEnd_);
	PetscPrintf(PETSC_COMM_WORLD,"    write_every = %i\n",WriteEvery_);
	PetscPrintf(PETSC_COMM_WORLD,"    print_every = %i\n",PrintEvery_);

	//Grid
	PetscPrintf(PETSC_COMM_WORLD,"  #Grid\n");
	ReadInt(Nx_,"Grid.nx");
	ReadInt(Ny_,"Grid.ny");
	ReadInt(Nz_,"Grid.nz");
	ReadScalar(Dx_,"Grid.dx");
	L_ [0] = Nx_ * Dx_; HL_[0] = L_[0]/2.;
	L_ [1] = Ny_ * Dx_; HL_[1] = L_[1]/2.;
	L_ [2] = Nz_ * Dx_; HL_[2] = L_[2]/2.;

	//Template
	PetscPrintf(PETSC_COMM_WORLD,"  #Template\n");
	ReadScalar(Thickness_,"Template.thickness");
	ReadScalar(Gap_,"Template.gap");
	ReadInt(Sides_,"Template.sides");

	//Hinge
	PetscPrintf(PETSC_COMM_WORLD,"  #Hinge\n");
	ReadScalar(Ec_,"Hinge.Ec");
	ReadScalar(Damping_,"Hinge.damping");

	//Fluid properties
	PetscPrintf(PETSC_COMM_WORLD,"  #Fluid\n");
	ReadScalar(Xi_,"Fluid.xi");
	ReadScalar(Re_,"Fluid.Re");
	ReadScalar(We_,"Fluid.We");
	ReadScalar(Pe_,"Fluid.Pe");
// 	ReadScalar(RhoRatio_,"Fluid.rhoRatio");
// 	ReadScalar(EtaRatio_,"Fluid.etaRatio");

	//Drop properties
	PetscPrintf(PETSC_COMM_WORLD,"  #Drop\n");
	ReadScalar(Volume_,"Drop.volume");
	ReadScalar(GrowthCoef_,"Drop.growth_coefficient");
	ReadScalar(ContactAngle_,"Drop.contact_angle");
	ReadScalar(IsoPhi_,"Drop.iso_phi");
	ReadScalar(SmoothingCoef_,"Drop.smoothing_coefficient");

	//External properties
	PetscPrintf(PETSC_COMM_WORLD,"  #External\n");
	Read3DScalarArray(Gravity_,"External.gravity");

	//Numerical
	PetscPrintf(PETSC_COMM_WORLD,"  #Numerical (phi)\n");
	ReadScalar(ThetaPhi_,"Numerical.phi.theta");

	//Message
	PetscPrintf(PETSC_COMM_WORLD,"  Input file read\n");
	return(0);
}
