#include "Context3D.h"


GetDirectionIndices();


using namespace EigenWrapper;


Context::Context
(
	DM da,
	PetscInt n
):
	da_{da},
	n_{n<0?0:n},
	base_{n_},
	gaps_{std::vector<Prism>(n_)},
	sides_{std::vector<TriangularPrism>(n_)},
	hinges_{std::vector<Hinge>(n_)},
	isoPhi_{Constant::Drop::IsoPhi},
	dropGrowthCoefficient_{Constant::Drop::GrowthCoef},
	contactAngle_{(Constant::Drop::ContactAngle)*(PETSC_PI/180.)},
	contactLines_{"tpcl"}, //tpcl = three phase contact line
	smoothingCoefficient_{Constant::Drop::SmoothingCoef},
	force_{EigenWrapper::EigenVec3::Zero()},
	torque_{EigenWrapper::EigenVec3::Zero()},
	save_{0},
	dt_{Constant::Time::DeltaT},
	dx_{Constant::Grid::Dx},
	dx2_{dx_*dx_},
	dx3_{dx_*dx2_},
	invDx1_{1./dx_},
	invDx2_{invDx1_*invDx1_},
	invDx3_{invDx2_*invDx1_},
	xi_{Constant::Fluid::Xi},
	xi2_{xi_*xi_},
	dtInvDx1_{dt_*invDx1_},
	dtInvPe_{dt_/(Constant::Fluid::Pe)},
	dtInvDx2InvPe_{dt_*invDx2_/(Constant::Fluid::Pe)},
	dtInvWe_{dt_/(Constant::Fluid::We)}
{
	PetscObjectReference((PetscObject)da);
	coords_.create(da_);
	IglWrapper::FlattenMarchingCubeCoordinates(da_,flattenedCoords_);
	phi_.create(da_,"phi");
	phiRHS_.create(da_,"phiRhs");
	mu_.create(da_,"mu");
	muBulk_.create(da_,"muBulk");
	phiWall_.create(da_,"phiWall");
	u_.create3D(da_,"u");
	base_.create(da_,"base");
	for (PetscInt s=0; s<n_; ++s)
	{
		gaps_[s].create(da_,"gap"+std::to_string(s));
		sides_[s].create(da_,"side"+std::to_string(s));
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Context created\n");
}


Context::~Context()
{
	coords_.destroy();
	phi_.destroy();
	phiRHS_.destroy();
	mu_.destroy();
	muBulk_.destroy();
	phiWall_.destroy();
	u_.destroy();
	base_.destroy();
	for (auto &gap:gaps_){gap.destroy();}
	for (auto &side:sides_){side.destroy();}
	DMDestroy(&da_);
	PetscPrintf(PETSC_COMM_WORLD,"  Context destroyed\n");
}


PetscErrorCode Context::write()
{
	phi_.writeInBinaryFormat(save_);
// 	u_.writeInBinaryFormat(save_);
	contactLines_.gatherAll();
	contactLines_.writeInBinaryFormat(save_);
	contactLines_.ungatherAll();
	for (const auto& hinge:hinges_){hinge.writeInBinaryFormat(save_);}
	if ( save_ < 1 )
	{
		base_.writeInBinaryFormat(-1);
		for (const auto& gap:gaps_){gap.writeInBinaryFormat(-1);}
	}
	for (const auto& side:sides_){side.writeInBinaryFormat(save_);}
	save_++;
	return(0);
}


PetscErrorCode Context::read
(
	PetscInt save
)
{
	save_ = save;
	phi_.readInBinaryFormat(save_);
	for (auto& hinge:hinges_)
	{
		hinge.readInBinaryFormat(save_);
	}
	base_.readInBinaryFormat(-1);
	base_.triangulateAndSetPhi(xi_);
	for (auto& gap:gaps_)
	{
		gap.readInBinaryFormat(-1);
		gap.triangulateAndSetPhi(xi_);
	}
	for (auto& side:sides_)
	{
		side.readInBinaryFormat(save_);
		side.triangulateAndSetPhi(xi_);
	}
	this->updatePhiWall();
	save_++;
	return(0);
}


PetscErrorCode Context::printInfo
(
	PetscInt t
)
{
	if (t>-1)
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Iteration %i\n",t);
	}
	PetscPrintf(PETSC_COMM_WORLD,"    V(drop) = %.4f\n",dropVolume_);
	PetscPrintf(PETSC_COMM_WORLD,"    V(drop)/V0 = %.4f\n",dropVolume_/prescribedDropVolume_);
	PetscPrintf(PETSC_COMM_WORLD,"     Fc = %.4f %.4f %.4f\n",force_.x(),force_.y(),force_.z());
	PetscPrintf(PETSC_COMM_WORLD,"     Tc = %.4f %.4f %.4f\n",torque_.x(),torque_.y(),torque_.z());
	if (n_>0)
	{
		PetscScalar load = hinges_.front().normalizedLoad(force_.norm());
		PetscPrintf(PETSC_COMM_WORLD,"    PL^2/(EI) = %.4f\n",load);
		PetscScalar c1 = sides_[0].Tm(2).z();
		PetscScalar c0 = base_.centre().z();
		PetscScalar posz = c1-c0;
		PetscPrintf(PETSC_COMM_WORLD,"    zc(side0) = %.4f\n",posz);
	}
	return(0);
}


PetscErrorCode Context::initialize()
{
	this->initializeTemplate();
	this->initializeHinges();
	this->initializeDrop((Constant::Grid::L[X_])/6);
	PetscPrintf(PETSC_COMM_WORLD,"  Initialization done\n");
// 	this->write();
	return(0);
}


PetscErrorCode Context::initialize
(
	PetscScalar area,
	PetscScalar height
)
{
	this->initializeTemplate(area,height);
	this->initializeHinges();
	this->initializeDrop(PetscSqrtReal(area));
	PetscPrintf(PETSC_COMM_WORLD,"  Initialization done\n");
// 	this->write();
	return(0);
}


PetscErrorCode Context::initializeTemplate()
{
	EigenVec3 centre
	(
		Constant::Grid::HL[X_],
		Constant::Grid::HL[Y_],
		Constant::Template::Thickness/2 + 5*dx_
	);
	PetscScalar thickness = (Constant::Template::Thickness);
	PetscScalar alpha = 0.95;
	EigenWrapper::EigenMat m = EigenWrapper::EigenMat(4,3);
	m.row(0) = centre + EigenWrapper::EigenVec3(+alpha*(Constant::Grid::HL[X_]),+alpha*(Constant::Grid::HL[Y_]),0);
	m.row(1) = centre + EigenWrapper::EigenVec3(-alpha*(Constant::Grid::HL[X_]),+alpha*(Constant::Grid::HL[Y_]),0);
	m.row(2) = centre + EigenWrapper::EigenVec3(-alpha*(Constant::Grid::HL[X_]),-alpha*(Constant::Grid::HL[Y_]),0);
	m.row(3) = centre + EigenWrapper::EigenVec3(+alpha*(Constant::Grid::HL[X_]),-alpha*(Constant::Grid::HL[Y_]),0);
	base_.updateMatrix(m);
	base_.height() = thickness;
	base_.triangulateAndSetPhi(xi_);
	this->updatePhiWall();
	return(0);
}


PetscErrorCode Context::initializeTemplate
(
	PetscScalar baseArea,
	PetscScalar pyramideHeight
)
{
	//Input values
	const EigenVec3 centre
	(
		Constant::Grid::HL[X_],
		Constant::Grid::HL[Y_],
		Constant::Template::Thickness/2. + 5.*dx_
	);
	const PetscScalar thickness = (Constant::Template::Thickness);
	const PetscScalar gap = (Constant::Template::Gap);

	//Unfolded pyramidal structure
	const PetscScalar n = static_cast<PetscScalar>(n_);
	if ( n>2 )
	{
		//         C
		//        / \
		//       /   \
		//      A-----B
		//    B C-----B A
		//   /| |     | |\
		//  / | |     | | \
		// C  | |     | |  C
		//  \ | |     | | /
		//   \| |     | |/
		//    A D-----A B
		//      B-----A      y
		///      \   /       |
		///       \ /        |
		///        C         O---->x
		//
		// A, B and C are placed in counterclockwise direction
		// the normal vector T = (B-A) x (C-A) points outwards

		//Dimensions of pyramid
		const PetscScalar alpha = 2.*(PETSC_PI/n);
		const PetscScalar baseRadius2  = 2.* (baseArea/n) / PetscSinReal(alpha  ); // Radius of the base
		const PetscScalar baseSide2    = 4.* (baseArea/n) * PetscTanReal(alpha/2); // Side length of the base
		const PetscScalar baseApothem2 =     (baseArea/n) / PetscTanReal(alpha/2); // Apothem of the base
		const PetscScalar sideHeight2 = baseApothem2 + pyramideHeight*pyramideHeight; // Height of side triangle
		const PetscScalar baseRadius  = PetscSqrtReal(baseRadius2);
		const PetscScalar baseSide    = PetscSqrtReal(baseSide2);
		const PetscScalar baseApothem = PetscSqrtReal(baseApothem2);
		const PetscScalar sideHeight = PetscSqrtReal(sideHeight2);
		PetscPrintf(PETSC_COMM_WORLD,"  Base area = %.3f\n",baseArea);
		PetscPrintf(PETSC_COMM_WORLD,"  Pyramid height = %.3f\n",pyramideHeight);
		PetscPrintf(PETSC_COMM_WORLD,"  Pyramid volume = %.3f\n",baseArea*pyramideHeight/3);
		PetscPrintf(PETSC_COMM_WORLD,"  Base radius = %.3f\n",baseRadius);
		PetscPrintf(PETSC_COMM_WORLD,"  Base side = %.3f\n",baseSide);
		PetscPrintf(PETSC_COMM_WORLD,"  Base apothem = %.3f\n",baseApothem);
		PetscPrintf(PETSC_COMM_WORLD,"  Side height = %.3f\n",sideHeight);

		//N-sided base in (XY)-plane, normal vector (B-A)x(C-A)/|(B-A)x(C-A)| = (0,0,1)^T
		EigenWrapper::EigenMat m = base_.m();
		for (PetscInt r=0; r<m.rows(); ++r)
		{
			PetscScalar alphai = r*alpha - alpha/2;
			m.row(r) <<
				centre.x() + baseRadius*PetscCosReal(alphai),
				centre.y() + baseRadius*PetscSinReal(alphai),
				centre.z();
		}
		base_.updateMatrix(m);
		base_.height() = thickness;
		base_.triangulateAndSetPhi(xi_);
		PetscPrintf(PETSC_COMM_WORLD,"  base initialized\n");

		// Correction closing the gap and
		// preventing the drop from migrating
		// to the down side of the template
		for (PetscInt s=0; s<n_; ++s)
		{
			m.resize(4,3);
			PetscInt snext = (s+1) % (n_);
			EigenVec3 A = base_.Tm().row(s);
			EigenVec3 B = base_.Tm().row(snext);
			EigenVec3 dir = (B-A).cross(base_.n()).normalized();
			m.row(0) = (A + (1.05*gap) * dir);
			m.row(1) = (B + (1.05*gap) * dir);
			m.row(2) = B;
			m.row(3) = A;;
			gaps_[s].updateMatrix(m);
			gaps_[s].height() = thickness;
			gaps_[s].triangulateAndSetPhi(xi_);
		}
		PetscPrintf(PETSC_COMM_WORLD,"  gaps initialized\n");

		//Triangular sides in (XY)-plane, normal vector (B-A)x(C-A)/|(B-A)x(C-A)| = (0,0,1)^T
		m.resize(3,3);
		for (PetscInt s=0; s<n_; ++s)
		{
			PetscInt snext = (s+1) % (n_);
			EigenVec3 As = base_.m(snext);
			EigenVec3 Bs = base_.m(s);
			EigenVec3 dir = ((As-Bs).cross(base_.n())).normalized();
			EigenVec3 Cs = (As+Bs)/2 + sideHeight * dir;
			m.row(0) = As.transpose();
			m.row(1) = Bs.transpose();
			m.row(2) = Cs.transpose();
			m = m.rowwise() + (centre + gap*dir).transpose();
			sides_[s].updateMatrix(m);
			sides_[s].height() = thickness;
			sides_[s].triangulateAndSetPhi(xi_);
		}
		PetscPrintf(PETSC_COMM_WORLD,"  sides initialized\n");

		//Update smooth wall
		this->updatePhiWall();
	}
	else
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Error: Template side must be greater than 3.\n");
		PetscEnd();
	}

	return(0);
}


PetscErrorCode Context::initializeHinges()
{
	const PetscScalar materialDensity = 1.;
	const PetscInt beadNumber = 10;
	for (PetscInt s=0; s<n_; ++s)
	{
		auto& hinge = hinges_[s];
		PetscInt snext = (s+1) % (n_);
		EigenVec3 A0 = base_.Tm(s);
		EigenVec3 B0 = base_.Tm(snext);
		EigenVec3 A1 = sides_[s].Tm(1);
		EigenVec3 B1 = sides_[s].Tm(0);
		EigenVec3 A = (A0+B0)/2.;
		EigenVec3 B = (A1+B1)/2.;
		hinge = Hinge
		(
			materialDensity,
			A,
			B,
			beadNumber,
			"hinge" + std::to_string(s)
		);
		hinge.damping() = (Constant::Hinge::Damping);
		for (auto& bead:hinge.beads())
		{
			bead.mass() = 1.;
			bead.inertia() = 1.;
		}
	}
	return(0);
}


PetscErrorCode Context::initializeDrop
(
	PetscScalar radius
)
{
	if ( (Constant::Drop::Volume) > 0 )
	{
		prescribedDropVolume_ = Constant::Drop::Volume;
		PetscScalar nomi = prescribedDropVolume_;
		PetscScalar denomi = 2./3. * PETSC_PI;
		prescribedDropRadius_ = PetscPowScalarReal(nomi/denomi,1./3.);
	}
	else
	{
		prescribedDropVolume_ = (2./3.) * PETSC_PI * PetscPowRealInt(radius,3);
		prescribedDropRadius_ = radius;
	}
	PetscScalar lx = (Constant::Grid::L[X_]), hLx = lx/2.;
	PetscScalar ly = (Constant::Grid::L[Y_]), hLy = ly/2.;
	PetscScalar lz = (Constant::Grid::L[Z_]), hLz = lz/2.;
	PetscScalar Xb = base_.centre().x();
	PetscScalar Yb = base_.centre().y();
	PetscScalar Zb = base_.centre().z() + base_.height()/2.;
	PetscScalar xi = 1e-6;
	PetscScalar lV = 0., gV;
	coords_.getArray();
	phi_.getArray();
	ForAllKJI(da_,
	{
		PetscScalar rx = Xb - coords_.x(k,j,i); //if (rx <= -hLx){rx += lx;} if (rx >= hLx){rx -= lx;}
		PetscScalar ry = Yb - coords_.y(k,j,i); //if (ry <= -hLy){ry += ly;} if (ry >= hLy){ry -= ly;}
		PetscScalar rz = Zb - coords_.z(k,j,i); //if (rz <= -hLz){rz += lz;} if (rz >= hLz){rz -= lz;}
		PetscScalar r  = PetscSqrtScalar(rx*rx + ry*ry + rz*rz);
		PetscScalar phi = rz > 0. ? 0. : 0.5 + 0.5 * PetscTanhScalar( (prescribedDropRadius_-r) / xi );
		phi_(k,j,i) = phi;
// 		lV += phi;
		lV += PetscSqr(phi);
	})
	coords_.restoreArray();
	phi_.restoreArray();
	MPI_Allreduce(&lV,&gV,1,MPIU_SCALAR,MPIU_SUM,PETSC_COMM_WORLD);
	dropVolume_ = gV*dx3_;
	PetscPrintf(PETSC_COMM_WORLD,"  Prescribed drop volume V0 = %.3f\n",prescribedDropVolume_);
	PetscPrintf(PETSC_COMM_WORLD,"  Prescribed radius R0 = %.3f\n",prescribedDropRadius_);
	PetscPrintf(PETSC_COMM_WORLD,"  Drop initialized with volume V(t=0) = %.3f\n",dropVolume_);
	return(0);
}


PetscErrorCode Context::updatePhiWall()
{
	for (auto &side:sides_){side.setPhi(xi_);}
	base_.phi().copyTo(phiWall_);
	phiWall_.getArray();
	for (auto &gap:gaps_){gap.phi().getArray();}
	for (auto &side:sides_){side.phi().getArray();}
	ForAllKJI(da_,
	{
		PetscScalar sum = phiWall_(k,j,i);
		for (const auto &side:sides_)
		{
			sum = PetscMax
			(
				sum,
				side.phi().a(k,j,i)
			);
		}
		for (const auto &gap:gaps_)
		{
			sum = PetscMax
			(
				sum,
				gap.phi().a(k,j,i)
			);
		}
		phiWall_(k,j,i) = sum;
	})
	phiWall_.restoreArray();
	for (auto &gap:gaps_){gap.phi().restoreArray();}
	for (auto &side:sides_){side.phi().restoreArray();}
	return(0);
}


PetscErrorCode Context::updatePhi()
{
	//V(t)
	PetscScalar lV=0., gV;
	phi_.getArray();
	ForAllKJI(da_,
	{
// 		lV += phi_(k,j,i);
		lV += PetscSqr( phi_(k,j,i) );
	})
	phi_.restoreArray();
	MPI_Allreduce(&lV,&gV,1,MPIU_SCALAR,MPIU_SUM,PETSC_COMM_WORLD);
	gV *= (dx3_);
	dropVolume_ = gV;

	//Bulk
	PetscScalar cosTheta = PetscCosReal(contactAngle_);
	PetscScalar corr = (dropVolume_)/(prescribedDropVolume_) - 1.; //corr < 0 [growth] else [shrinkage]
	PetscScalar phic = 0.5 + dropGrowthCoefficient_ * corr;
	GetArray(phi_,muBulk_);
	phiWall_.getGhostedArray();
	ForAllKJI(da_,
	{
		PetscScalar phi = phi_(k,j,i);
		PetscScalar phiw = phiWall_(k,j,i);
		PetscScalar nx = ( phiWall_(k  ,j  ,i+1) - phiWall_(k  ,j  ,i-1) )/2.;
		PetscScalar ny = ( phiWall_(k  ,j+1,i  ) - phiWall_(k  ,j-1,i  ) )/2.;
		PetscScalar nz = ( phiWall_(k+1,j  ,i  ) - phiWall_(k-1,j  ,i  ) )/2.;
		PetscScalar nmag = PetscSqrtReal( nx*nx + ny*ny + nz*nz );
		PetscScalar muBulk;
		muBulk  = 4. * phi * (phi-1.) * (phi - phic);
		muBulk -= 4. * phi * (1.-phi-phiw) * phiw;
		muBulk += 4. * (xi_) * phi * (phi-1.) * nmag * cosTheta;
		muBulk_(k,j,i) = muBulk;
	})
	phiWall_.restoreGhostedArray();
	RestoreArray(phi_,muBulk_);

	//RHS
	GetArray(phiRHS_,muBulk_,u_);
	GetGhostedArray(phi_,phiWall_);
	ForAllKJI(da_,
	{
		//All variables, including velocities, are calculated at node centre (k,j,i)
		PetscScalar lapPhi = Discretisation3D::SecondOrder::Laplacian(phi_(),k,j,i);
		PetscScalar gradPhiGradPhiw_x = ( phi_(k  ,j  ,i+1) - phi_(k  ,j  ,i-1) ) * ( phiWall_(k  ,j  ,i+1) - phiWall_(k  ,j  ,i-1) );
		PetscScalar gradPhiGradPhiw_y = ( phi_(k  ,j+1,i  ) - phi_(k  ,j-1,i  ) ) * ( phiWall_(k  ,j+1,i  ) - phiWall_(k  ,j-1,i  ) );
		PetscScalar gradPhiGradPhiw_z = ( phi_(k+1,j  ,i  ) - phi_(k-1,j  ,i  ) ) * ( phiWall_(k+1,j  ,i  ) - phiWall_(k-1,j  ,i  ) );
		PetscScalar gradPhiGradPhiw = 0.25 * ( gradPhiGradPhiw_x + gradPhiGradPhiw_y + gradPhiGradPhiw_z );
		PetscScalar uDir[3] =
		{
			u_.x(k,j,i),
			u_.y(k,j,i),
			u_.z(k,j,i)
		};
		PetscScalar uGradPhi = Discretisation3D::FirstOrderUpwind::UGrad(uDir,phi_(),k,j,i);
		phiRHS_(k,j,i) = phi_(k,j,i) - (dt_) * uGradPhi  - (dtInvPe_) * muBulk_(k,j,i) + (dtInvDx2InvPe_) * (lapPhi - gradPhiGradPhiw);
	})
	RestoreArray(phiRHS_,muBulk_,u_);
	RestoreGhostedArray(phi_,phiWall_);

	//Phi(d+dt)
	phiRHS_.copyTo(phi_);

	//Check
	phi_.check();

	return(0);
}


PetscErrorCode Context::updateContactLine
(
	const EigenMat &wallVertices,
	const EigenMatI &wallFaces,
	PetscInt time
)
{
	//Reset contact lines
	contactLines_.clearAll();

	//Tesselate drop
	EigenMat dropVertices;
	EigenMatI dropFaces;
	IglWrapper::MarchingCubes
	(
		flattenedCoords_,
		phi_,
		isoPhi_,
		dropVertices,
		dropFaces
	);

	//Intersect drop with template to get contact line (raw)
	Polylines contactLines
	(
		dropVertices,
		dropFaces,
		wallVertices,
		wallFaces
	);
	Polylines errRaw = contactLines();

	//Gather contact lines among all procs
	contactLines.gatherAll();

	//Contact lines are found
	if ( !contactLines.areEmpty() )
	{
		//Concatenate contact lines and share them among processors
		contactLines.concatenateAll();
		contactLines.ungatherAll();
		Polylines errCon = contactLines();

		//Clean contact lines
		PetscScalar epsilon = 0.1*dx_;
		contactLines.cleanAll(epsilon);
		Polylines errCln = contactLines();

		//First attempts at smoothing the three phase contact lines (tpcl)
		contactLines.smoothAll(epsilon,smoothingCoefficient_);

		//Second smoothing attempt
		if ( !contactLines.areSmooth() )
		{
			epsilon = 0.5*dx_;
			contactLines.cleanAll(epsilon);
			contactLines.smoothAll(epsilon,smoothingCoefficient_);
		}

		//third and last smoothing attempt
		if ( !contactLines.areSmooth() )
		{
			epsilon = 1.0*dx_;
			contactLines.cleanAll(epsilon);
			contactLines.smoothAll(epsilon,smoothingCoefficient_);
		}
		contactLines_ = contactLines();

		//Print error
		if ( !contactLines.areSmooth() )
		{
			std::cout << "  Warning: smoothing failed on one proc" << std::endl;
		}

		//Export data upon smoothing failure
		if ( !contactLines.gatheredAreSmooth() )
		{
			PetscInt source=0, rank;
			MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
			PetscPrintf(PETSC_COMM_WORLD,"  Warning: smoothing failed\n");
			std::string rk = std::to_string(rank);
			std::string it = std::to_string(time);
			IglWrapper::WriteOFF("phi_rk" + rk + "_it" + it + ".off",dropVertices,dropFaces);
			if (rank==source)
			{
				IglWrapper::WriteOFF("wall_it" + it + ".off",wallVertices,wallFaces);
			}
			errRaw.gatherAll(); errRaw.writeFromSourceInASCIIFormat("tpcl_it" + it + "_raw",-1);
			errCon.gatherAll(); errCon.writeFromSourceInASCIIFormat("tpcl_it" + it + "_con",-1);
			errCln.gatherAll(); errCln.writeFromSourceInASCIIFormat("tpcl_it" + it + "_cln",-1);
			contactLines.gatherAll(); contactLines.writeFromSourceInASCIIFormat("tpcl_it" + it + "_smo",-1);

		}
	}
	return(0);
}


PetscErrorCode Context::updateContactLine
(
	const SolidBody& sb,
	PetscInt iteration
)
{
	this->updateContactLine
	(
		sb.topVertices(),
		sb.topFaces(),
		iteration
	);
	return(0);
}


PetscErrorCode Context::updateContactLine
(
	PetscInt iteration
)
{
	const auto& side = sides_.front();
	this->updateContactLine(side,iteration);
	return(0);
}


PetscErrorCode Context::updateCapillaryForce()
{
	force_.setZero();
	torque_.setZero();
	if ( !contactLines_.areEmpty() )
	{
		PetscScalar sinTheta = PetscSinReal(contactAngle_);
		if (n_>0)
		{
			//Define reference force
			const PetscScalar Ec = Constant::Hinge::Ec;
			const auto& hg = hinges_.front();
			const PetscScalar referenceLength = prescribedDropRadius_;
			const PetscScalar referenceLength2 = PetscSqr(prescribedDropRadius_);
			const PetscScalar referenceForce = hg.referenceForce();

			//Force acting on template side
			const auto& side = sides_.front();
			EigenVec3 pivot = ( side.Tm(0) + side.Tm(1) )/2.;
			EigenVec3 n = side.Tn();
			for (const auto& pl:contactLines_())
			{
				force_ += sinTheta * pl.force(n);
				EigenVec3 actionPoint = pl.barycentre();
				{
					//Correction setting point of action in the plane defined by normal vector (u^v)
					EigenVec3 u = n;
					EigenVec3 v = side.centre() - pivot;
					EigenVec3 w = (u.cross(v)).normalized();
					actionPoint -= (actionPoint-pivot).dot(w) * w;
				}
				EigenVec3 leverArm = actionPoint - pivot;
				torque_ += leverArm.cross(force_);
			}

			//Normalize force
			force_  *= Ec / referenceLength  * referenceForce;
			torque_ *= Ec / referenceLength2 * referenceForce;
		}
		else
		{
			//Force acting on the base
			EigenVec3 n = base_.n();
			for (const auto& pl:contactLines_())
			{
				force_ += sinTheta * pl.force(n);
			}
		}
	}
	force_ = EigenWrapper::MPIReduce(force_);
	torque_ = EigenWrapper::MPIReduce(torque_);
	return(0);
}


PetscErrorCode Context::updateHinges()
{
	if ( n_>0 )
	{
		//Update first hinge
		hinges_[0].update(force_,torque_,dtInvWe_);

		//Update further hinges with h>0
		for (PetscInt h=1; h<n_; ++h)
		{
			//It would make more sense to use the first and second points of the base for the vector, that is u(s)=u(0)
			//u then rotates by s/(n-1) x 180 to get v(s) for s=1,..,n-1
			//It turns out, for n = 6 sides, the rotation from u(s=0)=(1,0,0) to v(s=3)=(-1,0,0) causes problems
			//The problem is bypassed by using the points s and s-1 for u, that is u(s)=v(s-1)
			PetscInt hprev = (h==0) ? (n_-1) : (h-1);
			PetscInt hnext = (h+1) % (n_);
			EigenVec3 u = (base_.m(h    ) + base_.m(hprev)).normalized();
			EigenVec3 v = (base_.m(hnext) + base_.m(h    )).normalized();
			EigenQuat quv;
			quv.setFromTwoVectors(u,v);
			quv.normalize();
			EigenMat Ruv = quv.toRotationMatrix();
			auto& hingeu = hinges_[hprev];
			auto& hingev = hinges_[h    ];
			for (PetscInt b=0; b<hingeu.size(); ++b)
			{
				auto& beadu = hingeu.bead(b);
				auto& beadv = hingev.bead(b);
				const EigenAngleAxis aau(beadu.quaternion());
				beadv.position() = base_.centre() + Ruv * ( beadu.position() - base_.centre() );
				beadv.quaternion() = EigenWrapper::EigenAngleAxis(aau.angle(),Ruv*aau.axis());
				beadv.vel() = Ruv * beadu.vel();
				beadv.omg() = Ruv * beadu.omg();
			}
		}

		//Check
		for (const auto& hinge:hinges_)
		{
			hinge.check();
		}
	}
	return(0);
}


PetscErrorCode Context::updateSides()
{
	for (PetscInt h=0; h<n_; ++h)
	{
		const auto& lastBead = hinges_[h].lastBead();
		auto& side = sides_[h];
		EigenVec3 arm = /* COM */ - ( side.m(0) + side.m(1) )/2.;
		arm = lastBead.quaternion().toRotationMatrix() * arm;
		side.centre() = lastBead.position() + arm;
		side.quaternion() = lastBead.quaternion();
		side.triangulate();//phi is set later
	}
	return(0);
}


PetscErrorCode Context::updateU()
{
	u_.zero();
	u_.getArray();
	coords_.getArray();
	for (PetscInt h=0; h<n_; ++h)
	{
		auto& side = sides_[h];
		side.phi().getArray();
		const EigenVec3 &velSolid = hinges_[h].lastBead().vel();
		const EigenVec3 &omgSolid = hinges_[h].lastBead().omg();
		const EigenVec3 &beadPosition = hinges_[h].lastBead().position();
		ForAllKJI(da_,
		{
			EigenVec3 gridPoint
			(
				coords_.x(k,j,i),
				coords_.y(k,j,i),
				coords_.z(k,j,i)
			);
			EigenVec3 arm = gridPoint - beadPosition;
			EigenVec3 vs = velSolid + omgSolid.cross(arm);
			for (PetscInt dim=0; dim<3; ++dim)
			{
				//at node centre (k,j,i)
				u_(dim,k,j,i) += side.phi().a(k,j,i) * vs(dim);
			}
		})
		side.phi().restoreArray();
	}
	u_.restoreArray();
	coords_.restoreArray();
	return(0);
}


PetscErrorCode Context::validateHingeModel()
{
	PetscScalar dt = 0.1;
	PetscInt tmax = 20000;
	PetscInt tsave = 100;
	PetscInt loadn = 31;
	PetscScalar loadMax = -8.;
	const auto &hg = hinges_[0];
// 	hg.ks() = Constant::Hinge::Ec;
// 	hg.damping() = 0.001;
	PetscPrintf(PETSC_COMM_WORLD,"  ks = %.3f\n",hg.ks());
	PetscPrintf(PETSC_COMM_WORLD,"  damping = %.3f\n",hg.damping());
	for (PetscInt loadi=0; loadi<loadn; ++loadi)
	{
		PetscScalar loadFrac = static_cast<PetscScalar>(loadi)/static_cast<PetscScalar>(loadn-1);
		PetscScalar load = loadFrac*loadMax*hg.referenceForce();
		PetscScalar normalizedLoad = hg.normalizedLoad(load);
		force_ = EigenWrapper::EigenVec3(0,0,load);
		torque_.setZero();
		dt_ = dt;
		PetscPrintf(PETSC_COMM_WORLD,"\n");
		PetscPrintf(PETSC_COMM_WORLD,"  loadi = %i\n"  ,loadi);
		PetscPrintf(PETSC_COMM_WORLD,"  load  = %.8f\n",load );
		PetscPrintf(PETSC_COMM_WORLD,"  PL^2/(EI) = %.8f\n",normalizedLoad);
		for (PetscInt h=0; h<n_; ++h)
		{
			hinges_[h].name() = "hinge" + std::to_string(h) +  "_load" + std::to_string(loadi);
			sides_ [h].name() = "side"  + std::to_string(h) +  "_load" + std::to_string(loadi);
		}
		for (PetscInt t=0; t<tmax; ++t)
		{
			this->updateHinges();
			this->updateSides();
			if (t % tsave == 0)
			{
				for (const auto& hinge:hinges_)
				{
					hinge.writeInBinaryFormat(t);
				}
				for (const auto& side:sides_)
				{
					side.writeInBinaryFormat(t);
				}
			}
		}
	}
	base_.writeInBinaryFormat(-1);
	return(0);
}


PetscErrorCode Context::validateCapillaryForceModel()
{
	PetscScalar dt = Constant::Time::DeltaT;
	PetscInt tmin  = Constant::Time::TBegin;
	PetscInt tmax  = Constant::Time::TEnd;
	PetscInt tend  = tmax-tmin;
	PetscInt save  = Constant::Time::WriteEvery;
	PetscInt print = Constant::Time::PrintEvery;
	PetscViewer viewer;
	PetscViewerASCIIOpen(PETSC_COMM_WORLD,"force.txt",&viewer);
	PetscViewerASCIIPrintf(viewer,"#iteration Fx Fy Fz Ty Ty Tz V\n");
	for (PetscInt t=0; t<tend; ++t)
	{
		this->updatePhi();
		this->updateContactLine(base_,t);
		this->updateCapillaryForce();
		const auto& fc = force_;
		const auto& tc = torque_;
		PetscViewerASCIIPrintf
		(
			viewer,
			"%i %.8f %.8f %.8f %.8f\n",
			t,
			fc.x(),
			fc.y(),
			fc.z(),
			dropVolume_
		);
		if ( (/*t>0 &&*/ t%save == 0) || t == (tend-1) )
		{
			this->write();
		}
		if ( (/*t>0 &&*/ t%print == 0) || t == (tend-1) )
		{
			this->printInfo(t);
		}
	}
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode Context::fold()
{
	const PetscScalar dt = Constant::Time::DeltaT;
	const PetscInt tmin  = Constant::Time::TBegin;
	const PetscInt tmax  = Constant::Time::TEnd;
	const PetscInt tend  = tmax-tmin;
	const PetscInt tsave  = Constant::Time::WriteEvery;
	const PetscInt tprint = Constant::Time::PrintEvery;
	this->write();
	this->printInfo();
	PetscViewer viewer;
	PetscViewerASCIIOpen(PETSC_COMM_WORLD,"force.txt",&viewer);
	PetscViewerASCIIPrintf(viewer,"#t qx qy qz qw Fx Fy Fz Ty Ty Tz V\n");
	std::vector<EigenVec3> rotationAxis(n_);
	std::vector<EigenVec3> fixedArm    (n_);
	std::vector<EigenVec3> pivotingArm (n_);
	for (PetscInt s=0; s<n_; ++s)
	{
		const auto& side = sides_[s];
		rotationAxis[s] = side.m(1) - side.m(0);
		pivotingArm[s] =  /* COM */ - ( side.m(0) + side.m(1) )/2.; //COM = [0,0,0]^T per definition
		fixedArm[s] = side.centre() - pivotingArm[s];
	}
	for (PetscInt t=0; t<tend; ++t)
	{
		PetscInt tcrop = tend - 5*tsave;
		PetscScalar frac = static_cast<PetscScalar>(t)/static_cast<PetscScalar>(tcrop-1);
		if (frac > 1)
		{
			frac = 1.;
		}
		EigenVec pt0 = sides_[0].Tm(2) + (sides_[0].height()/2) * sides_[0].Tn();
		EigenVec pt1 = sides_[1].Tm(2) + (sides_[1].height()/2) * sides_[1].Tn();
		if ( ( pt1 - pt0 ).norm() < (Constant::Template::Gap)/2. )
		{
			PetscPrintf(PETSC_COMM_WORLD,"  Self folding complete\n");
		}
		else
		{
			for (PetscInt s=0; s<n_; ++s)
			{
				PetscScalar maxAngle = 100 * (PETSC_PI/180.);
				EigenQuat q = EigenWrapper::AngleAxisToQuat(frac*maxAngle,rotationAxis[s]);
				sides_[s].quaternion() = q;
				sides_[s].centre() = fixedArm[s] + q.toRotationMatrix()*pivotingArm[s];
				sides_[s].triangulate();
				this->updatePhiWall();
			}
		}
		this->updatePhi();
		this->updateContactLine(t);
		this->updateCapillaryForce();
		const auto& fc = force_;
		const auto& tc = torque_;
		PetscViewerASCIIPrintf
		(
			viewer,
			"%i %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f\n",
			t,
			sides_[0].quaternion().x(),
			sides_[0].quaternion().y(),
			sides_[0].quaternion().z(),
			sides_[0].quaternion().w(),
		 	fc.x(),
			fc.y(),
			fc.z(),
			tc.x(),
			tc.y(),
			tc.z(),
			dropVolume_
		);
		if ( (t>0 && t%tsave  == 0) || t == (tend-1) )
		{
			this->write();
		}
		if ( (t>0 && t%tprint == 0) || t == (tend-1) )
		{
			this->printInfo(t);
		}
	}
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode Context::run()
{
	PetscPrintf(PETSC_COMM_WORLD,"  Self-folding starts\n");
	const PetscScalar dt = Constant::Time::DeltaT;
	const PetscInt tmin  = Constant::Time::TBegin;
	const PetscInt tmax  = Constant::Time::TEnd;
	const PetscInt tend  = tmax-tmin;
	const PetscInt tsave  = Constant::Time::WriteEvery;
	const PetscInt tprint = Constant::Time::PrintEvery;
	this->printInfo();
	PetscViewer viewer;
	PetscViewerASCIIOpen(PETSC_COMM_WORLD,"force.txt",&viewer);;
	PetscViewerASCIIPrintf(viewer,"#t qx qy qz qw Fx Fy Fz Ty Ty Tz V\n");
	for (PetscInt t=0; t<tend; ++t)
	{
		this->updatePhi();
		this->updateContactLine(t);
		this->updateCapillaryForce();
		const auto& side0 = sides_[0];
		const auto& side1 = sides_[1];
		EigenVec pt0 = side0.Tm(2) + (side0.height()/2.) * side0.Tn();
		EigenVec pt1 = side1.Tm(2) + (side1.height()/2.) * side1.Tn();
		if ( ( pt1 - pt0 ).norm() < (Constant::Template::Gap)/2. )
		{
			PetscPrintf(PETSC_COMM_WORLD,"  Template is folded. Only the drop deformation is solved.\n");
		}
		else
		{
			this->updateHinges();
			this->updateSides();
			this->updateU();
			this->updatePhiWall();
		}
		const auto& fc = force_;
		const auto& tc = torque_;
		PetscViewerASCIIPrintf
		(
			viewer,
			"%i %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f\n",
			t,
			side0.quaternion().x(),
			side0.quaternion().y(),
			side0.quaternion().z(),
			side0.quaternion().w(),
		 	fc.x(),
			fc.y(),
			fc.z(),
			tc.x(),
			tc.y(),
			tc.z(),
			dropVolume_
		);
		if ( (/*t>0 &&*/ t%tsave  == 0) || t == (tend-1) )
		{
			this->write();
		}
		if ( (/*t>0 &&*/ t%tprint == 0) || t == (tend-1) )
		{
			this->printInfo(t);
		}
	}
	PetscViewerDestroy(&viewer);
	PetscPrintf(PETSC_COMM_WORLD,"  Self-folding completed\n");
	return(0);
}


PetscErrorCode Context::equilibriate()
{
	PetscPrintf(PETSC_COMM_WORLD,"  Equilibrium starts\n");
	const PetscScalar dt = Constant::Time::DeltaT;
	const PetscInt tmin  = Constant::Time::TBeginEq;
	const PetscInt tmax  = Constant::Time::TEndEq;
	const PetscInt tend  = tmax-tmin;
	const PetscInt tsave  = Constant::Time::WriteEveryEq;
	const PetscInt tprint = Constant::Time::PrintEveryEq;
	for (PetscInt t=0; t<tend; ++t)
	{
		this->updatePhi();
		if ( (/*t>0 &&*/ t%tsave  == 0) || t == (tend-1) )
		{
			this->write();
		}
		if ( (/*t>0 &&*/ t%tprint  == 0) || t == (tend-1) )
		{
			PetscPrintf(PETSC_COMM_WORLD,"  Iteration = %i\n",t);
		}
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Equilibrium completed\n");
	return(0);
}
