#ifndef CONTEXT_H
#define CONTEXT_H

#include "PetscWrapper.h"
#include "IglWrapper.h"
#include "Constant.h"
#include "Field3D.h"
#include "SolidBody.h"
#include "Hinge.h"
#include "Discretisation3D.h"
#include "Polyline.h"
#include "Utils.h"
#include <vector>


/** \file
 *  \brief Solution context */


/** \class Context
	*  \brief Solution context */
class Context
{

public:

	//Dynamic array
	DM da_;/**< \brief Petsc dynamic array */

	//Grid coordinates
	PetscWrapper::GridCoordinates3D coords_; /**< \brief grid coordinates */
	EigenWrapper::EigenMat flattenedCoords_; /**< \brief grid coordinates rearranged into an Eigen matrix for use in marching cubes */

	//Fields
	ScalarField phi_;     /**< \brief function representing the drop */
	ScalarField phiRHS_;  /**< \brief right hand side of phi */
	ScalarField mu_;      /**< \brief chemical potential */
	ScalarField muBulk_;  /**< \brief bulk part of the chemical potential */
	ScalarField phiWall_; /**< \brief smooth wall sum */
	VectorField u_;       /**< \brief velocity */

	//Template
	PetscInt n_;                         /**< \brief number of sides in the template */
	Prism base_;                         /**< \brief fixed base */
	std::vector<Prism> gaps_;            /**< \brief gaps stopping the drop from migrating downwards */
	std::vector<TriangularPrism> sides_; /**< \brief mobile sides */

	//Hinges
	std::vector<Hinge> hinges_; /**< \brief hinges connecting support and sides */

	//Drop
	PetscScalar isoPhi_;                /**< \brief iso value representing the drop */
	PetscScalar dropVolume_;            /**< \brief drop volume V(t) */
	PetscScalar prescribedDropVolume_;  /**< \brief prescribed drop volume V(t=0) */
	PetscScalar prescribedDropRadius_;  /**< \brief prescribed drop radius R(t=0) */
	PetscScalar dropGrowthCoefficient_; /**< \brief drop growth coefficient */
	PetscScalar contactAngle_;          /**< \brief contact angle */

	//Three phase contact line
	Polylines contactLines_;           /**< \brief three-phase contact lines */
	PetscScalar smoothingCoefficient_; /**< \brief curve smoothing coefficient */
	EigenWrapper::EigenVec3 force_;    /**< \brief capillary force */
	EigenWrapper::EigenVec3 torque_;   /**< \brief capillary torque */

	//Integers
	PetscInt save_; /**< \brief incremental save index */

	//Scalars
	PetscScalar dt_;            /**< \brief time step */
	PetscScalar dx_;            /**< \brief dx */
	PetscScalar dx2_;           /**< \brief dx^2 */
	PetscScalar dx3_;           /**< \brief dx^3 */
	PetscScalar invDx1_;        /**< \brief 1 / dx */
	PetscScalar invDx2_;        /**< \brief 1 / (dx^2) */
	PetscScalar invDx3_;        /**< \brief 1 / (dx^3) */
	PetscScalar xi_;            /**< \brief xi */
	PetscScalar xi2_;           /**< \brief xi^2 */
	PetscScalar dtInvDx1_;      /**< \brief dt/dx */
	PetscScalar dtInvPe_;       /**< \brief dt/Pe */
	PetscScalar dtInvDx2InvPe_; /**< \brief dt / (Pe . dx^2) */
	PetscScalar dtInvWe_;       /**< \brief dt / We */

public:

	Context(DM,PetscInt n);
	~Context();
	PetscErrorCode write();
	PetscErrorCode read(PetscInt);
	PetscErrorCode printInfo(PetscInt t=-1);
	PetscErrorCode initialize(PetscScalar,PetscScalar);
	PetscErrorCode initialize();
	PetscErrorCode initializeTemplate(PetscScalar,PetscScalar);
	PetscErrorCode initializeTemplate();
	PetscErrorCode initializeDrop(PetscScalar radius);
	PetscErrorCode initializeHinges();
	PetscErrorCode updateHinges();
	PetscErrorCode updateSides();
	PetscErrorCode updateU();
	PetscErrorCode updatePhiWall();
	PetscErrorCode updatePhi();
	PetscErrorCode updateContactLine(const EigenWrapper::EigenMat&,const EigenWrapper::EigenMatI&,PetscInt);
	PetscErrorCode updateContactLine(const SolidBody &sb,PetscInt);
	PetscErrorCode updateContactLine(PetscInt);
	PetscErrorCode updateCapillaryForce();
	PetscErrorCode equilibriate();
	PetscErrorCode validateHingeModel();
	PetscErrorCode validateCapillaryForceModel();
	PetscErrorCode fold();
	PetscErrorCode run();
};


#endif
