#ifndef DISCRETISATION3D_H
#define DISCRETISATION3D_H


/** \file
 *  \brief Collection of space discretisation schemes */


#include "petsc.h"
#include "Utils.h"


namespace
{
	typedef PetscScalar(*ThreeDimensionalGradOperator)(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt);
	typedef PetscScalar(*ThreeDimensionalTransposedLaplacianOperator)(const PetscScalar *const, const PetscScalar *const *const *const *const,PetscInt,PetscInt,PetscInt);
	typedef PetscScalar(*ThreeDimensionalGrad2Operator)(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt);
}


/** \namespace Discretisation3D
 *  \brief Collection of space discretisation schemes */
namespace Discretisation3D
{


	/** \namespace Discretisation3D::SecondOrder
	 *  \brief Collection of second order space discretisation schemes */
	namespace SecondOrder
	{


		/** \brief 2-point stencil Grad_x(f)
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \partial_x f \right|_{k,j,i-1/2} \f$ */
		PetscScalar GradX
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 2-point stencil Grad_y(f)
		 *  \param[in] f three-dimensional array
		 * *\param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \partial_y f \right|_{k,j-1/2,i} \f$ */
		PetscScalar GradY
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 2-point stencil Grad_z(f)
		 *  \param[in] f three-dimensional array
		 * *\param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \partial_z f \right|_{k-1/2,j,i} \f$ */
		PetscScalar GradZ
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to second order %Grad */
		const ThreeDimensionalGradOperator GradPtr[] =
		{
			Discretisation3D::SecondOrder::GradX,
			Discretisation3D::SecondOrder::GradY,
			Discretisation3D::SecondOrder::GradZ
		};

		/** \brief 2-point stencil Grad(f)
		 *  \param[in] f   three-dimensional array
		 *  \param[in] dim space dimension
		 *  \param[in] k   node index
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        scalar \f$ \left. \partial_x f \right|_{k    ,j    ,i-1/2} \f$ if dim = 0 else
		 *                        \f$ \left. \partial_y f \right|_{k    ,j-1/2,i    } \f$ if dim = 1 else
		 *                        \f$ \left. \partial_z f \right|_{k-1/2,j    ,i    } \f$ if dim = 2 */
		PetscScalar Grad
		(
			const PetscScalar *const *const *const f,
			PetscInt dim,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil Grad_xx(f)
		 *  \note http://web.media.mit.edu/~crtaylor/calculator.html
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \partial_{xx} f \right|_{k,j,i-1/2} \f$ */
		PetscScalar Grad2X
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil Grad_yy(f)
		 *  \note http://web.media.mit.edu/~crtaylor/calculator.html
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \partial_{yy} f \right|_{k,j-1/2,i} \f$ */
		PetscScalar Grad2Y
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil Grad_zz(f)
		 *  \note http://web.media.mit.edu/~crtaylor/calculator.html
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \partial_{zz} f \right|_{k-1/2,j,i} \f$ */
		PetscScalar Grad2Z
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to second order %Grad_ii */
		const ThreeDimensionalGrad2Operator Grad2Ptr[] =
		{
			Discretisation3D::SecondOrder::Grad2X,
			Discretisation3D::SecondOrder::Grad2Y,
			Discretisation3D::SecondOrder::Grad2Z
		};


		/** \brief 4-point stencil Grad_ii(f)
		 *  \note http://web.media.mit.edu/~crtaylor/calculator.html
		 *  \param[in] f   two-dimensional array
		 *  \param[in] dim space dimension
		 *  \param[in] k   node index
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        scalar \f$ \left. \partial_{xx} f \right|_{k    ,j    ,i-1/2} \f$ if dim = 0 else
		 *                        \f$ \left. \partial_{yy} f \right|_{k    ,j-1/2,i    } \f$ if dim = 1 else
         *                        \f$ \left. \partial_{zz} f \right|_{k-1/2,j    ,i    } \f$ if dim = 2 */
		PetscScalar Grad2
		(
			const PetscScalar *const *const *const f,
			PetscInt dim,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 2-point stencil divergent Div(fx,fy,fz)
		 *  \note \f$ \textbf{f} = ( f_x|_{k,j,i-1/2} , f_y|_{k,j-1/2,i}, f_z|_{k-1/2,j,i}  )^\top \f$
		 *  \note \f$ \partial_x f_x |_{k,j,i} = f_x|_{k    ,j    ,i+1/2} - f_x|_{k    ,j    ,i-1/2} = f_x|_{k  ,j  ,I+1} - f_x|_{k,j,I} \f$
		 *  \note \f$ \partial_y f_y |_{k,j,i} = f_y|_{k    ,j+1/2,i    } - f_y|_{k    ,j-1/2,i    } = f_y|_{k  ,J+1,i  } - f_y|_{k,J,i} \f$
		 * 	\note \f$ \partial_z f_z |_{k,j,i} = f_z|_{k+1/2,j    ,i    } - f_z|_{k-1/2,j    ,i    } = f_z|_{K+1,j  ,i  } - f_z|_{K,j,i} \f$
		 *  \param[in] fx three-dimensional arrays \f$ f_x|_{k    ,j    ,i-1/2} \f$
		 *  \param[in] fy three-dimensional arrays \f$ f_y|_{k    ,j-1/2,i    } \f$
		 *  \param[in] fz three-dimensional arrays \f$ f_y|_{k-1/2,j    ,i    } \f$
		 *  \param[in] k  node index
		 *  \param[in] j  node index
		 *  \param[in] i  node index
		 *  \return       scalar \f$ \left. \nabla \cdot \textbf{f} \right|_{k,j,i} = \partial_x f_x + \partial_y f_y + \partial_z f_z \f$ */
		PetscScalar Divergent
		(
			const PetscScalar *const *const *const fx,
			const PetscScalar *const *const *const fy,
			const PetscScalar *const *const *const fz,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 2-point stencil divergent Div(F)
		 *  \note \f$ \textbf{f} = ( f_x|_{k,j,i-1/2} , f_y|_{k,j-1/2,i}, f_z|_{k-1/2,j,i}  )^\top \f$
		 *  \note \f$ \partial_x f_x |_{k,j,i} = f_x|_{k    ,j    ,i+1/2} - f_x|_{k    ,j    ,i-1/2} = f_x|_{k  ,j  ,I+1} - f_x|_{k,j,I} \f$
		 *  \note \f$ \partial_y f_y |_{k,j,i} = f_y|_{k    ,j+1/2,i    } - f_y|_{k    ,j-1/2,i    } = f_y|_{k  ,J+1,i  } - f_y|_{k,J,i} \f$
		 * 	\note \f$ \partial_z f_z |_{k,j,i} = f_z|_{k+1/2,j    ,i    } - f_z|_{k-1/2,j    ,i    } = f_z|_{K+1,j  ,i  } - f_z|_{K,j,i} \f$
		 *  \param[in] F array of three-dimensional arrays \f$ \textbf{f} = ( f_x|_{k,j,i-1/2} , f_y|_{k,j-1/2,i} , f_z|_{k-1/2,j,i}  )^\top \f$
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \nabla \cdot \textbf{f} \right|_{k,j,i} = \partial_x f_x + \partial_y f_y + \partial_z f_z \f$ */
		PetscScalar Divergent
		(
			const PetscScalar *const *const *const *const F,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 7-point stencil Laplacian Lap(f)
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \nabla^2 f \right|_{k,j,i} \f$ */
		PetscScalar Laplacian
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 7-point stencil Laplacian Lap(eta,f) = Div( eta Grad(f) )
		 *  \param[in] eta interpolated viscosity list \f$ \textbf{\eta} = \left[ \eta_C, \eta_F, \eta_B, \eta_N, \eta_S, \eta_E, \eta_W \right] \f$
		 *  \param[in] f   three-dimensional array
		 *  \param[in] k   node index
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        scalar \f$ \left. \nabla \cdot \left( \eta \nabla f \right) \right|_{k,j,i} \f$ */
		PetscScalar Laplacian
		(
			const PetscScalar *const eta,
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 7-point stencil x-component of transposed Laplacian Div( eta [Grad(F)]^T )
		 *  \param[in] eta one dimensional array of interpolated viscosities \f$ \textbf{\eta} = \left[ \eta_C, \eta_F, \eta_B, \eta_N, \eta_S, \eta_E, \eta_W \right] \f$
		 *  \param[in] F   array of three-dimensional arrays \f$ \textbf{f} = ( f_x|_{k,j,i-1/2},f_y|_{k,j-1/2,i},f_x|_{k-1/2,j,i} )^\top \f$
		 *  \param[in] k   node index
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        x-component of \f$ \left. \nabla \cdot \left[ \eta ( \nabla \textbf{f} )^\top \right] \right|_{k,j,i-1/2} \f$ */
		PetscScalar TransposedLaplacianX
		(
			const PetscScalar *const eta,
			const PetscScalar *const *const *const *const F,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 7-point stencil y-component of transposed Laplacian Div( eta [Grad(F)]^T )
		 *  \param[in] eta one dimensional array of interpolated viscosities \f$ \textbf{\eta} = \left[ \eta_C, \eta_F, \eta_B, \eta_N, \eta_S, \eta_E, \eta_W \right] \f$
		 *  \param[in] F   array of three-dimensional arrays \f$ \textbf{f} = ( f_x|_{k,j,i-1/2},f_y|_{k,j-1/2,i},f_x|_{k-1/2,j,i} )^\top \f$
		 *  \param[in] k   node index
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        y-component of \f$ \left. \nabla \cdot \left[ \eta ( \nabla \textbf{f} )^\top \right] \right|_{k,j-1/2,i} \f$ */
		PetscScalar TransposedLaplacianY
		(
			const PetscScalar *const eta,
			const PetscScalar *const *const *const *const F,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 7-point stencil z-component of transposed Laplacian Div( eta [Grad(F)]^T )
		 *  \param[in] eta one dimensional array of interpolated viscosities \f$ \textbf{\eta} = \left[ \eta_C, \eta_F, \eta_B, \eta_N, \eta_S, \eta_E, \eta_W \right] \f$
		 *  \param[in] F   array of three-dimensional arrays \f$ \textbf{f} = ( f_x|_{k,j,i-1/2},f_y|_{k,j-1/2,i},f_x|_{k-1/2,j,i} )^\top \f$
		 *  \param[in] k   node index
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        z-component of \f$ \left. \nabla \cdot \left[ \eta ( \nabla \textbf{f} )^\top \right] \right|_{k-1/2,j,i} \f$ */
		PetscScalar TransposedLaplacianZ
		(
			const PetscScalar *const eta,
			const PetscScalar *const *const *const *const F,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to transposed Laplacian */
		const ThreeDimensionalTransposedLaplacianOperator TransposedLaplacianPtr[] =
		{
			Discretisation3D::SecondOrder::TransposedLaplacianX,
			Discretisation3D::SecondOrder::TransposedLaplacianY,
			Discretisation3D::SecondOrder::TransposedLaplacianZ
		};


		/** \brief Compact 7-point stencil of transposed Laplacian Div( eta [Grad(F)]^T )
		 *  \param[in] eta one dimensional array of interpolated viscosities \f$ \textbf{\eta} = \left[ \eta_C, \eta_F, \eta_B, \eta_N, \eta_S, \eta_E, \eta_W \right] \f$
		 *  \param[in] F   array of two-dimensional arrays \f$ \textbf{f} = ( f_x|_{k,j,i-1/2},f_y|_{k,j-1/2,i},f_z|_{k-1/2,j,i} )^\top \f$
		 *  \param[in] dim space dimension
		 *  \param[in] k   node index
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        x-component of \f$ \left. \nabla \cdot \left[ \eta ( \nabla \textbf{f} )^\top \right] \right|_{k    ,j    ,i-1/2} \f$ if dim = 0 else
		 *                 y-component of \f$ \left. \nabla \cdot \left[ \eta ( \nabla \textbf{f} )^\top \right] \right|_{k    ,j-1/2,i    } \f$ if dim = 1 else
		 *                 z-component of \f$ \left. \nabla \cdot \left[ \eta ( \nabla \textbf{f} )^\top \right] \right|_{k-1/2,j    ,i    } \f$ if dim = 2 else */
		PetscScalar TransposedLaplacian
		(
			const PetscScalar *const eta,
			const PetscScalar *const *const *const *const F,
			PetscInt dim,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Biharmonic operator Bih(f)
		 *  \note See biharmonic.py
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      scalar \f$ \left. \nabla^4 f \right|_{k,j,i} \f$ */
		PetscScalar Biharmonic
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


	}


	/** \namespace Discretisation3D::FourthOrder
	 *  \brief Collection of fourth order central space discretisation schemes */
	namespace FourthOrder
	{


		/** \brief 4-point stencil Grad_x(f)
		 *  \note Yefet and Petropoulos, A staggered fourth order accurate explicit finite difference scheme for the time domain Maxwell's equations, JCP, 2001
		 *  \copydetails Discretisation3D::SecondOrder::GradX(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar GradX
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil Grad_y(f)
		 *  \note Yefet and Petropoulos, A staggered fourth order accurate explicit finite difference scheme for the time domain Maxwell's equations, JCP, 2001
		 *  \copydetails Discretisation3D::SecondOrder::GradY(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar GradY
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil Grad_z(f)
		 *  \note Yefet and Petropoulos, A staggered fourth order accurate explicit finite difference scheme for the time domain Maxwell's equations, JCP, 2001
		 *  \copydetails Discretisation3D::SecondOrder::GradZ(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar GradZ
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to fourth-order gradient
		 *  \copydetails Discretisation3D::SecondOrder::GradPtr() */
		const ThreeDimensionalGradOperator GradPtr[] =
		{
			Discretisation3D::FourthOrder::GradX,
			Discretisation3D::FourthOrder::GradY,
			Discretisation3D::FourthOrder::GradZ
		};


		/** \brief 4-point stencil Grad(f)
		*   \copydetails Discretisation3D::SecondOrder::Grad(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt,PetscInt) */
		PetscScalar Grad
		(
			const PetscScalar *const *const *const f,
			PetscInt dim,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Compact 19-point stencil Laplacian Lap(f)
		 *  \note  Spotz, 1996, High-order compact formulation for the 3D poisson equation
		 *  \copydetails Discretisation3D::SecondOrder::Laplacian(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar Laplacian
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Biharmonic operator Bih(f)
		 *  \copydetails Discretisation3D::SecondOrder::Biharmonic(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar Biharmonic
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


	}


	/** \namespace Discretisation3D::FifthOrder
	 *  \brief Collection of fifth order central space discretisation schemes */
	namespace FifthOrder
	{


		/** \brief Compact 27-point stencil Laplacian Lap(f)
		 *  \note  Spotz, 1996, High-order compact formulation for the 3D poisson equation
		 *  \copydetails Discretisation3D::SecondOrder::Laplacian(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar Laplacian
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Biharmonic operator Bih(f)
		 *  \copydetails Discretisation3D::SecondOrder::Biharmonic(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar Biharmonic
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


	}

	/** \namespace Discretisation3D::FirstOrderUpwind
	 *  \brief First-order upwind scheme */
	namespace FirstOrderUpwind
	{


		/** \brief First-order upwind scheme U.Grad(f)
		 *  \param[in] uDir convective velocity \f$ \textbf{u}|_{k,j,i} = \left( u_x, u_y, u_z \right )^\top \f$
		 *  \param[in] f    three-dimensional array
		 *  \param[in] k    node index
		 *  \param[in] j    node index
		 *  \param[in] i    node index
		 *  \return         U.Grad(f) = \f$ \left. \textbf{u} \cdot ( \nabla f ) \right|_{k,j,i} \f$ */
		PetscScalar UGrad
		(
			const PetscScalar *const  uDir,
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


	}


	/** \namespace Discretisation3D::SecondOrderUpwind
	 *  \brief Second-order upwind scheme */
	namespace SecondOrderUpwind
	{


		/** \brief Second-order upwind scheme U.Grad(f)
		 *  \copydetails Discretisation3D::FirstOrderUpwind::UGrad(const PetscScalar *const,const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar UGrad
		(
			const PetscScalar *const  uDir,
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


	}


	/** \namespace Discretisation3D::Weno
	 *  \brief %Weno convection scheme */
	namespace Weno
	{


		/** \brief %Weno scheme
		 *  \copydetails Discretisation3D::FirstOrderUpwind::UGrad(const PetscScalar *const,const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar UGrad
		(
			const PetscScalar *const  uDir,
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


	}


}


#endif
