#include <iostream>
#include "Constant.h"
#include "Field3D.h"

static char help[] = "Export scalar.dat to scalar.txt\n\n";

PetscErrorCode PrintUsage()
{
	std::cout
		<< "  USAGE:" << std::endl
		<< "  -h                 print usage" << std::endl
		<< "  -i                 input file" << std::endl
		<< "  -f                 field" << std::endl
		<< "  -v                 3d vector field" << std::endl
		//<< "  -viewer_binary_mpi petsc option" << std::endl
		<< "  -t t0:t1:dt        start, end, every" << std::endl
		<< "  Export field.dat to field.txt" << std::endl;
	return(0);
}


PetscErrorCode GetTimes
(
	PetscInt &t0,
	PetscInt &t1,
	PetscInt &dt
)
{
	PetscBool flg;
	char str[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-t",str,sizeof(str),&flg);
	if (!flg)
	{
		//SETERRQ(PETSC_COMM_WORLD,1,"Pass with -t t0:t1:dt");
		t0 = -1;
		t1 =  0;
		dt =  1;
		return(0);
	}
	char *p = strtok(str,":"); t0 = atoi(p);
	p = strtok(NULL,":"); t1 = (p == NULL) ? t0+1 : atoi(p);
	p = strtok(NULL,":"); dt = (p == NULL) ? 1 : atoi(p);
	return(0);
}


PetscInt main(int argc,char** args)
{
	PetscInitialize(&argc,&args,(char*)0,help);
	PetscBool usage(PETSC_FALSE); PetscOptionsGetBool(NULL,NULL,"-h",&usage,NULL);
	if (usage)
	{
		PrintUsage();
		PetscEnd();
	}
	char input[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-i",input,sizeof(input),NULL);
	if (*input == 0)
	{
		sprintf(input,"Input.in");
	}
	char field[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-f",field,sizeof(field),NULL);
	PetscBool vectorField(PETSC_FALSE); PetscOptionsGetBool(NULL,NULL,"-v",&vectorField,NULL);
	PetscInt t0,t1,dt; GetTimes(t0,t1,dt);
	Constant::ReadInputFile(input);
	DM da;
	DMDACreate3d
	(
		PETSC_COMM_WORLD,     //MPI communicator
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the x-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the y-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the z-direction
		DMDA_STENCIL_BOX,     //stencil type (DMDA_STENCIL_BOX or DMDA_STENCIL_STAR)
		(Constant::Grid::Nx), //global dimension in x-direction of the array
		(Constant::Grid::Ny), //global dimension in y-direction of the array
		(Constant::Grid::Nz), //global dimension in y-direction of the array
		PETSC_DECIDE,         //corresponding number of processors in x-dimension
		PETSC_DECIDE,         //corresponding number of processors in y-dimension
		PETSC_DECIDE,         //corresponding number of processors in z-dimension
		1,                    //number of degrees of freedom per node (i.e. number of field components)
		1,                    //stencil width (1, 2 or 3), 3 for WENO
		NULL,                 //arrays containing the number of nodes in each cell along the x coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the y coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the z coordinates
		&da                   //dynamic array
	);
	DMSetUp(da);
	DMDASetUniformCoordinates
	(
		da,
		0.,(Constant::Grid::L[0]),
		0.,(Constant::Grid::L[1]),
		0.,(Constant::Grid::L[2])
	);
	if (vectorField)
	{
		VectorField f;
		f.create3D(da,field);
		for (PetscInt t=t0; t<t1; t+=dt)
		{
			f.readInBinaryFormat(t);
			f.writeInVtsFormatWithXYZSubscripts(t);
		}
		f.destroy();
	}
	else
	{
		ScalarField f;
		f.create(da,field);
		for (PetscInt t=t0; t<t1; t+=dt)
		{
			f.readInBinaryFormat(t);
			f.writeInVtsFormat(t);
		}
		f.destroy();
	}
	PetscFinalize();
	return(0);
}
