#include <iostream>
#include "Constant.h"
#include "SolidBody.h"

static char help[] = "Export solid body data to txt, phi field to VTS, and mesh to STL\n\n";

PetscErrorCode PrintUsage()
{
	std::cout
		<< "  USAGE:" << std::endl
		<< "  -h     print usage" << std::endl
		<< "  -i     input file" << std::endl
		<< "  -f     solid body name" << std::endl
		<< "  -tp    triangularPrism" << std::endl
		<< "  -n     number of prism sides" << std::endl
		<< "  -stl   export stl file" << std::endl
		<< "  -vts   export vts file" << std::endl
		<< "  -txt   export txt file" << std::endl
		<< "  -t t0:t1:dt start, end, every" << std::endl
		<< "  Export solid body to txt and VTS format" << std::endl
	;
	return(0);
}

PetscErrorCode GetTimes
(
	PetscInt &t0,
	PetscInt &t1,
	PetscInt &dt
)
{
	PetscBool flg;
	char str[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-t",str,sizeof(str),&flg);
	if (!flg)
	{
		//SETERRQ(PETSC_COMM_WORLD,1,"Pass with -t t0:t1:dt");
		t0 = -1;
		t1 =  0;
		dt =  1;
		return(0);
	}
	char *p = strtok(str,":"); t0 = atoi(p);
	p = strtok(NULL,":"); t1 = (p == NULL) ? t0+1 : atoi(p);
	p = strtok(NULL,":"); dt = (p == NULL) ? 1 : atoi(p);
	return(0);
}

PetscInt main(int argc,char** args)
{
	PetscInitialize(&argc,&args,(char*)0,help);
	PetscBool usage(PETSC_FALSE); PetscOptionsGetBool(NULL,NULL,"-h",&usage,NULL);
	if (usage)
	{
		PrintUsage();
		PetscEnd();
	}
	char input[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-i",input,sizeof(input),NULL);
	if (*input == 0)
	{
		sprintf(input,"Input.in");
	}
	char solid[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-f",solid,sizeof(solid),NULL);
	PetscBool tp (PETSC_FALSE); PetscOptionsGetBool(NULL,NULL,"-tp",&tp,NULL);
	PetscBool stl(PETSC_FALSE); PetscOptionsGetBool(NULL,NULL,"-stl",&stl,NULL);
	PetscBool vts(PETSC_FALSE); PetscOptionsGetBool(NULL,NULL,"-vts",&vts,NULL);
	PetscBool txt(PETSC_FALSE); PetscOptionsGetBool(NULL,NULL,"-txt",&txt,NULL);
	PetscInt sides(-1); PetscOptionsGetInt(NULL,NULL,"-n",&sides,NULL);
	PetscInt t0,t1,dt; GetTimes(t0,t1,dt);
	Constant::ReadInputFile(input);
	if (sides<0)
	{
		sides = Constant::Template::Sides;
	}
	DM da;
	DMDACreate3d
	(
		PETSC_COMM_WORLD,     //MPI communicator
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the x-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the y-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the z-direction
		DMDA_STENCIL_BOX,     //stencil type (DMDA_STENCIL_BOX or DMDA_STENCIL_STAR)
		(Constant::Grid::Nx), //global dimension in x-direction of the array
		(Constant::Grid::Ny), //global dimension in y-direction of the array
		(Constant::Grid::Nz), //global dimension in y-direction of the array
		PETSC_DECIDE,         //corresponding number of processors in x-dimension
		PETSC_DECIDE,         //corresponding number of processors in y-dimension
		PETSC_DECIDE,         //corresponding number of processors in z-dimension
		1,                    //number of degrees of freedom per node (i.e. number of field components)
		1,                    //stencil width (1, 2 or 3), 3 for WENO
		NULL,                 //arrays containing the number of nodes in each cell along the x coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the y coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the z coordinates
		&da                   //dynamic array
	);
	DMSetUp(da);
	DMDASetUniformCoordinates
	(
		da,
		0.,(Constant::Grid::L[0]),
		0.,(Constant::Grid::L[1]),
		0.,(Constant::Grid::L[2])
	);
	if (tp)
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Triangular prism\n");
		TriangularPrism prism;
		prism.create(da,solid);
		for (PetscInt t=t0; t<t1; t+=dt)
		{
			prism.readInBinaryFormat(t);
			prism.triangulate();
			if (vts)
			{
				prism.setPhi(Constant::Fluid::Xi);
				prism.writeInVtsFormat(t);
			}
			if (stl){prism.writeInStlFormat(t);}
			if (txt){prism.writeInAsciiFormat(t);}
		}
		prism.destroy();
	}
	else
	{
		PetscPrintf(PETSC_COMM_WORLD,"  %i-sided prism\n",sides);
		Prism prism(sides);
		prism.create(da,solid);
		for (PetscInt t=t0; t<t1; t+=dt)
		{
			prism.readInBinaryFormat(t);
			prism.triangulate();
			if (vts)
			{
				prism.setPhi(Constant::Fluid::Xi);
				prism.writeInVtsFormat(t);
			}
			if (stl){prism.writeInStlFormat(t);}
			if (txt){prism.writeInAsciiFormat(t);}
		}
		prism.destroy();
	}
	PetscFinalize();
	return(0);
}
