#ifndef FIELD3D_H
#define FIELD3D_H

#include "PetscWrapper.h"

/** \file
 *  \brief %Field functions conveniently grouping a global Petsc vector, a local Petsc vector and a multidimensional C-array */

/** \class ScalarField
 *  \brief Structure holding a global PETSC vector, a local PETSC vector, and a 3-dimensional C-array */
class ScalarField
{

private:

	/** \brief Dynamic array */
	DM da_;

	/** \brief Global Petsc vector */
	Vec g_;

	 /** \brief Local Petsc vector */
	Vec l_;

	/** \brief 3-dimensional C-array */
	PetscScalar ***a_;

	public:

	/** \brief Dummy constructor */
	ScalarField(){}

	/** \brief Dummy destructor */
	~ScalarField(){}

	/** \brief Reference to global Petsc vector */
	Vec &g()
	{
		return g_;
	}

	/** \brief Reference to 3-dimensional C-array */
	PetscScalar ***a(){return a_;}

	/** \brief Reference to 3-dimensional C-array by overloading parenthesis operator */
	PetscScalar ***operator()(){return a_;}

	/** \brief Reference access to a[k][j][i]
	 *  \param[in] k node index
	 *  \param[in] j node index
	 *  \param[in] i node index
	 *  \return      reference to a[k][j][i] */
	PetscScalar &a(PetscInt k,PetscInt j,PetscInt i){return a_[k][j][i];}

	/** \brief Const reference access to a[k][j][i] */
	const PetscScalar &a(PetscInt k,PetscInt j,PetscInt i) const {return a_[k][j][i];}

	/** \brief Reference access to a[k][j][i] by overloading parenthesis operator
	 *  \param[in] k node index
	 *  \param[in] j node index
	 *  \param[in] i node index
	 *  \return      reference to a[k][j][i] */
	PetscScalar &operator()(PetscInt k,PetscInt j,PetscInt i){return a_[k][j][i];}

	/** \brief Const reference access to a[k][j][i] */
	const PetscScalar &operator()(PetscInt k,PetscInt j,PetscInt i) const {return a_[k][j][i];}

	/** \brief Constructor
	 *  \param[in] da   DMDA distributed array
	 *  \param[in] name global Petsc vector name */
	PetscErrorCode create(DM da,const std::string &name);

	/** \brief Destructor */
	PetscErrorCode destroy();

	/** \brief Check values */
	PetscErrorCode check();

	/** \brief Write %ScalarField in binary format
	 *  \param[in] t time index */
	PetscErrorCode writeInBinaryFormat(PetscInt t=-1) const;

	/** \brief Write %ScalarField in VTS format
	 *  \param[in] t time index */
	PetscErrorCode writeInVtsFormat(PetscInt t=-1) const;

	/** \brief Read %ScalarField in binary format
	 *  \param[in] t time index */
	PetscErrorCode readInBinaryFormat(PetscInt t=-1);

	/** \brief Get multidimensional C-array */
	PetscErrorCode getArray();

	/** \brief Restore multidimensional C-array */
	PetscErrorCode restoreArray();

	/** \brief Get multidimensional C-array containing ghost points */
	PetscErrorCode getGhostedArray();

	/** \brief Restore multidimensional C-array containing ghost points */
	PetscErrorCode restoreGhostedArray();

	/** \brief Copy one dimensional array with extra space for ghosted values at positions kw+1, jw+1 and iw+1 */
	PetscErrorCode flattenMarchingCubeArrayTo(std::vector<PetscScalar> &v,PetscInt &nx,PetscInt &ny,PetscInt &nz);

	/** \brief Copy the global Petsc vector
	 *  \param[out] to %ScalarField */
	PetscErrorCode copyTo(ScalarField &to);

	/** \brief Set elements of the global Petsc vector to a given value
	 *  \param[in] s scalar */
	PetscErrorCode set(PetscScalar s);

	/** \brief Set elements of the global Petsc vector to zero */
	PetscErrorCode zero();

	/** \brief Sum of the global Petsc vector elements
	 *  \param[in] s sum */
	PetscErrorCode sum(PetscScalar &s);

	/** \brief Arithmetic mean of global Petsc vector elements
	 *  \param[out] m arithmetic mean */
	PetscErrorCode arithmeticMean(PetscScalar &m);

};


/** \class VectorField
 *  \brief Structure holding an array of global PETSC vectors, an array of local PETSC vectors, and an array of 3-dimensional C-arrays */
class VectorField
{

private:

	/** \brief Dynamic array */
	DM da_;

	/** \brief vector size */
	PetscInt n_;

	/** \brief array of global Petsc vectors */
	Vec *g_;

	/** \brief array of local Petsc vectors */
	Vec *l_;

	 /** \brief array of 3-dimensional C-arrays */
	PetscScalar ****a_;

public:

	/** \brief Dummy constructor */
	VectorField(){}

	/** \brief Dummy destructor */
	~VectorField(){}

	/** \brief Const reference to array size */
	const PetscInt &n() const {return n_;}

	/** \brief Reference to global Petsc vector */
	Vec *g(){return g_;}

	/** \brief Reference to array of 3-dimensional C-arrays */
	PetscScalar ****a(){return a_;}

	/** \brief Reference to array of 3-dimensional C-arrays by overloading parenthesis operator */
	PetscScalar ****operator()(){return a_;}

	/** \brief Reference to 3-dimensional C-array a[dim]
	 *  \param[in] dim space dimension */
	PetscScalar ***a(PetscInt dim){return a_[dim];}

	/** \brief Reference to 3-dimensional C-array a[dim] by overloading parenthesis operator
	 *  \param[in] dim space dimension */
	PetscScalar ***operator()(PetscInt dim){return a_[dim];}

	/** \brief Reference to a[dim][k][j][i]
	 *  \param[in] dim dimension
	 *  \param[in] k   node index
	 *  \param[in] j   node index
	 *  \param[in] i   node index */
	PetscScalar &a(PetscInt dim,PetscInt k,PetscInt j,PetscInt i){return a_[dim][k][j][i];}

	/** \brief Const reference to a[dim][k][j][i] */
	const PetscScalar &a(PetscInt dim,PetscInt k,PetscInt j,PetscInt i) const {return a_[dim][k][j][i];}

	/** \brief Reference a[dim][k][j][i] by overloading parenthesis operator
	 *  \param[in] dim dimension
	 *  \param[in] k   node index
	 *  \param[in] j   node index
	 *  \param[in] i   node index */
	PetscScalar &operator()(PetscInt dim,PetscInt k,PetscInt j,PetscInt i){return a_[dim][k][j][i];}

	/** \brief Const reference to a[dim][k][j][i] by overloading parenthesis operator */
	const PetscScalar &operator()(PetscInt dim,PetscInt k,PetscInt j,PetscInt i) const {return a_[dim][k][j][i];}

	/** \brief Reference to a[0]  */
	PetscScalar ***x(){return a_[0];}

	/** \brief Reference to a[1] */
	PetscScalar ***y(){return a_[1];}

	/** \brief Reference to a[2] */
	PetscScalar ***z(){return a_[2];}

	/** \brief Reference to a[0][k][j][i]
	  * \param[in] k node index
	  * \param[in] j node index
	  * \param[in] i node index */
	PetscScalar &x(PetscInt k,PetscInt j,PetscInt i){return a_[0][k][j][i];}

	/** \brief Const reference to a[0][k][j][i]
	  * \param[in] k node index
	  * \param[in] j node index
	  * \param[in] i node index */
	const PetscScalar &x(PetscInt k,PetscInt j,PetscInt i) const {return a_[0][k][j][i];}

	/** \brief Reference to a[1][k][j][i]
	  * \param[in] k node index
	  * \param[in] j node index
	  * \param[in] i node index */
	PetscScalar &y(PetscInt k,PetscInt j,PetscInt i){return a_[1][k][j][i];}

	/** \brief Const reference to a[1][k][j][i]
	  * \param[in] k node index
	  * \param[in] j node index
	  * \param[in] i node index */
	const PetscScalar &y(PetscInt k,PetscInt j,PetscInt i) const {return a_[1][k][j][i];}

	/** \brief Reference to a[2][k][j][i]
	  * \param[in] k node index
	  * \param[in] j node index
	  * \param[in] i node index */
	PetscScalar &z(PetscInt k,PetscInt j,PetscInt i){return a_[2][k][j][i];}

	/** \brief Const reference to a[2][k][j][i]
	  * \param[in] k node index
	  * \param[in] j node index
	  * \param[in] i node index */
	const PetscScalar &z(PetscInt k,PetscInt j,PetscInt i) const {return a_[2][k][j][i];}

	/** \brief Constructor
	 *  \param[in] da   DMDA distributed array
	 *  \param[in] name global Petsc vector name
	 *  \param[in] n    size */
	PetscErrorCode create(DM da,const std::string &name,PetscInt n);

	/** \brief Three-dimensional constructor
	 *  \param[in] da   DMDA distributed array
	 *  \param[in] name global Petsc vector name */
	PetscErrorCode create3D(DM da,const std::string &name);

	/** \brief Destructor */
	PetscErrorCode destroy();

	/** \brief Check values */
	PetscErrorCode check();

	/** \brief Write %VectorField in binary format
	 *  \param[in] t time index */
	PetscErrorCode writeInBinaryFormat(PetscInt t=-1) const;

	/** \brief Write %VectorField in VTS format
	 *  \param[in] t time index */
	PetscErrorCode writeInVtsFormat(PetscInt t=-1) const;

	/** \brief Write %VectorField in VTS format with x,y,z subscripts
	 *  \param[in] t time index */
	PetscErrorCode writeInVtsFormatWithXYZSubscripts(PetscInt t=-1) const;

	/** \brief Read %VectorField in binary format
	 *  \param[in] t time index */
	PetscErrorCode readInBinaryFormat(PetscInt t=-1);

	/** \brief Get array of multidimensional C-arrays */
	PetscErrorCode getArray();

	/** \brief Restore array of multidimensional C-arrays */
	PetscErrorCode restoreArray();

	/** \brief Get array of multidimensional C-arrays containing ghost points */
	PetscErrorCode getGhostedArray();

	/** \brief Restore array of multidimensional C-arrays containing ghost points */
	PetscErrorCode restoreGhostedArray();

	/** \brief Copy the array of global Petsc vectors
	*  \param[out] to %VectorField */
	PetscErrorCode copyTo(VectorField &to);

	/** \brief Set elements of the global Petsc vector to a given value
	 *  \param[in] s scalar */
	PetscErrorCode set(PetscScalar s);

	/** \brief Set elements of the global Petsc vectors to a given value
	 *  \param[in] S one-dimensional scalar array S = [sa,sb,...] */
	PetscErrorCode set(const PetscScalar *const S);

	/** \brief Set elements of the global Petsc vector to zero */
	PetscErrorCode zero();

	/** \brief Sum of global Petsc vector elements
	 *  \param[out] S one-dimensional sum array */
	PetscErrorCode sum(PetscScalar *S);

	/** \brief Arithmetic mean of global Petsc vector elements
	 *  \param[out] M one-dimensional arithmetic mean array */
	PetscErrorCode arithmeticMean(PetscScalar *M);

	/** \brief Sum global Petsc vectors and copy the resulting Petsc vector to a %ScalarField
	 *  \param[out] f %ScalarField */
	PetscErrorCode addTo(ScalarField &f);

};

/** \brief Get multidimensional C-array(s)
 *  \param[in] f field */
template<typename T> PetscErrorCode GetArray(T &f);

/** \brief Restore multidimensional C-array(s)
 *  \param[in] f field */
template<typename T> PetscErrorCode RestoreArray(T &f);

/** \brief Get multidimensional C-array(s) with ghost points
 *  \param[in] f field */
template<typename T> PetscErrorCode GetGhostedArray(T &f);

/** \brief Restore multidimensional C-array(s) with ghost points
 *  \param[in] f field */
template<typename T> PetscErrorCode RestoreGhostedArray(T &f);

/** \brief Variadic function getting (n) multidimensional C-arrays */
template<typename T,typename ... Ts> PetscErrorCode GetArray(T &first,Ts &...args);

/** \brief Variadic function restoring (n) multidimensional C-arrays */
template<typename T,typename ... Ts> PetscErrorCode RestoreArray(T &first,Ts &...args);

/** \brief Variadic function getting (n) multidimensional C-arrays  with ghost points */
template<typename T,typename ... Ts> PetscErrorCode GetGhostedArray(T &first,Ts &...args);

/** \brief Variadic function restoring (n) multidimensional C-arrays  with ghost points */
template<typename T,typename ... Ts> PetscErrorCode RestoreGhostedArray(T &first,Ts &...args);


#endif
