#include "FitPackWrapper.h"
#include <algorithm>


extern "C"
{

	/** \brief Smoothing of an open parametric curve
	 *  \param [in]  iopt Flag.
	 *                    With iopt = -1, a weighted least-squares spline curve is determined.
	 *                    With iopt =  0 or 1, a smoothing spline curve is determined.
	 *                    opti = 0 indicates, the routine starts with am initial set of knots t(i) set by parcur
	 *                    opti = 1 indicates, the routine will continue with the knots found at the last call of the routine. attention: another call is needed here
	 *  \param [in]  ipar Flag
	 *                    With ipar = 0, the parameter values u(i),ub and ue are to be calculated by parcur (default)
	 *                    With ipar = 1, the user will supply these
	 *  \param [in]  idim Dimension of the curve, default is 3
	 *  \param [in]  m    Number of geometrical points
	 *  \param [in]  u    Array of parameter values (optional), with ipar = 0, the values u(i) are determined by parcur
	 *  \param [in]  mx   Dimension of the array m, in 3D, it is 3 x number of points
	 *  \param [in]  x    Array of dimension at least idim*m, x(idim*(i-1)+j) must contain the j-th coordinate of the i-th data point for i=1,2,...,m and j=1,2,...,idim
	 *  \param [in]  w    Weight coefficients, array of dimension m, default is unity
	 *  \param [in]  ub   First end-point of the parameters interval, default is u[0], b=begin
	 *  \param [in]  ue   Second end-point of the parameters interval, default is u[1], e=end
	 *  \param [in]  k    Degree of the splines, default is 3
	 *  \param [in]  s    Smoothing factor
	 *  \param [in]  nest An over-estimate of the total number of knots of the spline to help in determining the storage space. Always large enough is nest = m+k+1
	 *  \param [out] n    Total number of knots of the smoothing spline curve returned if iopt >=0. In case iopt=-1, the value of n must be specified on entry
	 *  \param [out] t    Knots of the spline curve
	 *  \param [in]  nc   Dimension of the array c
	 *  \param [in]  c    B-spline coefficients
	 *  \param [out] fp   Weighted sum of squared residuals of the spline curve returned
	 *  \param [in]  wrk  Array used as working space
	 *  \param [in]  lwrk Dimension of array wrk
	 *  \param [in]  iwrk Array used as working space
	 *  \param [out] ier  Unless the routine detects an error, ier contains a non-positive value on exit */
	void parcur_
	(
		int   &iopt,
		int   &ipar,
		int   &idim,
		int   &m,
		float *u,
		int   &mx,
		float *x,
		float *w,
		float &ub,
		float &ue,
		int   &k,
		float &s,
		int   &nest,
		int   &n,
		float *t,
		int   &nc,
		float *c,
		float &fp,
		float *wrk,
		int   &lwrk,
		int   *iwrk,
		int   &ier
	);

	/** \brief Smoothing of a closed parametric curve
	 *  \param [in]  iopt Flag.
	 *                    With iopt = -1,a weighted least-squares closed spline curve is determined.
	 *                    With iopt =  0 or 1, a smoothing closed spline curve is determined.
	 *                    opti = 0 indicates, the routine starts with an initial set of knots t(i) = f(u)
	 *                    opti = 1 indicates, the routine will continue with the knots found at the last call. A call with iopt=1 must always be immediately preceded by another call with iopt=1 or iopt=0.
	 *  \param [in]  ipar Flag.
	 *                    With ipar = 0, the parameter values u(i) are to be calculated by clocur
	 *                    With ipar = 1, the user will supply the parameter values u(i)
	 *  \param [in]  idim Dimension of the curve, default is 3
	 *  \param [in]  m    Number of geometrical points
	 *  \param [in]  u    Array of parameter values.  With ipar=0, on exit, the array will contain the values u(i) as determined by clocur
	 *  \param [in]  mx   Dimension of the array m, in 3D, it is 3 x number of points
	 *  \param [in]  x    Array of dimension at least idim*m, x(idim*(i-1)+j) must contain the j-th coordinate of the i-th data point for i=1,2,...,m and j=1,2,...,idim
	 *  \param [in]  w    Weight coefficients, array of dimension m, default is unity
	 *  \param [in]  k    Degree of the splines, default is 3
	 *  \param [in]  s    Smoothing factor
	 *  \param [in]  nest An over-estimate of the total number of knots of the spline to help in determining the storage space. Always large enough is nest = m+k+1
	 *  \param [out] n    Total number of knots of the smoothing spline curve returned if iopt >=0. In case iopt=-1, the value of n must be specified on entry
	 *  \param [out] t    Knots of the spline curve
	 *  \param [in]  nc   Dimension of the array c
	 *  \param [in]  c    B-spline coefficients
	 *  \param [out] fp   Weighted sum of squared residuals of the spline curve returned
	 *  \param [in]  wrk  Array used as working space
	 *  \param [in]  lwrk Dimension of array wrk
	 *  \param [in]  iwrk Array used as working space
	 *  \param [out] ier  Unless the routine detects an error, ier contains a non-positive value on exit */
	void clocur_
	(
		int   &iopt,
		int   &ipar,
		int   &idim,
		int   &m,
		float *u,
		int   &mx,
		float *x,
		float *w,
		int   &k,
		float &s,
		int   &nest,
		int   &n,
		float *t,
		int   &nc,
		float *c,
		float &fp,
		float *wrk,
		int   &lwrk,
		int   *iwrk,
		int   &ier
	);

	/** \brief Evaluation of a spline function
	 *  \param[in]  t   Array of length n, which contains the position of the knots.
	 *  \param[in]  n   Integer giving the total number of knots of s(x).
	 *  \param[in]  c   Array of length n, which contains the b-spline coefficients.
	 *  \param[in]  k   Integer giving the degree of s(x).
	 *  \param[in]  x   Array of length m, which contains the points where s(x) must
	 *  \param[out] y   Array of length m, giving the value of s(x) at the different points.
	 *  \param[in]  m   Integer giving the number of points where s(x) must be evaluated.
	 *  \param[out] ier Error flag, ier = 0 : normal return, ier =10 : invalid input data */
	void splev_
	(
		float *t,
		int   &n,
		float *c,
		int   &k,
		float *x,
		float *y,
		int   &m,
		int   &ier
	);

}


int FitPackWrapper::SmoothPolyline
(
	const Eigen::MatrixXf &mat,
	Eigen::MatrixXf &mats,
	float sf,
	bool isClosed,
	bool& error
)
{
	error = false;
	int    iopt = 0;
	int    ipar = 0;
	int    idim = mat.cols();
	int    m    = mat.rows();
	int    mx   = (idim*m);
	int    k    = 3;
	int    nest = std::max((isClosed?(m+2*k):(m+k+1)),(2*k+3));
	int    n    = -1; //dummy value
	int    nc   = (nest*idim);
	int    lwrk = isClosed ? m*(k + 1) + nest*(7 + idim + 5*k) : m*(k + 1) + nest*(6 + idim + 3*k);
	int    ier  = -1; //dummy value
	int   *iwrk = new int[nest];
	float  ub   = -1; //dummy value
	float  ue   = -1; //dummy value
	float  s   = sf;
	float  fp  = -1;  //dummy value
	float *u   = new float[m];
	float *x   = new float[mx];
	float *w   = new float[m];
	float *t   = new float[nest];
	float *c   = new float[nc];
	float *wrk = new float[lwrk];
	for (int dim=0; dim<idim; ++dim)
	{
		for (int pt=0; pt<m; ++pt)
		{
			int i = idim*pt + dim;
			x[i] = mat(pt,dim);
		}
	}
	for (int i=0; i<m; ++i)
	{
		w[i] = 1.;
	}
	if (isClosed)
	{
		clocur_(iopt,ipar,idim,m,u,mx,x,w      ,k,s,nest,n,t,nc,c,fp,wrk,lwrk,iwrk,ier);
	}
	else
	{
		parcur_(iopt,ipar,idim,m,u,mx,x,w,ub,ue,k,s,nest,n,t,nc,c,fp,wrk,lwrk,iwrk,ier);
		int lc = n-k-1;
		int first = 0;
		int last = lc-1;
		for (int dim=0; dim<idim; ++dim)
		{
			c[first+dim*n] = mat(0  ,dim);
			c[last +dim*n] = mat(m-1,dim);
		}
		//for (int i=0; i<lc; i++)
		//{
		//	std::cout << "  c[" << i << "] = (" << c[i    ] << " " << c[i+n  ] << " " << c[i+2*n] << ")" << std::endl;
		//}
	}
	if (ier>0)
	{
		error = true;
		if (isClosed)
		{
			std::cout << "  Error (ier=" << ier << ") returned by fortan function clocur" << std::endl;
		}
		else
		{
			std::cout << "  Error (ier=" << ier << ") returned by fortan function parcur" << std::endl;
		}
// 		std::cout
// 			<< "  Fitpack parameters: "
// 			<< "iopt=" << iopt << " "
// 			<< "ipar=" << ipar << " "
// 			<< "idim=" << idim << " "
// 			<< "m="    << m    << " "
// 			<< "mx="   << mx   << " "
// 			<< "k="    << k    << " "
// 			<< "nest=" << nest << " "
// 			<< "n="    << n    << " "
// 			<< "nc="   << nc   << " "
// 			<< "lwrk=" << lwrk << " "
// 			<< "ier="  << ier  << " "
// 			<< "ub="   << ub   << " "
// 			<< "ue="   << ue   << " "
// 			<< "s="    << s    << " "
// 			<< "fp="   << fp   << std::endl;
// 			std::cout << " x = [ ";
// 			for (int i=0; i<mx-1; ++i)
// 			{
// 				std::cout << x[i] << " ";
// 			}
// 			std::cout << x[mx-1] << "]" << std::endl;
	}
	int    ms = m;
	float *cs = new float[n];
	float *xs = new float[ms];
	float *ys = new float[ms];
	for (int i=0; i<ms; i++)
	{
		xs[i] = static_cast<float>(i) / static_cast<float>(ms-1);
	}
	mats = Eigen::MatrixXf(ms,idim);
	for (int dim=0; dim<idim; ++dim)
	{
		for (int i=0; i<n; ++i)
		{
			cs[i] = c[i + dim*n];
		}
		splev_(t,n,cs,k,xs,ys,ms,ier);
		if (ier>0)
		{
			error = true;
			std::cout << "  Error returned by fortan function splev"  << std::endl;
		}
		for (int pt=0; pt<ms; ++pt)
		{
			mats(pt,dim) = ys[pt];
		}
	}
	delete [] iwrk;
	delete [] u;
	delete [] x;
	delete [] w;
	delete [] t;
	delete [] c;
	delete [] wrk;
	delete [] cs;
	delete [] xs;
	delete [] ys;
	return(0);
}


template<typename Type>
int FitPackWrapper::SmoothPolyline
(
	const Eigen::Matrix<Type,Eigen::Dynamic,Eigen::Dynamic> &mat,
	Eigen::Matrix<Type,Eigen::Dynamic,Eigen::Dynamic> &mats,
	Type s,
	bool isClosed,
	bool& error
)
{
	Eigen::MatrixXf matf
	(
		mat.rows(),
		mat.cols()
	);
	for (int k=0; k<matf.size(); k++)
	{
		//https://stackoverflow.com/questions/16283000/most-efficient-way-to-loop-through-an-eigen-matrix
		*(matf.data() + k) = static_cast<float>( *(mat.data() + k) );
	}
	Eigen::MatrixXf matfs;
	FitPackWrapper::SmoothPolyline(matf,matfs,s,isClosed,error);
	mats = Eigen::Matrix<Type,Eigen::Dynamic,Eigen::Dynamic>
	(
		matfs.rows(),
		matfs.cols()
	);
	for (int k=0; k<mats.size(); k++)
	{
		*(mats.data() + k) = static_cast<Type>( *(matfs.data() + k) );
	}
	return (0);
}


template int FitPackWrapper::SmoothPolyline<PetscScalar>
(
	const Eigen::Matrix<PetscScalar,Eigen::Dynamic,Eigen::Dynamic> &,
	Eigen::Matrix<PetscScalar,Eigen::Dynamic,Eigen::Dynamic> &,
	PetscScalar,
	bool,
	bool&
);
