#ifndef HINGE_H
#define HINGE_H

#include "EigenWrapper.h"
#include "PetscWrapper.h"
#include <vector>

/** \class Bead */
class Bead
{
private:
	/** \brief bead density */
	PetscScalar density_ = 1.;

	/** \brief bead radius */
	PetscScalar radius_ = 1.;

	/** \brief bead mass */
	PetscScalar mass_ = 1.;

	/** \brief bead inertia */
	PetscScalar inertia_ = 1.;

	 /** \brief bead position */
	EigenWrapper::EigenVec3 position_ = EigenWrapper::EigenVec3::Zero();

	/** \brief bead quaternion */
	EigenWrapper::EigenQuat quaternion_ = EigenWrapper::EigenQuat::Identity();

	/** \brief bead translational velocity */
	EigenWrapper::EigenVec3 vel_ = EigenWrapper::EigenVec3::Zero();

	/** \brief bead rotational velocity */
	EigenWrapper::EigenVec3 omg_  = EigenWrapper::EigenVec3::Zero();

    /** \brief bead force */
	EigenWrapper::EigenVec3 force_ = EigenWrapper::EigenVec3::Zero();

	/** \brief bead torque */
	EigenWrapper::EigenVec3 torque_ = EigenWrapper::EigenVec3::Zero();

	/** \brief bead force used in Adams–Bashforth scheme */
	EigenWrapper::EigenVec3 forceOld_ = EigenWrapper::EigenVec3::Zero();

	/** \brief bead torque used in Adams–Bashforth scheme */
	EigenWrapper::EigenVec3 torqueOld_ = EigenWrapper::EigenVec3::Zero();

public:
	Bead();
	Bead(const Bead& rhs);
	~Bead();
	Bead& operator=(const Bead& rhs);
	PetscErrorCode setProperties(PetscScalar density,PetscScalar radius);
	PetscScalar &density(){return density_;}
	PetscScalar &radius(){return radius_;}
	PetscScalar &mass(){return mass_;}
	PetscScalar &inertia(){return inertia_;}
	EigenWrapper::EigenVec3 &position(){return position_;}
	EigenWrapper::EigenQuat &quaternion(){return quaternion_;}
	EigenWrapper::EigenVec3 &vel(){return vel_;}
	EigenWrapper::EigenVec3 &omg(){return omg_;}
	EigenWrapper::EigenVec3 &force(){return force_;}
	EigenWrapper::EigenVec3 &torque(){return torque_;}
	EigenWrapper::EigenVec3 &forceOld() {return forceOld_;}
	EigenWrapper::EigenVec3 &torqueOld() {return torqueOld_;}

	const PetscScalar &density() const {return density_;}
	const PetscScalar &radius() const {return radius_;}
	const PetscScalar &mass() const {return mass_;}
	const PetscScalar &inertia() const {return inertia_;}
	const EigenWrapper::EigenVec3 &position() const {return position_;}
	const EigenWrapper::EigenQuat &quaternion() const {return quaternion_;}
	const EigenWrapper::EigenVec3 &vel() const {return vel_;}
	const EigenWrapper::EigenVec3 &omg() const {return omg_;}
	const EigenWrapper::EigenVec3 &force() const {return force_;}
	const EigenWrapper::EigenVec3 &torque() const {return torque_;}
	const EigenWrapper::EigenVec3 &forceOld() const {return forceOld_;}
	const EigenWrapper::EigenVec3 &torqueOld() const {return torqueOld_;}
};

/** \class Hinge
*   \brief Chain of beads connected through a virtual beam */
class Hinge
{
private:

	 /** \brief Name */
	std::string name_;

	/** \brief Fiber density */
	PetscScalar density_;

	/** \brief Fiber radius */
	PetscScalar radius_;

	 /** \brief Fiber length */
	PetscScalar length_;

	/** \brief Fiber stiffness \f$ ks = \pi r_b E / 2 \f$ */
	PetscScalar ks_ = 1.;

	/** \brief Damping factor */
	PetscScalar damping_ = 0.001;

	 /** \brief Array of beads */
	std::vector<Bead> beads_;

	/** \brief Bead number */
	PetscInt size_;

public:

	/** \brief Empty Constructor */
	Hinge();

	/** \brief Constructor
	 *  \param[in] name hinge name */
	Hinge(const std::string& name);

	/** \brief Constructor
	 *  \param[in] density material density
	 *  \param[in] A       first bead's centre of mass
	 *  \param[in] B       last bead's centre of mas
	 *  \param[in] n       bead number
	 *  \param[in] name    hinge name */
	Hinge(PetscScalar density,const EigenWrapper::EigenVec3& A,const EigenWrapper::EigenVec3& B,PetscInt n,const std::string& name);

	/** \brief Constructor
	 *  \param[in] A       first bead's centre of mass
	 *  \param[in] B       last bead's centre of mas
	 *  \param[in] n       bead number
	 *  \param[in] name    hinge name */
	Hinge(const EigenWrapper::EigenVec3& A,const EigenWrapper::EigenVec3& B,PetscInt n,const std::string& name);

	/** \brief Copy constructor */
	Hinge(const Hinge& rhs);

	/** \brief Assignment operator */
	Hinge& operator=(const Hinge& rhs);

	/** \brief Destructor */
	~Hinge();

	/** \brief Set beads */
	PetscErrorCode set(PetscScalar density,const EigenWrapper::EigenVec3& A,const EigenWrapper::EigenVec3& B,PetscInt n);

	PetscErrorCode setProperties(PetscScalar density,PetscScalar radius);

	/** \brief Write hinge
	 *  \param[in] t time */
	PetscErrorCode writeInASCIIFormat
	(
		PetscInt t
	) const;

	/** \brief Write hinge
	 *  \param[in] t time */
	PetscErrorCode writeInBinaryFormat
	(
		PetscInt t
	) const;

	/** \brief Read hinge
	 *  \param[in] t time */
	PetscErrorCode readInBinaryFormat
	(
		PetscInt t = -1
	);

	/** \brief Calculate beam torques and forces */
	PetscErrorCode updateBeadForces();

	/** \brief Update bead velocities
	 *  \param[in] dt time step */
	PetscErrorCode updateBeadVelocities(PetscScalar dt);

	/** \brief Update bead quaternions and positions
	 *  \param[in] dt time step */
	PetscErrorCode updateBeadPositions(PetscScalar dt);

	/** \brief Reference structural force given by |F_0| = EI/L^2 */
	PetscScalar referenceForce() const;

	/** \brief Normalize an external load with hinge properties
	 *  \param[in] P external load */
	PetscScalar normalizedLoad(PetscScalar P) const;

	/** \brief Normalize an external load with hinge properties
	 *  \param[in] P external load */
	EigenWrapper::EigenVec3 normalizedLoad(const EigenWrapper::EigenVec3& P) const;

	/** \brief Check that beads do not skyrocket into infinity */
	PetscErrorCode check() const;

	/** \brief Update all beads given a load acting on the last bead. The first bead is fixed.
	 *  \param[in] Fext external force acting on last bead
	 *  \param[in] Text external torque acting in last bead
	 *  \param[in] dt   time step */
	PetscErrorCode update(const EigenWrapper::EigenVec3 &Fext,const EigenWrapper::EigenVec3 &Text,PetscScalar dt);

	std::string &name(){return name_;}
	PetscInt &size(){return size_;}
	PetscScalar &length(){return length_;}
	PetscScalar &ks(){return ks_;}
	PetscScalar &damping(){return damping_;}
	PetscScalar &radius(){return radius_;}
	std::vector<Bead> &beads(){return beads_;}
	Bead &bead(PetscInt b){return beads_[b];}
	Bead &firstBead(){return beads_.front();}
	Bead &lastBead(){return beads_.back();}

	const std::string &name() const {return name_;};
	const PetscInt &size() const {return size_;}
	const PetscScalar &length() const {return length_;};
	const PetscScalar &ks() const {return ks_;}
	const PetscScalar &damping() const {return damping_;}
	const PetscScalar &radius() const {return radius_;}
	const std::vector<Bead> &beads() const {return beads_;}
	const Bead &bead(PetscInt b) const {return beads_[b];}
	const Bead &firstBead() const {return beads_.front();}
	const Bead &lastBead() const {return beads_.back();}

};

PetscErrorCode TestHinge(PetscScalar load=8.);

#endif
