#include "IglWrapper.h"
#include "igl/writeOFF.h"
#include "igl/writeSTL.h"
#include "igl/writeOBJ.h"
#include "igl/combine.h"
#include "igl/copyleft/cgal/remesh_self_intersections.h"
#include "igl/remove_unreferenced.h"
#include "igl/unique_edge_map.h"
#include "igl/extract_non_manifold_edge_curves.h"
#include "igl/is_boundary_edge.h"
#include "igl/marching_cubes.h"
#include "igl/signed_distance.h"


int IglWrapper::WriteOFF
(
	const std::string &name,
	const EigenWrapper::EigenMat &vertices,
	const EigenWrapper::EigenMatI &faces
)
{
	igl::writeOFF(name,vertices,faces);
	return(0);
}


int IglWrapper::WriteSTL
(
	const std::string &name,
	const EigenWrapper::EigenMat &vertices,
	const EigenWrapper::EigenMatI &faces
)
{
	igl::writeSTL(name,vertices,faces);
	return(0);
}


int IglWrapper::WriteOBJ
(
	const std::string &name,
	const EigenWrapper::EigenMat &vertices,
	const EigenWrapper::EigenMatI &faces
)
{
	igl::writeOBJ(name,vertices,faces);
	return(0);
}


int IglWrapper::FlattenCoordinates
(
	DM da,
	EigenWrapper::EigenMat &mat
)
{
	PetscWrapper::GridCoordinates3D coords;
	coords.create(da);
	std::vector<PetscScalar> a; PetscInt rows,cols;
	coords.flattenArrayTo(a,rows,cols);
	mat = Eigen::Map<const EigenWrapper::EigenMat>(a.data(),rows,cols);
	coords.destroy();
	return(0);
}


int IglWrapper::FlattenMarchingCubeCoordinates
(
	DM da,
	EigenWrapper::EigenMat &mat
)
{
	PetscWrapper::GridCoordinates3D coords;
	coords.create(da);
	std::vector<PetscScalar> a; PetscInt rows,cols;
	coords.flattenMarchingCubeArrayTo(a,rows,cols);
	mat = Eigen::Map<const EigenWrapper::EigenMat>(a.data(),rows,cols);
	coords.destroy();
	return(0);
}


int IglWrapper::FlattenedSignedDistance
(
	const EigenWrapper::EigenMat &vertices,
	const EigenWrapper::EigenMatI &faces,
	const EigenWrapper::EigenMat &flattenedGrid,
	EigenWrapper::EigenVec &signedDistance
)
{
	igl::SignedDistanceType signedDistanceType = igl::SignedDistanceType::SIGNED_DISTANCE_TYPE_PSEUDONORMAL;
	//EigenWrapper::EigenVec signedDistance; //list of signed distances
	EigenWrapper::EigenVecI faceIndice;    //list of face indices with smallest signed distances
	EigenWrapper::EigenMat closestPoint;   //list of closest points
	EigenWrapper::EigenMat closestNormal;  //list of closest normal vectors
	igl::signed_distance
	(
		flattenedGrid,
		vertices,
		faces,
		signedDistanceType,
		-10000,
		+10000,
		signedDistance,
		faceIndice,
		closestPoint,
		closestNormal
	);
	return(0);
}


int IglWrapper::MarchingCubes
(
	const EigenWrapper::EigenMat &flattenedGrid,
	ScalarField &field,
	const PetscScalar &iso,
	EigenWrapper::EigenMat &vertices,
	EigenWrapper::EigenMatI &faces
)
{
	std::vector<PetscScalar> flattenedField;
	PetscInt nx,ny,nz;
	field.flattenMarchingCubeArrayTo(flattenedField,nx,ny,nz);
	igl::marching_cubes
	(
		Eigen::Map<const EigenWrapper::EigenVec>
		(
			flattenedField.data(),
			flattenedField.size()
		),
		flattenedGrid,
		nx,
		ny,
		nz,
		iso,
		vertices,
		faces
	);
	return(0);
}


int IglWrapper::IntersectMeshes
(
	const EigenWrapper::EigenMat  &Va,
	const EigenWrapper::EigenMatI &Fa,
	const EigenWrapper::EigenMat  &Vb,
	const EigenWrapper::EigenMatI &Fb,
	std::vector<EigenWrapper::EigenMat> &segments
)
{
	if ( (Va.size() > 0) && (Vb.size() > 0) )
	{
		//Combine meshes
		std::vector<EigenWrapper::EigenMat> listVertices =
		{
			Va,
			Vb
		};
		std::vector<EigenWrapper::EigenMatI> listFaces =
		{
			Fa,
			Fb
		};
		EigenWrapper::EigenMat combinedVertices;
		EigenWrapper::EigenMatI combinedFaces;
		igl::combine(listVertices,listFaces,combinedVertices,combinedFaces);

		//Resolve intersections (copy/paste from mesh_boolean.cpp)
		EigenWrapper::EigenMat intersectedVertices;
		EigenWrapper::EigenMatI intersectedFaces,IF;
		EigenWrapper::EigenVecI J,IM;
		igl::copyleft::cgal::RemeshSelfIntersectionsParam params;
		params.stitch_all = false;
		igl::copyleft::cgal::remesh_self_intersections
		(
			combinedVertices,
			combinedFaces,
			params,
			intersectedVertices,
			intersectedFaces,
			IF,
			J,
			IM
		);

		//Check if there is an intersection
		if ( IF.size()>0 )
		{
			//Merge coinciding vertices into non-manifold vertices
			std::for_each(intersectedFaces.data(),intersectedFaces.data()+intersectedFaces.size(),[&IM](int &a){a=IM(a);});

			//Remove unreferenced vertices
			EigenWrapper::EigenMat unreferencedVertices;
			EigenWrapper::EigenMatI unreferencedFaces;
			EigenWrapper::EigenVecI UIM;
			igl::remove_unreferenced(intersectedVertices,intersectedFaces,unreferencedVertices,unreferencedFaces,UIM);

			//Compute edges of (F) --> (E,uE,EMAP,uE2E)
			EigenWrapper::EigenMatI E,uE;
			EigenWrapper::EigenVecI EMAP;
			std::vector<std::vector<size_t> > uE2E;
			igl::unique_edge_map(unreferencedFaces,E,uE,EMAP,uE2E);

			//Non manifold edges
			std::vector<std::vector<size_t> > curves;
			igl::extract_non_manifold_edge_curves(unreferencedFaces,EMAP,uE2E,curves);

			//Edge boundaries
			EigenWrapper::EigenVecI boundaryEdge;
			igl::is_boundary_edge(uE,unreferencedFaces,boundaryEdge);

			//Polylines
			for (PetscInt cj=0; cj<curves.size(); ++cj)
			{
				EigenWrapper::EigenMat segment;
				for (PetscInt ci=0; ci<curves[cj].size(); ++ci)
				{
					PetscInt index = curves[cj][ci];
					if ( !boundaryEdge(index) )
					{
						PetscInt A = uE.row(index)(0);
						PetscInt B = uE.row(index)(1);
						segment = EigenWrapper::EigenMat(2,3);
						segment.row(0) << unreferencedVertices.row(A);
						segment.row(1) << unreferencedVertices.row(B);
					}
					else
					{
						break;
					}
				}

				//Add segment to list
				if ( segment.rows() > 0 )
				{
					segments.push_back(segment);
				}
			}
		}
	}

	return(0);
}
