#include "Interpolation3D.h"


GetDirectionIndices();


namespace
{
	const PetscScalar EDGE_INTERP_01_BY_02_ = 1./2.; /**< \brief Low-order edge interpolation coefficient = 1/2 */
	const PetscScalar EDGE_INTERP_09_BY_16_ = 9./16.; /**< \brief High-order edge interpolation coefficient = 9/16 */
	const PetscScalar EDGE_INTERP_01_BY_16_ = 1./16.; /**< \brief High-order edge interpolation coefficient = 1/16 */
	const PetscScalar FACE_INTERP_01_BY_04_ = 1./4.; /**< \brief Low-order face interpolation coefficient = 9/32 */
	const PetscScalar FACE_INTERP_09_BY_32_ = 9./32.; /**< \brief High-order face interpolation coefficient = 9/32 */
	const PetscScalar FACE_INTERP_01_BY_32_ = 1./32.; /**< \brief High-order face interpolation coefficient = 1/32 */
}


PetscScalar Interpolation3D::LowOrder::InterpolateEdgeX
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		EDGE_INTERP_01_BY_02_ * ( f[k  ][j  ][i-1] + f[k  ][j  ][i  ] )
	);
}


PetscScalar Interpolation3D::HighOrder::InterpolateEdgeX
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  EDGE_INTERP_09_BY_16_ * ( f[k  ][j  ][i-1] + f[k  ][j  ][i  ] )
		- EDGE_INTERP_01_BY_16_ * ( f[k  ][j  ][i-2] + f[k  ][j  ][i+1] )
	);
}



PetscScalar Interpolation3D::LowOrder::InterpolateEdgeY
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		EDGE_INTERP_01_BY_02_ * ( f[k  ][j-1][i  ] + f[k  ][j  ][i  ] )
	);
}


PetscScalar Interpolation3D::HighOrder::InterpolateEdgeY
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  EDGE_INTERP_09_BY_16_ * ( f[k  ][j-1][i  ] + f[k  ][j  ][i  ] )
		- EDGE_INTERP_01_BY_16_ * ( f[k  ][j-2][i  ] + f[k  ][j+1][i  ] )
	);
}


PetscScalar Interpolation3D::LowOrder::InterpolateEdgeZ
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		EDGE_INTERP_01_BY_02_ * ( f[k-1][j  ][i  ] + f[k  ][j  ][i  ] )
	);
}


PetscScalar Interpolation3D::HighOrder::InterpolateEdgeZ
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  EDGE_INTERP_09_BY_16_ * ( f[k-1][j  ][i  ] + f[k  ][j  ][i  ] )
		- EDGE_INTERP_01_BY_16_ * ( f[k-2][j  ][i  ] + f[k+1][j  ][i  ] )
	);
}


PetscScalar Interpolation3D::LowOrder::InterpolateEdge
(
	const PetscScalar *const *const *const f,
	PetscInt dim,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		Interpolation3D::LowOrder::InterpolateEdgePtr[dim](f,k,j,i)
	);
}


PetscScalar Interpolation3D::HighOrder::InterpolateEdge
(
	const PetscScalar *const *const *const f,
	PetscInt dim,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		Interpolation3D::HighOrder::InterpolateEdgePtr[dim](f,k,j,i)
	);
}


PetscScalar Interpolation3D::LowOrder::InterpolateFaceX
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		FACE_INTERP_01_BY_04_*
		(
			  f[k-1][j-1][i  ]
			+ f[k-1][j  ][i  ]
			+ f[k  ][j-1][i  ]
			+ f[k  ][j  ][i  ]
		)
	);
}


PetscScalar Interpolation3D::HighOrder::InterpolateFaceX
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		FACE_INTERP_09_BY_32_*
		(
			  f[k-1][j-1][i  ]
			+ f[k-1][j  ][i  ]
			+ f[k  ][j-1][i  ]
			+ f[k  ][j  ][i  ]
		)

		- FACE_INTERP_01_BY_32_*
		(
			  f[k-2][j-2][i  ]
			+ f[k-2][j+1][i  ]
			+ f[k+1][j-2][i  ]
			+ f[k+1][j+1][i  ]
		)
	);
}


PetscScalar Interpolation3D::LowOrder::InterpolateFaceY
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		FACE_INTERP_01_BY_04_*
		(
			  f[k-1][j  ][i-1]
			+ f[k-1][j  ][i  ]
			+ f[k  ][j  ][i-1]
			+ f[k  ][j  ][i  ]
		)
	);
}


PetscScalar Interpolation3D::HighOrder::InterpolateFaceY
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		FACE_INTERP_09_BY_32_*
		(
			  f[k-1][j  ][i-1]
			+ f[k-1][j  ][i  ]
			+ f[k  ][j  ][i-1]
			+ f[k  ][j  ][i  ]
		)

		- FACE_INTERP_01_BY_32_*
		(
			  f[k-2][j  ][i-2]
			+ f[k-2][j  ][i+1]
			+ f[k+1][j  ][i-2]
			+ f[k+1][j  ][i+1]
		)
	);
}


PetscScalar Interpolation3D::LowOrder::InterpolateFaceZ
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		FACE_INTERP_01_BY_04_*
		(
			  f[k  ][j-1][i-1]
			+ f[k  ][j-1][i  ]
			+ f[k  ][j  ][i-1]
			+ f[k  ][j  ][i  ]
		)
	);
}


PetscScalar Interpolation3D::HighOrder::InterpolateFaceZ
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		FACE_INTERP_09_BY_32_*
		(
			  f[k  ][j-1][i-1]
			+ f[k  ][j-1][i  ]
			+ f[k  ][j  ][i-1]
			+ f[k  ][j  ][i  ]
		)

		- FACE_INTERP_01_BY_32_*
		(
			  f[k  ][j-2][i-2]
			+ f[k  ][j-2][i+1]
			+ f[k  ][j+1][i-2]
			+ f[k  ][j+1][i+1]
		)
	);
}


PetscScalar Interpolation3D::LowOrder::InterpolateFace
(
	const PetscScalar *const *const *const f,
	PetscInt dim,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return Interpolation3D::LowOrder::InterpolateFacePtr[dim](f,k,j,i);
}


PetscScalar Interpolation3D::HighOrder::InterpolateFace
(
	const PetscScalar *const *const *const f,
	PetscInt dim,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return Interpolation3D::HighOrder::InterpolateFacePtr[dim](f,k,j,i);
}


PetscErrorCode Interpolation3D::LowOrder::InterpolateViscositiesX
(
	const PetscScalar *const *const *const eta,
	PetscInt k,
	PetscInt j,
	PetscInt i,
	PetscScalar etaInterp[7]
)
{
	etaInterp[F_] = Interpolation3D::LowOrder::InterpolateFaceY(eta,k+1,j  ,i  );
	etaInterp[B_] = Interpolation3D::LowOrder::InterpolateFaceY(eta,k  ,j  ,i  );
	etaInterp[N_] = Interpolation3D::LowOrder::InterpolateFaceZ(eta,k  ,j+1,i  );
	etaInterp[S_] = Interpolation3D::LowOrder::InterpolateFaceZ(eta,k  ,j  ,i  );
	etaInterp[E_] = eta[k  ][j  ][i  ];
	etaInterp[W_] = eta[k  ][j  ][i-1];
	etaInterp[C_] = etaInterp[F_] + etaInterp[B_] + etaInterp[N_] + etaInterp[S_] + etaInterp[E_] + etaInterp[W_];
	return (0);
}


PetscErrorCode Interpolation3D::LowOrder::InterpolateViscositiesY
(
	const PetscScalar *const *const *const eta,
	PetscInt k,
	PetscInt j,
	PetscInt i,
	PetscScalar etaInterp[7]
)
{
	etaInterp[F_] = Interpolation3D::LowOrder::InterpolateFaceX(eta,k+1,j  ,i  );
	etaInterp[B_] = Interpolation3D::LowOrder::InterpolateFaceX(eta,k  ,j  ,i  );
	etaInterp[N_] = eta[k  ][j  ][i  ];
	etaInterp[S_] = eta[k  ][j-1][i  ];
	etaInterp[E_] = Interpolation3D::LowOrder::InterpolateFaceZ(eta,k  ,j  ,i+1);
	etaInterp[W_] = Interpolation3D::LowOrder::InterpolateFaceZ(eta,k  ,j  ,i  );
	etaInterp[C_] = etaInterp[F_] + etaInterp[B_] + etaInterp[N_] + etaInterp[S_] + etaInterp[E_] + etaInterp[W_];
	return (0);
}

PetscErrorCode Interpolation3D::LowOrder::InterpolateViscositiesZ
(
	const PetscScalar *const *const *const eta,
	PetscInt k,
	PetscInt j,
	PetscInt i,
	PetscScalar etaInterp[7]
)
{
	etaInterp[F_] = eta[k  ][j  ][i  ];
	etaInterp[B_] = eta[k-1][j  ][i  ];
	etaInterp[N_] = Interpolation3D::LowOrder::InterpolateFaceX(eta,k  ,j+1,i  );
	etaInterp[S_] = Interpolation3D::LowOrder::InterpolateFaceX(eta,k  ,j  ,i  );
	etaInterp[E_] = Interpolation3D::LowOrder::InterpolateFaceY(eta,k  ,j  ,i+1);
	etaInterp[W_] = Interpolation3D::LowOrder::InterpolateFaceY(eta,k  ,j  ,i  );
	etaInterp[C_] = etaInterp[F_] + etaInterp[B_] + etaInterp[N_] + etaInterp[S_] + etaInterp[E_] + etaInterp[W_];
	return (0);
}


PetscErrorCode Interpolation3D::LowOrder::InterpolateViscosities
(
	const PetscScalar *const *const *const eta,
	PetscInt dim,
	PetscInt k,
	PetscInt j,
	PetscInt i,
	PetscScalar etaInterp[7]
)
{
	return Interpolation3D::LowOrder::InterpolateViscositiesPtr[dim](eta,k,j,i,etaInterp);
}
