#ifndef INTERPOLATION3D_H
#define INTERPOLATION3D_H


/** \file
 *  \brief Collection of interpolation schemes */


#include "petsc.h"
#include "Utils.h"


namespace
{


	typedef PetscScalar(*ThreeDimensionalEdgeInterpolator)(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt);
	typedef PetscScalar(*ThreeDimensionalFaceInterpolator)(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt);
	typedef PetscErrorCode(*ThreeDimensionalViscosityInterpolator)(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt,PetscScalar*);


}


/** \namespace Interpolation3D
  * \brief Collection of interpolation schemes */
namespace Interpolation3D
{


	/** \namespace Interpolation3D::LowOrder
	 *  \brief Collection of low-order interpolation schemes */
	namespace LowOrder
	{


		/** \brief 2-point stencil edge interpolation along x-axis
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      interpolated scalar \f$ f|_{k,j,i-1/2} \f$ */
		PetscScalar InterpolateEdgeX
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 2-point stencil edge interpolation along y-axis
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      interpolated scalar \f$ f|_{k,j-1/2,i} \f$ */
		PetscScalar InterpolateEdgeY
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 2-point stencil edge interpolation along z-axis
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      interpolated scalar \f$ f|_{k-1/2,j,i} \f$ */
		PetscScalar InterpolateEdgeZ
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 2-point stencil edge interpolation along axis
		 *  \param[in] f   three-dimensional array
		 *  \param[in] dim space dimension
		 *  \param[in] k   node index
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        interpolated scalar \f$ f|_{k    ,j    ,i-1/2} \f$ if dim = 0 else
		 *                                     \f$ f|_{k    ,j-1/2,i    } \f$ if dim = 1 else
		 *                                     \f$ f|_{k-1/2,j    ,i    } \f$ if dim = 2 */
		PetscScalar InterpolateEdge
		(
			const PetscScalar *const *const *const f,
			PetscInt dim,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to 2-point-stencil edge interpolator */
		const ThreeDimensionalEdgeInterpolator InterpolateEdgePtr[] =
		{
			Interpolation3D::LowOrder::InterpolateEdgeX,
			Interpolation3D::LowOrder::InterpolateEdgeY,
			Interpolation3D::LowOrder::InterpolateEdgeZ
		};


		/** \brief 4-point stencil face interpolation in x-plane
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      interpolated scalar \f$ f_{k-1/2,j-1/2,i} \f$ */
		PetscScalar InterpolateFaceX
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil face interpolation in y-plane
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      interpolated scalar \f$ f_{k-1/2,j,i-1/2} \f$ */
		PetscScalar InterpolateFaceY
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil face interpolation in z-plane
		 *  \param[in] f three-dimensional array
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return      interpolated scalar \f$ f_{k,j-1/2,i-1/2} \f$ */
		PetscScalar InterpolateFaceZ
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to 4-point-stencil face interpolator */
		const ThreeDimensionalFaceInterpolator InterpolateFacePtr[] =
		{
			Interpolation3D::LowOrder::InterpolateFaceX,
			Interpolation3D::LowOrder::InterpolateFaceY,
			Interpolation3D::LowOrder::InterpolateFaceZ
		};


		/** \brief 4-point stencil face interpolation
		 *  \param[in] f   three-dimensional array
		 *  \param[in] dim space dimension
		 *  \param[in] k   node index
		 *  \param[in] j   node index
		 *  \param[in] i   node index
		 *  \return        interpolated scalar \f$ f_{k-1/2,j-1/2,    i} \f$ if dim = 0 else
		 *                                     \f$ f_{k-1/2,j    ,i-1/2} \f$ if dim = 1 else
		 *                                     \f$ f_{k    ,j-1/2,i-1/2} \f$ if dim = 2 else */
		PetscScalar InterpolateFace
		(
			const PetscScalar *const *const *const f,
			PetscInt dim,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief One-dimensional array of interpolated viscosities needed in 7-point stencil Laplacian(eta,ux)
		 *  \note \f$ \eta_C = -(\eta_F + \eta_B + \eta_N + \eta_S + \eta_E + \eta_W) \f$
		 *  \note \f$ \eta_F = \eta_{ k+1/2, j    , I     } = \eta_{ k+1/2, j    , i-1/2 } \f$
		 *  \note \f$ \eta_B = \eta_{ k-1/2, j    , I     } = \eta_{ k-1/2, j    , i-1/2 } \f$
		 *  \note \f$ \eta_N = \eta_{ k    , j+1/2, I     } = \eta_{ k    , j+1/2, i-1/2 } \f$
		 *  \note \f$ \eta_S = \eta_{ k    , j-1/2, I     } = \eta_{ k    , j-1/2, i-1/2 } \f$
		 *  \note \f$ \eta_E = \eta_{ k    , j    , I+1/2 } = \eta_{ k    , j    , i     } \f$
		 *  \note \f$ \eta_W = \eta_{ k    , j    , I-1/2 } = \eta_{ k    , j    , i-1   } \f$
		 *  \param[in]  eta       three-dimensional array
		 *  \param[in]  k         node index
		 *  \param[in]  j         node index
		 *  \param[in]  i         node index
		 *  \param[out] etaInterp viscosity array needed in \f$ \left. \nabla \cdot \left( \eta \nabla u_x \right ) \right|_{k,j,i-1/2} \f$ */
		PetscErrorCode InterpolateViscositiesX
		(
			const PetscScalar *const *const *const eta,
			PetscInt k,
			PetscInt j,
			PetscInt i,
			PetscScalar etaInterp[7]
		);


		/** \brief One-dimensional array of interpolated viscosities needed in 7-point stencil Laplacian(eta,uy)
		 *  \note \f$ \eta_C = -(\eta_F + \eta_B + \eta_N + \eta_S + \eta_E + \eta_W) \f$
		 *  \note \f$ \eta_F = \eta_{ k+1/2, J    , i     } = \eta_{ k+1/2, j-1/2, i     } \f$
		 *  \note \f$ \eta_B = \eta_{ k-1/2, J    , i     } = \eta_{ k-1/2, j-1/2, i     } \f$
		 *  \note \f$ \eta_N = \eta_{ k    , J+1/2, i     } = \eta_{ k    , j    , i     } \f$
		 *  \note \f$ \eta_S = \eta_{ k    , J-1/2, i     } = \eta_{ k    , j-1  , i     } \f$
		 *  \note \f$ \eta_E = \eta_{ k    , J    , i+1/2 } = \eta_{ k    , j-1/2, i+1/2 } \f$
		 *  \note \f$ \eta_W = \eta_{ k    , J    , i-1/2 } = \eta_{ k    , j-1/2, i-1/2 } \f$
		 *  \param[in]  eta       three-dimensional array
		 *  \param[in]  k         node index
		 *  \param[in]  j         node index
		 *  \param[in]  i         node index
		 *  \param[out] etaInterp viscosity array needed in \f$ \left. \nabla \cdot \left( \eta \nabla u_y \right ) \right|_{k,j-1/2,i} \f$ */
		PetscErrorCode InterpolateViscositiesY
		(
			const PetscScalar *const *const *const eta,
			PetscInt k,
			PetscInt j,
			PetscInt i,
			PetscScalar etaInterp[7]
		);


		/** \brief One-dimensional array of interpolated viscosities needed in 7-point stencil Laplacian(eta,uz)
		 *  \note \f$ \eta_C = -(\eta_F + \eta_B + \eta_N + \eta_S + \eta_E + \eta_W) \f$
		 *  \note \f$ \eta_F = \eta_{ K+1/2, j    , i     } = \eta_{ k    , j    , i     } \f$
		 *  \note \f$ \eta_B = \eta_{ K-1/2, j    , i     } = \eta_{ k-1  , j    , i     } \f$
		 *  \note \f$ \eta_N = \eta_{ K    , j+1/2, i     } = \eta_{ k-1/2, j+1/2, i     } \f$
		 *  \note \f$ \eta_S = \eta_{ K    , j-1/2, i     } = \eta_{ k-1/2, j-1/2, i     } \f$
		 *  \note \f$ \eta_E = \eta_{ K    , j    , i+1/2 } = \eta_{ k-1/2, j    , i+1/2 } \f$
		 *  \note \f$ \eta_W = \eta_{ K    , j    , i-1/2 } = \eta_{ k-1/2, j    , i-1/2 } \f$
		 *  \param[in]  eta       three-dimensional array
		 *  \param[in]  k         node index
		 *  \param[in]  j         node index
		 *  \param[in]  i         node index
		 *  \param[out] etaInterp viscosity array needed in \f$ \left. \nabla \cdot \left( \eta \nabla u_z \right ) \right|_{k-1/2,j,i} \f$ */
		PetscErrorCode InterpolateViscositiesZ
		(
			const PetscScalar *const *const *const eta,
			PetscInt k,
			PetscInt j,
			PetscInt i,
			PetscScalar etaInterp[7]
		);


		/** \brief Pointer to 7-point-stencil viscosity interpolator needed in Laplacian(eta,u[dim]) */
		const ThreeDimensionalViscosityInterpolator InterpolateViscositiesPtr[] =
		{
			Interpolation3D::LowOrder::InterpolateViscositiesX,
			Interpolation3D::LowOrder::InterpolateViscositiesY,
			Interpolation3D::LowOrder::InterpolateViscositiesZ
		};


		/** brief One-dimensional array of interpolated viscosities needed in 7-point stencil Laplacian(eta,u[dim])
		 *  \param[in]  eta       three-dimensional array
		 *  \param[in]  dim       space dimension
		 *  \param[in]  k         node index
		 *  \param[in]  j         node index
		 *  \param[in]  i         node index
		 *  \param[out] etaInterp viscosity array needed in \f$ \nabla \cdot \left( \eta \nabla f_{\alpha=0,1,2} \right ) \f$
		 *                        @ \f$ |_{k    ,j    ,i-1/2} \f$ if dim = 0 else
		 *                        @ \f$ |_{k    ,j-1/2,    i} \f$ if dim = 1 else
		 *                        @ \f$ |_{k-1/2,j    ,i    } \f$ if dim = 2  */
		PetscErrorCode InterpolateViscosities
		(
			const PetscScalar *const *const *const eta,
			PetscInt dim,
			PetscInt k,
			PetscInt j,
			PetscInt i,
			PetscScalar etaInterp[7]
		);


	}


	/** \namespace Interpolation3D::HighOrder
	 *  \brief Collection of high-order interpolation schemes */
	namespace HighOrder
	{


		/** \brief 4-point stencil edge interpolation along x-axis
		    \copydetails Interpolation3D::LowOrder::InterpolateEdgeX(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar InterpolateEdgeX
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil edge interpolation along y-axis
		    \copydetails Interpolation3D::LowOrder::InterpolateEdgeY(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar InterpolateEdgeY
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil edge interpolation along z-axis
		    \copydetails Interpolation3D::LowOrder::InterpolateEdgeZ(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar InterpolateEdgeZ
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil edge interpolation along axis
		    \copydetails Interpolation3D::LowOrder::InterpolateEdge(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt,PetscInt) */
		PetscScalar InterpolateEdge
		(
			const PetscScalar *const *const *const f,
			PetscInt dim,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to 4-point-stencil edge interpolator */
		const ThreeDimensionalEdgeInterpolator InterpolateEdgePtr[] =
		{
			Interpolation3D::LowOrder::InterpolateEdgeX,
			Interpolation3D::LowOrder::InterpolateEdgeY,
			Interpolation3D::LowOrder::InterpolateEdgeZ
		};


		/** \brief 8-point stencil face interpolation in x-plane
		 *  \copydetails Interpolation3D::LowOrder::InterpolateFaceX(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar InterpolateFaceX
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 4-point stencil face interpolation in y-plane
		 *  \copydetails Interpolation3D::LowOrder::InterpolateFaceY(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar InterpolateFaceY
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief 8-point stencil face interpolation in z-plane
		 *  \copydetails Interpolation3D::LowOrder::InterpolateFaceX(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt) */
		PetscScalar InterpolateFaceZ
		(
			const PetscScalar *const *const *const f,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);


		/** \brief Pointer to 8-point-stencil face interpolator */
		const ThreeDimensionalFaceInterpolator InterpolateFacePtr[] =
		{
			Interpolation3D::HighOrder::InterpolateFaceX,
			Interpolation3D::HighOrder::InterpolateFaceY,
			Interpolation3D::HighOrder::InterpolateFaceZ
		};


		/** \brief 8-point stencil face interpolation
		*  \copydetails Interpolation3D::LowOrder::InterpolateFace(const PetscScalar *const *const *const,PetscInt,PetscInt,PetscInt,PetscInt) */
		PetscScalar InterpolateFace
		(
			const PetscScalar *const *const *const f,
			PetscInt dim,
			PetscInt k,
			PetscInt j,
			PetscInt i
		);



	}


}


#endif
