#include "PetscWrapper.h"


namespace
{


	/** \brief Number of time digits in file name set using ".*" */
	const PetscInt TIMEWIDTH_ = 5;


	/** \brief Check if file exists
	 *  \param[in] file file name */
	PetscErrorCode CheckIfFileExist
	(
		const std::string &file
	)
	{
		PetscBool fileExists;
		PetscTestFile(file.c_str(),'\0',&fileExists);
		if (!fileExists)
		{
			PetscPrintf(PETSC_COMM_WORLD,"  Error: %s not found\n",file.c_str());
			PetscEnd();
		}
		return(0);
	};


	/** \brief Create file name with time suffix and extension
	 *  \param[in] name name
	 *  \param[in] t    time
	 *  \param[in] ext  file extension
	 *  \return file name */
	std::string GetFileName
	(
		const std::string &name,
		PetscInt t,
		const std::string &ext
	)
	{
		char tmp[1000];
		if (t<0)
		{
			sprintf(tmp,"%s.%s",name.c_str(),ext.c_str());
		}
		else
		{
			sprintf(tmp,"%s_t%.*d.%s",name.c_str(),TIMEWIDTH_,t,ext.c_str());
		}
		return std::string(tmp);
	}


}


PetscErrorCode PetscWrapper::GridCoordinates2D::create
(
	DM da
)
{
	if (da)
	{
		PetscObjectReference((PetscObject)da);
		da_ = da;
	}
	else
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Error: dynamic array is undefined\n");
		PetscEnd();
	}
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates3D::create
(
	DM da
)
{
	if (da)
	{
		PetscObjectReference((PetscObject)da);
		da_ = da;
	}
	else
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Error: dynamic array is undefined\n");
		PetscEnd();
	}
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates2D::destroy()
{
// 	PetscObjectDereference((PetscObject)da);
	DMDestroy(&da_);
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates3D::destroy()
{
// 	PetscObjectDereference((PetscObject)da);
	DMDestroy(&da_);
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates2D::getArray()
{
	DMGetCoordinateDM(da_,&cda_);
	DMGetCoordinates(da_,&g_);
	DMDAVecGetArrayRead(cda_,g_,&a_);
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates3D::getArray()
{
	DMGetCoordinateDM(da_,&cda_);
	DMGetCoordinates(da_,&g_);
	DMDAVecGetArrayRead(cda_,g_,&a_);
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates2D::restoreArray()
{
	DMDAVecRestoreArrayRead(cda_,g_,&a_);
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates3D::restoreArray()
{
	DMDAVecRestoreArrayRead(cda_,g_,&a_);
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates2D::getGhostedArray()
{
	DMGetCoordinateDM(da_,&cda_);
	DMGetCoordinatesLocal(da_,&g_);
	DMDAVecGetArrayRead(cda_,g_,&a_);
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates3D::getGhostedArray()
{
	DMGetCoordinateDM(da_,&cda_);
	DMGetCoordinatesLocal(da_,&g_);
	DMDAVecGetArrayRead(cda_,g_,&a_);
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates2D::restoreGhostedArray()
{
	DMDAVecRestoreArrayRead(cda_,g_,&a_);
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates3D::restoreGhostedArray()
{
	DMDAVecRestoreArrayRead(cda_,g_,&a_);
	return(0);
}


PetscErrorCode PetscWrapper::GridCoordinates3D::copyGridPointTo
(
	PetscInt k,
	PetscInt j,
	PetscInt i,
	PetscScalar pt[3]
)
{
	pt[0] = a_[k][j][i].x; //this->x(k,j,i);
	pt[1] = a_[k][j][i].y; //this->y(k,j,i);
	pt[2] = a_[k][j][i].z; //this->z(k,j,i);
	return 0;
};


PetscErrorCode PetscWrapper::GridCoordinates3D::flattenArrayTo
(
	std::vector<PetscScalar> &v,
	PetscInt &rows,
	PetscInt &cols
)
{
	PetscInt i0,j0,k0,iw,jw,kw,n;
	DMDAGetCorners(da_,&i0,&j0,&k0,&iw,&jw,&kw);
	n = iw*jw*kw;
	v.resize(3*n);
	this->getArray();
	for (PetscInt k=k0; k<k0+kw; ++k)
	{
		for (PetscInt j=j0; j<j0+jw; ++j)
		{
			for (PetscInt i=i0; i<i0+iw; ++i)
			{
				PetscInt index = (i-i0) + iw * ( (j-j0) + jw * (k-k0) );
				v[index    ] = a_[k][j][i].x;
				v[index+  n] = a_[k][j][i].y;
				v[index+2*n] = a_[k][j][i].z;
			}
		}
	}
	this->restoreArray();
	rows = n;
	cols = 3;
	return(0);
}



PetscErrorCode PetscWrapper::GridCoordinates3D::flattenMarchingCubeArrayTo
(
	std::vector<PetscScalar> &v,
	PetscInt &rows,
	PetscInt &cols
)
{
	PetscInt i0,j0,k0,iw,jw,kw,n;
	DMDAGetCorners(da_,&i0,&j0,&k0,&iw,&jw,&kw);
	iw+=1; jw+=1; kw+=1; n=iw*jw*kw;
	v.resize(3*n);
	this->getGhostedArray();
	for (PetscInt k=k0; k<k0+kw; ++k)
	{
		for (PetscInt j=j0; j<j0+jw; ++j)
		{
			for (PetscInt i=i0; i<i0+iw; ++i)
			{
				PetscInt index = (i-i0) + iw * ( (j-j0) + jw * (k-k0) );
				v[index    ] = a_[k][j][i].x;
				v[index+  n] = a_[k][j][i].y;
				v[index+2*n] = a_[k][j][i].z;
			}
		}
	}
	this->restoreGhostedArray();
	rows = n;
	cols = 3;
	return(0);
}


PetscErrorCode PetscWrapper::CreateGlobalPetscVector
(
	DM da,
	Vec (&g),
	const std::string &name
)
{
	DMCreateGlobalVector(da,&g);
	VecSet(g,0.);
	PetscObjectSetName((PetscObject)g,name.c_str());
	PetscPrintf(PETSC_COMM_WORLD,"  Global Petsc vector %s created \n", name.c_str());
	return(0);
}


PetscErrorCode PetscWrapper::CreateGlobalPetscVector
(
	DM da,
	Vec* (&g),
	const std::string &name,
	PetscInt n
)
{
	Vec vtmp;
	DMCreateGlobalVector(da,&vtmp);
	VecDuplicateVecs(vtmp,n,&g);
	VecDestroy(&vtmp);
	for (PetscInt i=0;i<n;++i)
	{
		VecSet(g[i],0.);
		PetscObjectSetName((PetscObject)g[i],name.c_str());
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Array of %i global Petsc vectors %s created\n",n,name.c_str());
	return(0);
}


PetscErrorCode PetscWrapper::CreateLocalPetscVector
(
	DM da,
	Vec (&l)
)
{
	DMCreateLocalVector(da,&l);
	PetscPrintf(PETSC_COMM_WORLD,"  Local Petsc vector created\n");
	return(0);
}


PetscErrorCode PetscWrapper::CreateLocalPetscVector
(
	DM da,
	Vec* (&l),
	PetscInt n
)
{
	Vec v;
	DMCreateLocalVector(da,&v);
	VecDuplicateVecs(v,n,&l);
	VecDestroy(&v);
	PetscPrintf(PETSC_COMM_WORLD,"  Array of %i local Petsc vectors created\n",n);
	return(0);
}


PetscErrorCode PetscWrapper::DestroyPetscVector
(
	Vec (&v)
)
{
	VecDestroy(&v);
	PetscPrintf(PETSC_COMM_WORLD,"  Petsc vector destroyed\n");
	return(0);
}


PetscErrorCode PetscWrapper::DestroyPetscVector
(
	Vec* (&v),
	PetscInt n
)
{
	VecDestroyVecs(n,&v);
	PetscPrintf(PETSC_COMM_WORLD,"  Array of %i Petsc vectors destroyed\n",n);
	return(0);
}


PetscErrorCode PetscWrapper::GetLocalPetscVector
(
	DM da,
	Vec (&l)
)
{
	DMGetLocalVector(da,&l);
	return(0);
}


PetscErrorCode PetscWrapper::GetLocalPetscVector
(
	DM da,
	Vec* (&l),
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMGetLocalVector(da,&l[i]);
	}
	return(0);
}


PetscErrorCode PetscWrapper::RestoreLocalPetscVector
(
	DM da,
	Vec (&l)
)
{
	DMRestoreLocalVector(da,&l);
	return(0);
}


PetscErrorCode PetscWrapper::RestoreLocalPetscVector
(
	DM da,
	Vec* (&l),
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMRestoreLocalVector(da,&l[i]);
	}
	return(0);
}


template<typename T>
PetscErrorCode PetscWrapper::GetPetscArray
(
	DM da,
	Vec g,
	T (&a)
)
{
	DMDAVecGetArray(da,g,&a);
	return(0);
}


template<typename T>
PetscErrorCode PetscWrapper::GetPetscArray
(
	DM da,
	Vec* g,
	T* (&a),
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMDAVecGetArray(da,g[i],&a[i]);
	}
	return(0);
}


//2D/3D explicit instantiation
template PetscErrorCode PetscWrapper::GetPetscArray<PetscScalar**>(DM,Vec,PetscScalar**&);
template PetscErrorCode PetscWrapper::GetPetscArray<PetscScalar**>(DM,Vec*,PetscScalar***&,PetscInt);
template PetscErrorCode PetscWrapper::GetPetscArray<PetscScalar***>(DM,Vec ,PetscScalar***&);
template PetscErrorCode PetscWrapper::GetPetscArray<PetscScalar***>(DM,Vec*,PetscScalar****&,PetscInt);


template<typename T>
PetscErrorCode PetscWrapper::RestorePetscArray
(
	DM da,
	Vec g,
	T (&a)
)
{
	DMDAVecRestoreArray(da,g,&a);
	return(0);
}


template<typename T>
PetscErrorCode PetscWrapper::RestorePetscArray
(
	DM da,
	Vec* g,
	T* (&a),
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMDAVecRestoreArray(da,g[i],&a[i]);
	}
	return(0);
}


//2D/3D explicit instantiation
template PetscErrorCode PetscWrapper::RestorePetscArray<PetscScalar**>(DM,Vec,PetscScalar**&);
template PetscErrorCode PetscWrapper::RestorePetscArray<PetscScalar**>(DM,Vec*,PetscScalar***&,PetscInt);
template PetscErrorCode PetscWrapper::RestorePetscArray<PetscScalar***>(DM,Vec ,PetscScalar***&);
template PetscErrorCode PetscWrapper::RestorePetscArray<PetscScalar***>(DM,Vec*,PetscScalar****&,PetscInt);


template<typename T>
PetscErrorCode PetscWrapper::GetGhostedPetscArray
(
	DM da,
	Vec g,
	Vec l,
	T (&a)
)
{
	DMGlobalToLocalBegin(da,g,INSERT_VALUES,l);
	DMGlobalToLocalEnd(da,g,INSERT_VALUES,l);
	DMDAVecGetArrayRead(da,l,&a);
	return(0);
}


template<typename T>
PetscErrorCode PetscWrapper::GetGhostedPetscArray
(
	DM da,
	Vec* g,
	Vec* l,
	T* (&a),
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMGlobalToLocalBegin(da,g[i],INSERT_VALUES,l[i]);
		DMGlobalToLocalEnd(da,g[i],INSERT_VALUES,l[i]);
		DMDAVecGetArrayRead(da,l[i],&a[i]);
	}
	return(0);
}


//2D/3D explicit instantiation
template PetscErrorCode PetscWrapper::GetGhostedPetscArray<PetscScalar**>(DM,Vec,Vec,PetscScalar**&);
template PetscErrorCode PetscWrapper::GetGhostedPetscArray<PetscScalar**>(DM,Vec*,Vec*,PetscScalar***&,PetscInt);
template PetscErrorCode PetscWrapper::GetGhostedPetscArray<PetscScalar***>(DM,Vec ,Vec ,PetscScalar***&);
template PetscErrorCode PetscWrapper::GetGhostedPetscArray<PetscScalar***>(DM,Vec*,Vec*,PetscScalar****&,PetscInt);


template<typename T>
PetscErrorCode PetscWrapper::RestoreGhostedPetscArray
(
	DM da,
	Vec l,
	T (&a)
)
{
	DMDAVecRestoreArrayRead(da,l,&a);
	return(0);
}


template<typename T>
PetscErrorCode PetscWrapper::RestoreGhostedPetscArray
(
	DM da,
	Vec* l,
	T* (&a),
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		DMDAVecRestoreArrayRead(da,l[i],&a[i]);
	}
	return(0);
}


//2D/3D explicit instantiation
template PetscErrorCode PetscWrapper::RestoreGhostedPetscArray<PetscScalar**>(DM,Vec,PetscScalar**&);
template PetscErrorCode PetscWrapper::RestoreGhostedPetscArray<PetscScalar**>(DM,Vec*,PetscScalar***&,PetscInt);
template PetscErrorCode PetscWrapper::RestoreGhostedPetscArray<PetscScalar***>(DM,Vec,PetscScalar***&);
template PetscErrorCode PetscWrapper::RestoreGhostedPetscArray<PetscScalar***>(DM,Vec*,PetscScalar****&,PetscInt);


PetscErrorCode PetscWrapper::WritePetscVectorInBinaryFormat
(
	Vec g,
	PetscInt t
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g,&vname);
	std::string fname = GetFileName(vname,t,"dat");
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname.c_str(),FILE_MODE_WRITE,&viewer);
	VecView(g,viewer);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WritePetscVectorInBinaryFormat
(
	Vec* g,
	PetscInt t,
	PetscInt n
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g[0],&vname);
	for(PetscInt i=0;i<n;++i)
	{
		char fname[1000];
		if (t<0)
		{
			sprintf(fname,"%s%.*d.dat",vname,2,i);
		}
		else
		{
			sprintf(fname,"%s%.*d_t%.*d.dat",vname,2,i,TIMEWIDTH_,t);
		}
		PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
		PetscViewer viewer;
		PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
		VecView(g[i],viewer);
		PetscViewerDestroy(&viewer);
	}
	return(0);
}


PetscErrorCode PetscWrapper::ReadPetscVectorInBinaryFormat
(
	Vec g,
	PetscInt t
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g,&vname);
	std::string fname = GetFileName(vname,t,"dat");
	CheckIfFileExist(fname);
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname.c_str(),FILE_MODE_READ,&viewer);
	VecLoad(g,viewer);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::ReadPetscVectorInBinaryFormat
(
	Vec* g,
	PetscInt t,
	PetscInt n
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g[0],&vname);
	for(PetscInt i=0;i<n;++i)
	{
		char fname[1000];
		if (t<0)
		{
			sprintf(fname,"%s%.*d.dat",vname,2,i);
		}
		else
		{
			sprintf(fname,"%s%.*d_t%.*d.dat",vname,2,i,TIMEWIDTH_,t);
		}
		CheckIfFileExist(fname);
		PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname);
		PetscViewer viewer;
		PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_READ,&viewer);
		VecLoad(g[i],viewer);
		PetscViewerDestroy(&viewer);
	}
	return(0);
}


PetscErrorCode PetscWrapper::WritePetscVectorInVtsFormat
(
	Vec g,
	PetscInt t
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g,&vname);
	std::string fname = GetFileName(vname,t,"vts");
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerVTKOpen(PETSC_COMM_WORLD,fname.c_str(),FILE_MODE_WRITE,&viewer);
	VecView(g,viewer);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WritePetscVectorInVtsFormat
(
	Vec* g,
	PetscInt t,
	PetscInt n
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g[0],&vname);
	for(PetscInt i=0;i<n;++i)
	{
		char fname[1000];
		if (t<0)
		{
			sprintf(fname,"%s%.*d.vts",vname,2,i);
		}
		else
		{
			sprintf(fname,"%s%.*d_t%.*d.vts",vname,2,i,TIMEWIDTH_,t);
		}
		PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
		PetscViewer viewer;
		PetscViewerVTKOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
		VecView(g[i],viewer);
		PetscViewerDestroy(&viewer);
	}
	return(0);
}


PetscErrorCode PetscWrapper::WritePetscVectorInVtsFormatWithXYZSubscripts
(
	Vec* g,
	PetscInt t,
	PetscInt n
)
{
	const char *vname;
	PetscObjectGetName((PetscObject)g[0],&vname);
	if (n>3)
	{
		PetscPrintf(PETSC_COMM_WORLD,"Error: %s has more than three dimensions\n",vname);
		PetscEnd();
	}
	char subscript[] = ("xyz");
	for(PetscInt i=0;i<n;++i)
	{
		char fname[1000];
		if (t<0)
		{
			sprintf(fname,"%s%c.vts",vname,subscript[i]);
		}
		else
		{
			sprintf(fname,"%s%c_t%.*d.vts",vname,subscript[i],TIMEWIDTH_,t);
		}
		PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
		PetscViewer viewer;
		PetscViewerVTKOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
		VecView(g[i],viewer);
		PetscViewerDestroy(&viewer);
	}
	return(0);
}


PetscErrorCode PetscWrapper::WriteScalarInBinaryFormat
(
	PetscInt s,
	const std::string &name,
	PetscInt t
)
{
	std::string fname = GetFileName(name,t,"dat");
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname.c_str(),FILE_MODE_WRITE,&viewer);
	PetscViewerBinaryWrite(viewer,&s,1,PETSC_INT);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WriteScalarInBinaryFormat
(
	const PetscInt *const s,
	const std::string &name,
	PetscInt t,
	PetscInt n
)
{
	std::string fname = GetFileName(name,t,"dat");
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname.c_str(),FILE_MODE_WRITE,&viewer);
	PetscViewerBinaryWrite(viewer,s,n,PETSC_INT);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WriteScalarInBinaryFormat
(
	PetscScalar s,
	const std::string &name,
	PetscInt t
)
{
	std::string fname = GetFileName(name,t,"dat");
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname.c_str(),FILE_MODE_WRITE,&viewer);
	PetscViewerBinaryWrite(viewer,&s,1,PETSC_SCALAR);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WriteScalarInBinaryFormat
(
	const PetscScalar *const s,
	const std::string &name,
	PetscInt t,
	PetscInt n
)
{
	std::string fname = GetFileName(name,t,"dat");
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname.c_str(),FILE_MODE_WRITE,&viewer);
	PetscViewerBinaryWrite(viewer,s,n,PETSC_SCALAR);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WriteScalarInAsciiFormat
(
	PetscInt s,
	const std::string &name,
	PetscInt t
)
{
	std::string fname = GetFileName(name,t,"txt");
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerASCIIOpen(PETSC_COMM_WORLD,fname.c_str(),&viewer);
	PetscViewerASCIIPrintf(viewer,"%d\n",s);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WriteScalarInAsciiFormat
(
	PetscScalar s,
	const std::string &name,
	PetscInt t
)
{
	std::string fname = GetFileName(name,t,"txt");
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerASCIIOpen(PETSC_COMM_WORLD,fname.c_str(),&viewer);
	PetscViewerASCIIPrintf(viewer,"%.8f\n",s);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::WriteScalarInAsciiFormat
(
	const PetscScalar *const s,
	const std::string &name,
	PetscInt t,
	PetscInt n
)
{
	std::string fname = GetFileName(name,t,"txt");
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerASCIIOpen(PETSC_COMM_WORLD,fname.c_str(),&viewer);
	for (PetscInt i=0; i<n; ++i)
	{
		PetscViewerASCIIPrintf(viewer,"%.8f\n",s[i]);
	}
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::ReadScalarInBinaryFormat
(
	PetscInt &s,
	const std::string &name,
	PetscInt t
)
{
	std::string fname = GetFileName(name,t,"dat");
	CheckIfFileExist(fname);
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname.c_str(),FILE_MODE_READ,&viewer);
	PetscViewerBinaryRead(viewer,&s,1,NULL,PETSC_INT);
	MPI_Bcast(&s,1,MPIU_INT,0,PETSC_COMM_WORLD);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::ReadScalarInBinaryFormat
(
	PetscScalar &s,
	const std::string &name,
	PetscInt t
)
{
	std::string fname = GetFileName(name,t,"dat");
	CheckIfFileExist(fname);
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname.c_str(),FILE_MODE_READ,&viewer);
	PetscViewerBinaryRead(viewer,&s,1,NULL,PETSC_SCALAR);
	MPI_Bcast(&s,1,MPIU_SCALAR,0,PETSC_COMM_WORLD);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::ReadScalarInBinaryFormat
(
	PetscInt *s,
	const std::string &name,
	PetscInt t,
	PetscInt n
)
{
	std::string fname = GetFileName(name,t,"dat");
	CheckIfFileExist(fname);
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname.c_str(),FILE_MODE_READ,&viewer);
	PetscViewerBinaryRead(viewer,s,n,NULL,PETSC_INT);
	MPI_Bcast(s,n,MPIU_INT,0,PETSC_COMM_WORLD);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::ReadScalarInBinaryFormat
(
	PetscScalar *s,
	const std::string &name,
	PetscInt t,
	PetscInt n
)
{
	std::string fname = GetFileName(name,t,"dat");
	CheckIfFileExist(fname);
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname.c_str());
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname.c_str(),FILE_MODE_READ,&viewer);
	PetscViewerBinaryRead(viewer,s,n,NULL,PETSC_SCALAR);
	MPI_Bcast(s,n,MPIU_SCALAR,0,PETSC_COMM_WORLD);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode PetscWrapper::SetPetscVector
(
	Vec g,
	PetscScalar s
)
{
	VecSet(g,s);
	return(0);
}


PetscErrorCode PetscWrapper::SetPetscVector
(
	Vec* g,
	PetscScalar s,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		VecSet(g[i],s);
	}
	return(0);
}


PetscErrorCode PetscWrapper::SetPetscVector
(
	Vec* g,
	const PetscScalar *const s,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		VecSet(g[i],s[i]);
	}
	return(0);
}


PetscErrorCode PetscWrapper::ZeroPetscVector
(
	Vec g
)
{
	VecSet(g,0.);
	return(0);
}


PetscErrorCode PetscWrapper::ZeroPetscVector
(
	Vec* g,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		VecSet(g[i],0.);
	}
	return(0);
}


PetscErrorCode PetscWrapper::ComputeSumOfPetscVectorElements
(
	Vec v,
	PetscScalar &s
)
{
	VecSum(v,&s);
	return(0);
}


PetscErrorCode PetscWrapper::ComputeSumOfPetscVectorElements
(
	Vec* v,
	PetscScalar *s,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		VecSum(v[i],&s[i]);
	}
	return(0);
}


PetscErrorCode PetscWrapper::ComputeArithmeticMeanOfPetscVectorElements
(
	Vec g,
	PetscScalar &m
)
{
	PetscInt size;
	VecGetSize(g,&size);
	VecSum(g,&m);
	m /= static_cast<PetscScalar>(size);
	return(0);
}


PetscErrorCode PetscWrapper::ComputeArithmeticMeanOfPetscVectorElements
(
	Vec* g,
	PetscScalar *m,
	PetscInt n
)
{
	PetscInt size;
	for (PetscInt i=0;i<n;++i)
	{
		VecGetSize(g[i],&size);
		VecSum(g[i],&m[i]);
		m[i] /= static_cast<PetscScalar>(size);
	}
	return(0);
}


// PetscErrorCode PetscWrapper::CopyPetscVectorToAndRaiseElements
// (
// 	Vec a,
// 	PetscScalar exp,
// 	Vec b
// )
// {
// 	VecCopy(a,b);
// 	VecPow(b,exp);
// 	return(0);
// }


// PetscErrorCode PetscWrapper::CopyPetscVectorToAndRaiseElements
// (
// 	Vec* a,
// 	PetscScalar exp,
// 	Vec* b,
// 	PetscInt n
// )
// {
// 	for (PetscInt i=0;i<n;++i)
// 	{
// 		VecCopy(a[i],b[i]);
// 		VecPow(b[i],exp);
// 	}
// 	return(0);
// }


PetscErrorCode PetscWrapper::CopyPetscVectorTo
(
	Vec from,
	Vec to
)
{
	VecCopy(from,to);
	return(0);
}


PetscErrorCode PetscWrapper::CopyPetscVectorTo
(
	Vec* from,
	Vec* to,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		VecCopy(from[i],to[i]);
	}
	return(0);
}


PetscErrorCode PetscWrapper::SumArrayOfPetscVectors
(
	Vec* g,
	Vec sum,
	PetscInt n
)
{
	PetscScalar *sa; VecGetArray(sum,&sa);
	PetscInt localN; VecGetLocalSize(sum,&localN);
	const PetscScalar *ga;
	PetscInt v=0;
	{
		//g0
		VecGetArrayRead(g[v],&ga);
		for (PetscInt i=0; i<localN; i++)
		{
			sa[i] = ga[i];
		}
		VecRestoreArrayRead(g[v],&ga);
	}
	for (v=1; v<n; ++v)
	{
		//g1,g2,...gn-1
		VecGetArrayRead(g[v],&ga);
		for (PetscInt i=0; i<localN; i++)
		{
			sa[i] += ga[i];
		}
		VecRestoreArrayRead(g[v],&ga);
	}
	VecRestoreArray(sum,&sa);
	return(0);
}


PetscBool PetscWrapper::ArePetscVectorsEqual
(
	Vec va,
	Vec vb,
	PetscScalar eps
)
{
	if (va == vb)
	{
		return PETSC_TRUE;//double check
	}
	Vec w;
	VecDuplicate(va,&w);
	VecWAXPY(w,-1,va,vb);
	PetscScalar norm;
	VecNorm(w,NORM_2,&norm);
	VecDestroy(&w);
	if ( norm < eps )
	{
		return PETSC_TRUE;
	}
	return PETSC_FALSE;
}


PetscErrorCode PetscWrapper::CheckPetscVector
(
	Vec g
)
{
	PetscScalar min; VecMin(g,NULL,&min);
	PetscScalar max; VecMax(g,NULL,&max);
	max = PetscMax
	(
		PetscAbsReal(min),
		PetscAbsReal(max)
	);
	if ( PetscIsInfOrNanScalar(max) || max > 10000. )
	{
		const char *name;
		PetscObjectGetName((PetscObject)g,&name);
		std::string msg = "  Error: " + std::string(name) + " -> infinity";
		PetscPrintf(PETSC_COMM_WORLD,"%s\n",msg.c_str());
		PetscEnd();
	}
	return(0);
}


PetscErrorCode PetscWrapper::CheckPetscVector
(
	Vec* g,
	PetscInt n
)
{
	for (PetscInt i=0;i<n;++i)
	{
		PetscScalar min; VecMin(g[i],NULL,&min);
		PetscScalar max; VecMax(g[i],NULL,&max);
		max = PetscMax
		(
			PetscAbsReal(min),
			PetscAbsReal(max)
		);
		if ( PetscIsInfOrNanScalar(max) || max > 10000. )
		{
			const char *name;
			PetscObjectGetName((PetscObject)g[0],&name);
			std::string msg = "  Error: " + std::string(name) + " -> infinity";
			PetscPrintf(PETSC_COMM_WORLD,"%s\n",msg.c_str());
			PetscEnd();
		}
	}
	return(0);
}


PetscErrorCode PetscWrapper::FlattenMarchingCubeArrayTo
(
	DM da,
	Vec g,
	Vec l,
	PetscScalar*** (&a),
	std::vector<PetscScalar> &v,
	PetscInt &nx,
	PetscInt &ny,
	PetscInt &nz
)
{
	DMGlobalToLocalBegin(da,g,INSERT_VALUES,l);
	DMGlobalToLocalEnd(da,g,INSERT_VALUES,l);
	DMDAVecGetArray(da,l,&a);
	PetscInt i0,j0,k0,iw,jw,kw;
	DMDAGetCorners(da,&i0,&j0,&k0,&iw,&jw,&kw);
	iw+=1; jw+=1; kw+=1;
	v.resize(iw*jw*kw);
	for (PetscInt k=k0; k<k0+kw; ++k)
	{
		for (PetscInt j=j0; j<j0+jw; ++j)
		{
			for (PetscInt i=i0; i<i0+iw; ++i)
			{
				PetscInt index = (i-i0) + iw * ( (j-j0) + jw * (k-k0) );
				v[index] = a[k][j][i];
			}
		}
	}
	DMDAVecRestoreArray(da,l,&a);
	nx = iw;
	ny = jw;
	nz = kw;
	return(0);
}
