#ifndef PETSCWRAPPER_H
#define PETSCWRAPPER_H


#include "petsc.h"
#include <vector>


/** \file
 *  \brief Collection of Petsc functions */


/** \namespace PetscWrapper
  * \brief Collection of Petsc functions */
namespace PetscWrapper
{


	/** \class %GridCoordinates2D
	 *  \brief Object for rapid and easy access to DM grid coordinates */
	class GridCoordinates2D
	{

	public:

		/** \brief Petsc dynamic array */
		DM da_;

		/** \brief Global Petsc vector holding coordinates */
		Vec g_;

		/** \brief Petsc coordinate DM */
		DM cda_;

		/** \brief Petsc structure holding 2d coordinates. */
		const DMDACoor2d **a_;

		/** \brief Set the Petsc dynamic array
		 *  \param[in] da DMDA distributed array */
		PetscErrorCode create(DM da);

		/** \brief Destroy da */
		PetscErrorCode destroy();

		/** \brief Const reference to x[j][i]
		*   \param[in] j node index
		*   \param[in] i node index
		*   \return const reference to a[j][i].x */
		const PetscScalar &x(PetscInt j,PetscInt i) const {return a_[j][i].x;}

		/** \brief Const reference to y[j][i]
		*   \param[in] j node index
		*   \param[in] i node index
		*   \return const reference to a[j][i].y */
		const PetscScalar &y(PetscInt j,PetscInt i) const {return a_[j][i].y;}

		/** \brief Get Petsc coordinate array */
		PetscErrorCode getArray();

		/** \brief Restore coordinate array */
		PetscErrorCode restoreArray();

		/** \brief Get Petsc coordinate array with extra space for ghost values */
		PetscErrorCode getGhostedArray();

		/** \brief Restore coordinate array */
		PetscErrorCode restoreGhostedArray();

	};


	/** \class %GridCoordinates3D
	 *  \brief Object for rapid and easy access to DM grid coordinates */
	class GridCoordinates3D
	{

	public:

		/** \brief Petsc dynamic array */
		DM da_;

		/** \brief Global Petsc vector holding coordinates */
		Vec g_;

		/** \brief Petsc coordinate DM */
		DM cda_;

		/** \brief Set the Petsc dynamic array
		 *  \param[in] da DMDA distributed array */
		PetscErrorCode create(DM da);

		/** \brief Destroy da */
		PetscErrorCode destroy();

		/** \brief Petsc structure holding 3d coordinates. */
		const DMDACoor3d*** a_;

		/** \brief Const reference to x[k][j][i]
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return const reference to a[k][j][i].x */
		const PetscScalar &x(PetscInt k,PetscInt j,PetscInt i) const {return a_[k][j][i].x;}

		/** \brief Const reference to y[k][j][i]
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return const reference to a[k][j][i].y */
		const PetscScalar &y(PetscInt k,PetscInt j,PetscInt i) const {return a_[k][j][i].y;}

		/** \brief Const reference to z[k][j][i]
		 *  \param[in] k node index
		 *  \param[in] j node index
		 *  \param[in] i node index
		 *  \return const reference to a[k][j][i].y */
		const PetscScalar &z(PetscInt k,PetscInt j,PetscInt i) const {return a_[k][j][i].z;}

		/** \brief Get Petsc coordinate array */
		PetscErrorCode getArray();

		/** \brief Restore coordinate array */
		PetscErrorCode restoreArray();

		/** \brief Get Petsc coordinate array with extra space for ghost values */
		PetscErrorCode getGhostedArray();

		/** \brief Restore coordinate array */
		PetscErrorCode restoreGhostedArray();

		/** \brief Copy grid point coordinate to array
		 *  \param[in]  k  node index
		 *  \param[in]  j  node index
		 *  \param[in]  i  node index
		 *  \param[out] pt (x,y,z) */
		PetscErrorCode copyGridPointTo(PetscInt i,PetscInt j,PetscInt k,PetscScalar pt[3]);

		/** \brief Copy grid coordinates to one-dimensional array.
		 *         Use Map<%EigenMat>(v.data(),rows,cols) to convert to Eigen matrix without using new memory.
		 *  \param[out] v std vector array v[x0, x1, ..., x_{rows-1}, y0, y1, ..., y_{rows-1}, z0, z1, ..., z_{rows-1}]
		 *  \param[out] rows number of rows
		 *  \param[out] cols number of cols = 3 */
		PetscErrorCode flattenArrayTo(std::vector<PetscScalar> &v,PetscInt &rows,PetscInt &cols);

		/** \brief Copy grid coordinates to one-dimensional array needed in marching cube function.
		 *         Use Eigen::Map<%const Eigen::Mat>(v.data(),rows,cols) to convert to Eigen matrix without creating new memory
		 *  \param[out] v std vector array v[x0, x1, ..., x_{rows-1}, y0, y1, ..., y_{rows-1}, z0, z1, ..., z_{rows-1}]
		 *  \param[out] rows number of rows
		 *  \param[out] cols number of cols = 3 */
		PetscErrorCode flattenMarchingCubeArrayTo(std::vector<PetscScalar> &v,PetscInt &rows,PetscInt &cols);

	};


	/** \brief Create global Petsc vector
	 *  \param[in]  da   DMDA distributed array
	 *  \param[out] g    global Petsc vector
	 *  \param[in]  name Petsc vector name */
	PetscErrorCode CreateGlobalPetscVector(DM da,Vec (&g),const std::string &name);


	/** \brief Create array of global Petsc vectors
	 *  \param[in]  da   DMDA distributed array
	 *  \param[out] g    array of global Petsc vectors g = [g0,g1,...,g_{n-1}]
	 *  \param[in]  name Petsc vector name
	 *  \param[in]  n    size */
	PetscErrorCode CreateGlobalPetscVector(DM da,Vec* (&g),const std::string &name,PetscInt n);


	/** \brief Create local Petsc vector with extra space for ghost values
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  local vector */
	PetscErrorCode CreateLocalPetscVector(DM da,Vec (&l));


	/** \brief Create array of local Petsc vectors with extra space for ghost values
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  array of local Petsc vectors l = [l0,l1,...,l_{n-1}]
	 *  \param[in]  n  size */
	PetscErrorCode CreateLocalPetscVector(DM da,Vec* (&l),PetscInt n);


	/** \brief Destroy Petsc vector
	 *  \param[out] v global/local vector */
	PetscErrorCode DestroyPetscVector(Vec (&v));


	/** \brief Destroy array of Petsc vectors
	 *  \param[out] v array of global/local vector  v = [v0,v1,...,v_{n-1}]
	 *  \param[in]  n size */
	PetscErrorCode DestroyPetscVector(Vec* (&v),PetscInt n);


	/** \brief Get local Petsc vector with extra space for ghost values
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  local vector */
	PetscErrorCode GetLocalPetscVector(DM da,Vec (&l));


	/** \brief Get array of local Petsc vectors with extra space for ghost values
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  array of local Petsc vectors l = [l0,l1,...,l_{n-1}]
	 *  \param[in]  n  size */
	PetscErrorCode GetLocalPetscVector(DM da,Vec* (&l),PetscInt n);


	/** \brief Restore local Petsc vector
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  local vector */
	PetscErrorCode RestoreLocalPetscVector(DM da,Vec (&l));


	/** \brief Restore array of local Petsc vectors
	 *  \param[in]  da DMDA distributed array
	 *  \param[out] l  array of local Petsc vectors l = [l0,l1,...,l_{n-1}]
	 *  \param[in]  n  size */
	PetscErrorCode RestoreLocalPetscVector(DM da,Vec* (&l),PetscInt n);


	/** \brief Get multidimensional C-array from global Petsc vector
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  global vector
	 *  \param[out] a  multidimensional C-array */
	template<typename T> PetscErrorCode GetPetscArray(DM da,Vec g,T (&a));


	/** \brief Get array of multidimensional C-arrays from array of global Petsc vectors
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  array of global Petsc vectors g = [g0,g1,...,g_{n-1}]
	 *  \param[out] a  array of multidimensional C-arrays a = [a0,a0,...,a_{n-1}]
	 *  \param[in]  n  size */
	template<typename T> PetscErrorCode GetPetscArray(DM da,Vec* g,T* (&a),PetscInt n);


	/** \brief Restore global Petsc vector from multidimensional C-array
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  global vector
	 *  \param[out] a  multidimensional C-array */
	template<typename T> PetscErrorCode RestorePetscArray(DM da,Vec g,T (&a));


	/** \brief Restore array of global Petsc vectors from array of multidimensional C-arrays
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  array of global Petsc vectors g = [g0,g1,...,g_{n-1}]
	 *  \param[out] a  array of multidimensional C-arrays a = [a0,a1,...,a_{n-1}]
	 *  \param[in]  n  size */
	template<typename T> PetscErrorCode RestorePetscArray(DM da,Vec* g,T* (&a),PetscInt n);


	/** \brief Get multidimensional C-array containing ghost points. This is done after scattering global Petsc vector to local parts
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  global vector
	 *  \param[out] l  local vector
	 *  \param[out] a  multidimensional C-array */
	template<typename T> PetscErrorCode GetGhostedPetscArray(DM da,Vec g,Vec l,T (&a));


	/** \brief Get array of multidimensional C-arrays containing ghost points. This is done after scattering global Petsc vectors to local parts
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  array of global Petsc vectors g = [g0,g1,...,g_{n-1}]
	 *  \param[out] l  array of local Petsc vectors l = [l0,l1,...,l_{n-1}]
	 *  \param[out] a  array of multidimensional C-arrays a = [a0,a1,...,a_{n-1}]
	 *  \param[in]  n  size */
	template<typename T> PetscErrorCode GetGhostedPetscArray(DM da,Vec* g,Vec *l,T* (&a),PetscInt n);


	/** \brief Restore local Petsc vector from multidimensional C-array containing ghost points
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  l  local vector
	 *  \param[out] a  multidimensional C-array */
	template<typename T> PetscErrorCode RestoreGhostedPetscArray(DM da,Vec l,T (&a));


	/** \brief Restore array of local Petsc vectors from array of multidimensional C-arrays containing ghost points
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  l  array of local Petsc vectors l = [l0,l1,...,l_{n-1}]
	 *  \param[out] a  array of multidimensional C-arrays a = [a0,a1,...,a_{n-1}]
	 *  \param[in]  n  size */
	template<typename T> PetscErrorCode RestoreGhostedPetscArray(DM da,Vec *l,T* (&a),PetscInt n);


	/** \brief Write Petsc vector in binary format
	 *  \param[in] g global Petsc vector
	 *  \param[in] t time index */
	PetscErrorCode WritePetscVectorInBinaryFormat(Vec g,PetscInt t);


	/** \brief Write array of Petsc vectors in binary format
	 *  \param[in] g array of global Petsc vector g = [g0,g1,...,g_{n-1}]
	 *  \param[in] t time index
	 *  \param[in] n size */
	PetscErrorCode WritePetscVectorInBinaryFormat(Vec* g,PetscInt t,PetscInt n);


	/** \brief Read Petsc vector in binary format
	 *  \param[out] g global Petsc vector
	 *  \param[in]  t time index */
	PetscErrorCode ReadPetscVectorInBinaryFormat(Vec g,PetscInt t);


	/** \brief Read array of Petsc vectors in binary format
	 *  \param[out] g array of global Petsc vectors g = [g0,g1,...,g_{n-1}]
	 *  \param[in]  t time index
	 *  \param[in]  n size */
	PetscErrorCode ReadPetscVectorInBinaryFormat(Vec* g,PetscInt t,PetscInt n);


	/** \brief Write Petsc vector in VTS format
	 *  \param[in] g global Petsc vector
	 *  \param[in] t time index */
	PetscErrorCode WritePetscVectorInVtsFormat(Vec g,PetscInt t);


	/** \brief Write array of Petsc vectors in VTS format
	 *  \param[in] g array of global Petsc vector g = [g0,g1,...,g_{n-1}]
	 *  \param[in] t time index
	 *  \param[in] n size */
	PetscErrorCode WritePetscVectorInVtsFormat(Vec* g,PetscInt t,PetscInt n);


	/** \brief Write array of Petsc vectors in VTS format with x,y,z subscripts
	 *  \param[in] g array of global Petsc vector g = [g0,g1,g2]
	 *  \param[in] t time index
	 *  \param[in] n size */
	PetscErrorCode WritePetscVectorInVtsFormatWithXYZSubscripts(Vec* g,PetscInt t,PetscInt n);


	/** \brief Write integer
	 *  \param[in] s    integer
	 *  \param[in] name file name
	 *  \param[in] t    time index */
	PetscErrorCode WriteScalarInBinaryFormat(PetscInt s,const std::string &name,PetscInt t);


	/** \brief Write one-dimensional integer array
	 *  \param[in] s    integer
	 *  \param[in] name file name
	 *  \param[in] t    time index
	 *  \param[in] n    size */
	PetscErrorCode WriteScalarInBinaryFormat(const PetscInt *const s,const std::string &name,PetscInt t,PetscInt n);


	/** \brief Write scalar
	 *  \param[in] s    scalar
	 *  \param[in] name file name
	 *  \param[in] t    time index */
	PetscErrorCode WriteScalarInBinaryFormat(PetscScalar s,const std::string &name,PetscInt t);


	/** \brief Write one-dimensional scalar array
	 *  \param[in] s    one-dimensional scalar array
	 *  \param[in] name file name
	 *  \param[in] t    time index
	 *  \param[in] n    size */
	PetscErrorCode WriteScalarInBinaryFormat(const PetscScalar *const s,const std::string &name,PetscInt t,PetscInt n);


	/** \brief Write integer
	 *  \param[in] s    integer
	 *  \param[in] name file name
	 *  \param[in] t    time index */
	PetscErrorCode WriteScalarInAsciiFormat(PetscInt s,const std::string &name,PetscInt t);


	/** \brief Write scalar
	 *  \param[in] s    scalar
	 *  \param[in] name file name
	 *  \param[in] t    time index */
	PetscErrorCode WriteScalarInAsciiFormat(PetscScalar s,const std::string &name,PetscInt t);


	/** \brief Write one-dimensional scalar array
	 *  \param[in] s    one-dimensional scalar array
	 *  \param[in] name file name
	 *  \param[in] t    time index
	 *  \param[in] n    size */
	PetscErrorCode WriteScalarInAsciiFormat(const PetscScalar *const s,const std::string &name,PetscInt t,PetscInt n);


	/** \brief Read integer
	 *  \param[in] s    integer
	 *  \param[in] name file name
	 *  \param[in] t    time index */
	PetscErrorCode ReadScalarInBinaryFormat(PetscInt &s,const std::string &name,PetscInt t);


	/** \brief Read one-dimensional integer array
	 *  \param[in] s    integer
	 *  \param[in] name file name
	 *  \param[in] t    time index
	 *  \param[in] n    size */
	PetscErrorCode ReadScalarInBinaryFormat(PetscInt *s,const std::string &name,PetscInt t,PetscInt n);


	/** \brief Read scalar
	 *  \param[in] s    scalar
	 *  \param[in] name file name
	 *  \param[in] t    time index */
	PetscErrorCode ReadScalarInBinaryFormat(PetscScalar &s,const std::string &name,PetscInt t);


	/** \brief Read one-dimensional scalar array
	 *  \param[in] s    one-dimensional scalar array
	 *  \param[in] name file name
	 *  \param[in] t    time index
	 *  \param[in] n    size */
	PetscErrorCode ReadScalarInBinaryFormat(PetscScalar *s,const std::string &name,PetscInt t,PetscInt n);


	/** \brief Set vector elements to a scalar value
	 *  \param[out] g global Petsc vector
	 *  \param[in]  s scalar */
	PetscErrorCode SetPetscVector(Vec g,PetscScalar s);


	/** \brief Set vector elements to a scalar value
	 *  \param[out] g array of global Petsc vectors g = [g0,g1,...,g_{n-1}]
	 *  \param[in]  s scalar
	 *  \param[in]  n size */
	PetscErrorCode SetPetscVector(Vec* g,PetscScalar s,PetscInt n);


	/** \brief Set vector elements to a scalar value
	 *  \param[out] g array of global Petsc vector g = [g0,g1,...,g_{n-1}]
	 *  \param[in]  s scalar array s = [s0,s1,...,s_{n-1}]
	 *  \param[in]  n size*/
	PetscErrorCode SetPetscVector(Vec* g,const PetscScalar *const s,PetscInt n);


	/** \brief Set vector elements to zero
	 *  \param[out] g global Petsc vector */
	PetscErrorCode ZeroPetscVector(Vec g);


	/** \brief Set vector elements to zero
	 *  \param[out] g global Petsc vector
	 *  \param[in]  n size */
	PetscErrorCode ZeroPetscVector(Vec* g,PetscInt n);


	/** \brief Sum elements of a global Petsc vector
	*   \param[in]  g global Petsc vector
	*   \param[out] s sum */
	PetscErrorCode ComputeSumOfPetscVectorElements(Vec g,PetscScalar &s);


	/** \brief Sum elements of each Petsc vector
	*   \param[in]  g array of global Petsc vectors g = [g0,g1,...,g_{n-1}]
	*   \param[out] s sum array s = [s0,s1,...,s_{n-1}]
	*   \param[in]  n size */
	PetscErrorCode ComputeSumOfPetscVectorElements(Vec* g,PetscScalar *s,PetscInt n);


	/** \brief Arithmetic mean of elements in Petsc vector
	 *  \param[in]  g global Petsc vector
	 *  \param[out] m mean */
	PetscErrorCode ComputeArithmeticMeanOfPetscVectorElements(Vec g,PetscScalar &m);


	/** \brief Arithmetic mean of elements in Petsc vector
	 *  \param[in]  g array of Petsc vectors g = [g0,g1,...,g_{n-1}]
	 *  \param[out] m array of means m = [m0,m1,...,m_{n-1}]
	 *  \param[in]  n size*/
	PetscErrorCode ComputeArithmeticMeanOfPetscVectorElements(Vec* g,PetscScalar *m,PetscInt n);


// 	/** \brief Copy and raise each element of a global Petsc vector
// 	 *  \param[in]  a   global Petsc vector
// 	 *  \param[in]  exp exponent
// 	 *  \param[out] b   global Petsc vector */
// 	PetscErrorCode CopyPetscVectorToAndRaiseElements(Vec a,PetscScalar exp,Vec b);


// 	/** \brief Copy and raise each element of a global Petsc vector
// 	 *  \param[in]  a   array of global Petsc vectors
// 	 *  \param[in]  exp exponent
// 	 *  \param[out] b   array of global Petsc vector */
// 	PetscErrorCode CopyPetscVectorToAndRaiseElements(Vec *a,PetscScalar exp,Vec *b,PetscInt n);


	/** \brief Copy global Petsc vector
	 *  \param[in]  from global Petsc vector
	 *  \param[out] to   global Petsc vector */
	PetscErrorCode CopyPetscVectorTo(Vec from,Vec to);


	/** \brief Copy array of global Petsc vectors
	*   \param[in]  from array of Petsc vectors from = [from0,from1,...,from_{n-1}]
	*   \param[out] to   array of Petsc vectors to = [to0,to1,...,to_{n-1}]
	*   \param[in]  n    size */
	PetscErrorCode CopyPetscVectorTo(Vec *from,Vec *to,PetscInt n);


	/** \brief Addition an array of Petsc vectors
	 *  \param[in]  g   array of Petsc vectors g = [g0,g1,...,g_{n-1}]
	 *  \param[out] sum resulting sum as global Petsc vector
	 *  \param[in]  n   size */
	PetscErrorCode SumArrayOfPetscVectors(Vec* g,Vec sum,PetscInt n);


	/** \brief Check whether two Petsc vectors are equal
	 *  \param[in] a   global Petsc vector
	 *  \param[in] b   global Petsc vector
	 *  \param[in] eps true if |va-vb| < epsilon */
	PetscBool ArePetscVectorsEqual(Vec a,Vec b,PetscScalar eps=1e-8);


	/** \brief Check convergence of a global Petsc vector
	 *  \param[in] g global Petsc vector */
	PetscErrorCode CheckPetscVector(Vec g);


	/** \brief Check convergence of a array of Petsc vectors
	 *  \param[in] g array of global Petsc vector g = [g0,g1,...,g_{n-1}]
	 *  \param[in] n size */
	PetscErrorCode CheckPetscVector(Vec* g,PetscInt n);


	/** \brief Copy Petsc data to one-dimensional array with extra space for ghosted values at positions kw+1, jw+1 and iw+1
	 *         Use Eigen::Map<%const Eigen::Vec>(v.data(),v.size()) to convert to Eigen vector
	 *  \param[in]  da DMDA distributed array
	 *  \param[in]  g  global Petsc vector
	 *  \param[in]  l  local Petsc vector
	 *  \param[in]  a  multidimensional C-array
	 *  \param[out] v  one dimensional std vector array, ordered as v[i + nx x ( j + ny x k )]
	 *  \param[out] nx data number in x direction
	 *  \param[out] ny data number in y direction
	 *  \param[out] nz data number in z direction */
	PetscErrorCode FlattenMarchingCubeArrayTo(DM da,Vec g,Vec l,PetscScalar*** (&a),std::vector<PetscScalar> &v,PetscInt &nx,PetscInt &ny,PetscInt &nz);

}


#endif


/*
	Read-only access is done as follows
	const PetscScalar ***a;
	DMDAVecGetPetscArrayRead(da,g,(&a));
	... = a[k][j][i];
	DMDAVecRestorePetscArrayRead(da,g,(&a));
*/
