#include "PhiEquation3D.h"
#include "Discretisation3D.h"
#include "Interpolation3D.h"
#include "PetscWrapper.h"
#include "Utils.h"


using namespace PetscWrapper;
using namespace Field3D;
using namespace Interpolation3D;
using namespace Discretisation3D;


GetDirectionIndices();


PetscErrorCode PhiEquation3D::ComputeBubbleVolume
(
	Context3D::Ctx &ctx
)
{
	GetFieldArray(ctx.phi);
	ForAllB(ctx.phi.n)
	{
		PetscScalar lV = 0.;
		PetscScalar gV;
		ForAllKJI(ctx.da,
		{
			lV += ctx.phi(b,k,j,i);
		})
		MPI_Allreduce(&lV,&gV,1,MPIU_SCALAR,MPIU_SUM,PETSC_COMM_WORLD);
		ctx.bubbleVolume[b] = gV;
	}
	RestoreFieldArray(ctx.phi);
	return(0);
}


PetscErrorCode PhiEquation3D::ComputeSqrBubbleVolume
(
	Context3D::Ctx &ctx
)
{
	GetFieldArray(ctx.phi);
	ForAllB(ctx.phi.n)
	{
		PetscScalar lV = 0.;
		PetscScalar gV;
		ForAllKJI(ctx.da,
		{
			lV += PetscSqr( ctx.phi(b,k,j,i) );
		})
		MPI_Allreduce(&lV,&gV,1,MPIU_SCALAR,MPIU_SUM,PETSC_COMM_WORLD);
		ctx.bubbleVolume[b] = gV;
	}
	RestoreFieldArray(ctx.phi);
	return(0);
}


PetscErrorCode PhiEquation3D::ComputeBulk
(
	Context3D::Ctx &ctx
)
{
	PetscScalar theta = 40;
	PetscScalar cosTheta = PetscCosReal( M_PI * theta/180. );
	ComputeBubbleVolume(ctx);
	GetFieldArray(ctx.muBulk,ctx.phi);
	GetGhostedFieldArray(ctx.phis);
	ForAllB(ctx.nb)
	{
		//corr < 0 [growth] else [shrinkage]
		PetscScalar corr = (ctx.bubbleVolume[b])/(ctx.initialBubbleVolume[b]) - 1.;
		PetscScalar phic = 0.5 + (ctx.bubbleGrowthCoefficient) * corr;
		ForAllJI(ctx.da,
		{
			PetscScalar nx = ( ctx.phis(k  ,j  ,i+1) - ctx.phis(k  ,j  ,i-1) )/2.;
			PetscScalar ny = ( ctx.phis(k  ,j+1,i  ) - ctx.phis(k  ,j-1,i  ) )/2.;
			PetscScalar nz = ( ctx.phis(k+1,j  ,i  ) - ctx.phis(k-1,j  ,i  ) )/2.;
			PetscScalar nmag = PetscSqrtReal( nx*nx + ny*ny + nz*nz);
			PetscScalar phi  = ctx.phi(b,k,j,i);
			PetscScalar phis = ctx.phis(k,j,i);
			ctx.muBulk(b,j,i)  = 4. * phi * ( phi - 1. ) * ( phi - phic ) ;
			ctx.muBulk(b,j,i) -= 4. * phi * (1-phi-phis) * phis;
			ctx.muBulk(b,j,i) += 4. * (ctx.xi) * phi * (phi-1.) * nmag * cosTheta;
		})
// 		PetscPrintf(PETSC_COMM_WORLD,"  phi_correction = %4.2f\n",corr);
	}
	RestoreFieldArray(ctx.muBulk,ctx.phi);
	RestoreGhostedFieldArray(ctx.phis);
	return(0);
}


PetscErrorCode PhiEquation3D::ComputeRepulsion
(
	Context3D::Ctx &ctx
)
{
	//TODO
	return(0);
}


PetscErrorCode PhiEquation3D::ComputeRHS
(
	Context3D::Ctx &ctx
)
{
	GetFieldArray(ctx.phiRHS);
	GetGhostedFieldArray(ctx.u,ctx.phi,ctx.muBulk,ctx.muRep,ctx.phis);
	ForAllB(ctx.nb)
	{
		ForAllJI(ctx.da,
		{
			PetscScalar uDir[2] =
			{
				LowOrder::InterpolateEdge(ctx.u.x(),X_,k  ,j  ,i+1), //ux @ (k,j,i) = (k    ,j    ,I+1/2)
				LowOrder::InterpolateEdge(ctx.u.y(),Y_,k  ,j+1,i  )  //uy @ (k,j,i) = (k    ,J+1/2,i    )
				LowOrder::InterpolateEdge(ctx.u.z(),Z_,k+1,j  ,i  )  //uz @ (k,j,i) = (K+1/2,j    ,i    )
			};
			PetscScalar uGradPhi = FirstOderUpwind::UGrad(uDir,ctx.phi(b),j,i);
			PetscScalar lapPhi = FourthOrder::Laplacian(ctx.phi(b),j,i);
			PetscScalar gradPhiPhis_x = 0.25 * ( ctx.phi(b,k  ,j  ,i+1) - ctx.phi(b,k  ,j  ,i-1) ) * ( ctx.phis(k  ,j  ,i+1) - ctx.phis(k  ,j  ,i-1) ); //@(k,j,i)
			PetscScalar gradPhiPhis_y = 0.25 * ( ctx.phi(b,k  ,j+1,i  ) - ctx.phi(b,k  ,j-1,i  ) ) * ( ctx.phis(k  ,j+1,i  ) - ctx.phis(k  ,j-1,i  ) ); //@(k,j,i)
			PetscScalar gradPhiPhis_z = 0.25 * ( ctx.phi(b,k+1,j  ,i  ) - ctx.phi(b,k-1,j  ,i  ) ) * ( ctx.phis(k+1,j  ,i  ) - ctx.phis(k-1,j  ,i  ) ); //@(k,j,i)

			PetscScalar divPhiPhis = gradPhiPhis_x + gradPhiPhis_y;
			ctx.phiRHS(b,j,i) =
			(
				  ctx.phi(b,j,i)
				- (ctx.dtInvDx1) * uGradPhi
				- (ctx.dtInvPe) * ctx.muBulk(b,j,i)
// 				- (ctx.dtInvPe) * ctx.muRep(b,j,i)
				+ (ctx.xi2DtInvPeExp) * lapPhi
				- (ctx.xi2DtInvPeExp) * divPhiPhis
			);
		})
	}
	RestoreFieldArray(ctx.phiRHS);
	RestoreGhostedFieldArray(ctx.u,ctx.phi,ctx.muBulk,ctx.muRep,ctx.phis);
	return(0);
}


PetscErrorCode PhiEquation3D::ComputeSurfaceForce
(
	Context3D::Ctx &ctx
)
{
	return(0);
}


PetscErrorCode PhiEquation3D::MakeFoam
(
	Context3D::Ctx &ctx
)
{
	SumVectorField(ctx.phi,ctx.phiSum);
	return(0);
}


PetscErrorCode PhiEquation3D::ComputeChemicalPotential
(
	Context3D::Ctx &ctx
)
{
	//to do
	return(0);
}


PetscErrorCode PhiEquation3D::FormSNESFunction
(
	SNES  snes,
	Vec   g_X,
	Vec   g_F,
	void *ptr
)
{
	return(0);
}


PetscErrorCode PhiEquation3D::FormSNESJacobian
(
	SNES  snes,
	Vec   g_X,
	Mat   jac,
	Mat   pjac,
	void *ptr
)
{
	return(0);
}

// ForAllB(ctx.phi.n){
// PetscInt rank; MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
// printf(" rank = %i V[%i] = %4.5f V0[%i] = %4.5f\n",rank,b,ctx.V[b],ctx.V0[b]);}
