#include "Polyline.h"

using namespace EigenWrapper;

namespace
{
	/** \brief Are two points identical ? */
	inline bool SamePoints
	(
		const EigenVec3 &A,
		const EigenVec3 &B
	)
	{
		return (B-A).norm() < 1e-10;
	};
}


Polyline::Polyline()
{}


Polyline::Polyline
(
	const std::string &name,
	PetscInt rows
):
	name_{name},
	m_{EigenMat(rows,3)}
{}


Polyline::Polyline
(
	PetscInt rows
):
	m_{EigenMat(rows,3)}
{}


Polyline::Polyline
(
	const std::string &name,
	const EigenMat &m
):
	name_{name},
	m_{m}
{}


Polyline::Polyline
(
	const EigenMat &m
):
	m_{m}
{}


Polyline::Polyline
(
	const std::string &name
):
	name_{name}
{}


Polyline::Polyline
(
	const Polyline &rhs
):
	name_{rhs.name_},
	isClosed_{rhs.isClosed_},
	isSmooth_{rhs.isSmooth_},
	m_{rhs.m_}
{}


Polyline::~Polyline()
{}


Polyline& Polyline::operator=
(
	const Polyline &rhs
)
{
	if ( this != &rhs )
	{
		name_ = rhs.name_;
		isClosed_ = rhs.isClosed_;
		m_ = rhs.m_;
	}
	return *this;
}


Polyline& Polyline::operator=
(
	const EigenMat &rhs
)
{
	if ( &m_ != &rhs )
	{
		m_ = rhs;
	}
	return *this;
}


PetscErrorCode Polyline::writeInAsciiFormat
(
	const std::string &name,
	PetscInt t
) const
{
	PetscInt rank; MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
	PetscInt source = 0;
	if (rank==source)
	{
		EigenWrapper::WriteInASCIIFormat(m_,name,t);
	}
	return(0);
}


PetscErrorCode Polyline::writeInAsciiFormat
(
	PetscInt t
) const
{
	this->writeInAsciiFormat(name_,t);
	return(0);
}


PetscErrorCode Polyline::writeInBinaryFormat
(
	const std::string &name,
	PetscInt t
) const
{
	char fname[1000];
	if (t<0)
	{
		sprintf(fname,"%s.dat",name.c_str());
	}
	else
	{
		sprintf(fname,"%s_t%05d.dat",name.c_str(),t);
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
	PetscInt mdim[2] =
	{
		static_cast<PetscInt>(m_.rows()),
		static_cast<PetscInt>(m_.cols())
	};
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
	PetscViewerBinaryWrite(viewer,mdim,2,PETSC_INT);
	PetscViewerBinaryWrite(viewer,m_.data(),m_.size(),PETSC_SCALAR);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode Polyline::writeInBinaryFormat
(
	PetscInt t
) const
{
	this->writeInBinaryFormat(name_,t);
	return(0);
}


PetscErrorCode Polyline::readInBinaryFormat
(
	const std::string &name,
	PetscInt t
)
{
	name_ = name;
	this->readInBinaryFormat(t);
	return(0);
}


PetscErrorCode Polyline::readInBinaryFormat
(
	PetscInt t
)
{
	char fname[1000];
	if (t<0)
	{
		sprintf(fname,"%s.dat",name_.c_str());
	}
	else
	{
		sprintf(fname,"%s_t%05d.dat",name_.c_str(),t);
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_READ,&viewer);
	PetscInt mdim[2];
	PetscViewerBinaryRead(viewer,mdim,2,NULL,PETSC_INT);
	MPI_Bcast(mdim,2,MPIU_INT,0,PETSC_COMM_WORLD);
	m_ = EigenWrapper::EigenMat
	(
		static_cast<int>(mdim[0]),
		static_cast<int>(mdim[1])
	);
	PetscViewerBinaryRead(viewer,m_.data(),m_.size(),NULL,PETSC_SCALAR);
	MPI_Bcast(m_.data(),m_.size(),MPIU_SCALAR,0,PETSC_COMM_WORLD);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode Polyline::clear()
{
	name_.clear();
	isClosed_ = false;
	isSmooth_ = false;
	m_.resize(0,0);
	return(0);
}


PetscErrorCode Polyline::clean
(
	PetscScalar epsilon
)
{
	if (m_.rows() > 2)
	{
		std::vector<PetscInt> keepRows = {0};
		PetscInt lastr = m_.rows()-1;
		for (PetscInt r=1; r<lastr; ++r)
		{
			EigenVec3 segment = m_.row(r) - m_.row(keepRows.back());
			if ( segment.norm() > epsilon )
			{
				keepRows.push_back(r);
			}
		}
		EigenVec3 lastSegment = m_.row(lastr) - m_.row(keepRows.back());
		if ( lastSegment.norm() > epsilon || keepRows.size() == 1 )
		{
			keepRows.push_back(lastr);
		}
		else
		{
			keepRows.back() = lastr;
		}
		EigenVecI keepCols = EigenVecI::LinSpaced(m_.cols(),0,m_.cols());
		EigenMat m = m_(keepRows,keepCols);
		m_ = m;
	}
	if ( this->length() < epsilon )
	{
		this->clear();
	}
	return(0);
}


PetscErrorCode Polyline::smooth
(
	PetscScalar epsilon,
	PetscScalar s
)
{
	isClosed_ = false;
	isSmooth_ = false;
	if (m_.size() == 0)
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Warning: trying to smooth an empty polyline");
		isSmooth_ = true;
	}
	if (m_.rows() == 1)
	{;
		PetscPrintf(PETSC_COMM_WORLD,"  Warning: trying to smooth a polyline made of only one point");
		isSmooth_ = true;
	}
	if ( m_.rows() == 2 )
	{
		isSmooth_ = true;
	}
	if ( m_.rows() == 3 )
	{
		EigenMat m(2,3);
		m.row(0) = m_.row(0);
		m.row(1) = m_.row(1);
		m_ = m;
		isSmooth_ = true;
	}
	if ( m_.rows() > 3 )
	{
		EigenVec3 segment = m_.row(m_.rows()-1) - m_.row(0);
		if ( segment.norm() < epsilon )
		{
			isClosed_ = true;
			m_.row(m_.rows()-1) = m_.row(0);
		}
		bool err;
		EigenMat m;
		FitPackWrapper::SmoothPolyline(m_,m,s,isClosed_,err);
		if (!err)
		{
			isSmooth_ = true;
			m_ = m;
		}
	}
	return (0);
}


bool Polyline::isConnectedTo
(
	const Polyline &rhs
) const
{
	if (this == &rhs)
	{
		//PetscInt rank; MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
		//std::cout << "  Warning(rank " << rank << "): checking whether two identical polylines are connected" << std::endl;
		return true;
	}
	const EigenVec3 &A = m_.row(0);
	const EigenVec3 &B = m_.row(m_.rows()-1);
	const EigenVec3 &C = rhs().row(0);
	const EigenVec3 &D = rhs().row(rhs().rows()-1);
	if ( SamePoints(B,C) )
	{
		return true;
	}
	if ( SamePoints(B,D) )
	{
		return true;
	}
	if ( SamePoints(A,C) )
	{
		return true;
	}
	if ( SamePoints(A,D) )
	{
		return true;
	}
	return false;
}


PetscErrorCode Polyline::concatenateWith
(
	const Polyline &rhs
)
{
	PetscInt rank; MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
	if (this == &rhs)
	{
		//std::cout << "  Warning(rank " << rank << "): connecting two identical polylines" << std::endl;
		return(0);
	}
	if ( m_.size() + rhs().size() == 0 )
	{
		//std::cout << "  Warning(rank " << rank << "): connecting two polylines that are null" << std::endl;
		return(0);
	}
	if ( m_.size()>0 && rhs().size() == 0 )
	{
		//std::cout << "  Warning(rank " << rank << "): connecting polyline with one that is null" << std::endl;
		return(0);
	}
	if ( m_.size()==0 && rhs().size()>0 )
	{
		//std::cout << "  Warning(rank " << rank << "): connecting polyline with one that is null" << std::endl;
		m_ = rhs();
		return(0);
	}
	EigenMat lhs = m_;
	PetscInt nl = lhs.rows();
	PetscInt nr = rhs().rows();
	const EigenVec3 &A = lhs.row(0);
	const EigenVec3 &B = lhs.row(nl-1);
	const EigenVec3 &C = rhs().row(0);
	const EigenVec3 &D = rhs().row(nr-1);
	m_ = EigenMat(nl+nr-1,3);
	if ( SamePoints(B,C) )
	{
		//std::cout << "  B == C" << std::endl;
		//[A--B][C--D] x
		//[D--C][B--A]
		m_ << lhs,rhs().bottomRows(nr-1);
	}
	else if ( SamePoints(A,D) )
	{
		//std::cout << "  A == D" << std::endl;
		//[B--A][D--C]
		//[C--D][A--B] x
		m_ << rhs(),lhs.bottomRows(nl-1);
	}
	else if ( SamePoints(B,D) )
	{
		//std::cout << "  B == D" << std::endl;
		//[A--B][D--C] x
		//[C--D][B--A]
		m_ << lhs.topRows(nl-1),rhs().colwise().reverse();
	}
	else if ( SamePoints(A,C) )
	{
		//std::cout << "  A == C" << std::endl;
		//[B--A][C--D] x
		//[D--C][A--B]
		m_ << lhs.colwise().reverse(),rhs().bottomRows(nr-1);
	}
	else
	{
		std::cout << "  Error (rank " << rank << "): polylines' extremities do not meet" << std::endl;
		PetscEnd();
	}
	return(0);
}


PetscScalar Polyline::length() const
{
	PetscScalar length = 0;
	if (m_.size() > 0)
	{
		for (PetscInt r=1; r<m_.rows(); ++r)
		{
			const EigenVec3 &A = m_.row(r-1);
			const EigenVec3 &B = m_.row(r  );
			length += (B-A).norm();
		}
	}
	return length;
}


EigenVec3 Polyline::force
(
	const EigenVec3 &dir
) const
{
	PetscScalar force = this->length();
	return force*dir/dir.norm();
}


EigenVec3 Polyline::barycentre() const
{
	return m_.colwise().mean().transpose();
}


Polylines::Polylines()
{}


Polylines::Polylines
(
	const std::string &name
):
	name_{name}
{}


Polylines::Polylines
(
	const std::vector<Polyline> &pls,
	const std::string &name
):
	pls_{pls},
	name_{name}
{}


Polylines::Polylines
(
	const std::vector<Polyline> &pls
):
	pls_{pls}
{}


Polylines::Polylines
(
	const Polylines &rhs
):
	name_{rhs.name_},
	pls_{rhs.pls_}
{}


Polylines::Polylines
(
	const EigenWrapper::EigenMat  &Va,
	const EigenWrapper::EigenMatI &Fa,
	const EigenWrapper::EigenMat  &Vb,
	const EigenWrapper::EigenMatI &Fb
)
{
	this->extractFromTwoIntersectingMeshes(Va,Fa,Vb,Fb);
}


Polylines::~Polylines()
{}


Polylines& Polylines::operator=
(
	const Polylines& rhs
)
{
	if (this != &rhs)
	{
		name_ = rhs.name_;
		pls_ = rhs.pls_;
	}
	return *this;
}


Polylines& Polylines::operator=
(
	const std::vector<Polyline>& rhs
)
{
	if ( &pls_ != &rhs )
	{
		pls_ = rhs;
	}
	return *this;
}


PetscErrorCode Polylines::writeInBinaryFormat
(
	const std::string &name,
	PetscInt t
) const
{
	char fname[1000];
	if (t<0)
	{
		sprintf(fname,"%s.dat",name.c_str());
	}
	else
	{
		sprintf(fname,"%s_t%05d.dat",name.c_str(),t);
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
	PetscInt size = static_cast<PetscInt>(pls_.size());
	PetscViewerBinaryWrite(viewer,&size,1,PETSC_INT);
	for (const auto& pl:pls_)
	{
		PetscInt matDim[2] =
		{
			static_cast<PetscInt>(pl.m().rows()),
			static_cast<PetscInt>(pl.m().cols())
		};
		PetscViewerBinaryWrite(viewer,matDim,2,PETSC_INT);
		PetscViewerBinaryWrite(viewer,pl.m().data(),pl.m().size(),PETSC_SCALAR);
	}
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode Polylines::writeInBinaryFormat
(
	PetscInt t
) const
{
	this->writeInBinaryFormat(name_,t);
	return(0);
}


PetscErrorCode Polylines::readInBinaryFormat
(
	const std::string &name,
	PetscInt t
)
{
	name_ = name;
	this->readInBinaryFormat(t);
	return(0);
}


PetscErrorCode Polylines::readInBinaryFormat
(
	PetscInt t
)
{
	this->clearAll();
	char fname[1000];
	if (t<0)
	{
		sprintf(fname,"%s.dat",name_.c_str());
	}
	else
	{
		sprintf(fname,"%s_t%05d.dat",name_.c_str(),t);
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_READ,&viewer);
	PetscInt size;
	PetscViewerBinaryRead(viewer,&size,1,NULL,PETSC_INT);
	MPI_Bcast(&size,1,MPIU_INT,0,PETSC_COMM_WORLD);
	pls_.resize(static_cast<int>(size));
	for (auto& pl:pls_)
	{
		PetscInt matDim[2];
		PetscViewerBinaryRead(viewer,matDim,2,NULL,PETSC_INT);
		MPI_Bcast(matDim,2,MPIU_INT,0,PETSC_COMM_WORLD);
		pl.m() = EigenWrapper::EigenMat
		(
			static_cast<int>(matDim[0]),
			static_cast<int>(matDim[1])
		);
		PetscViewerBinaryRead(viewer,pl.m().data(),pl.m().size(),NULL,PETSC_SCALAR);
		MPI_Bcast(pl.m().data(),pl.m().size(),MPIU_SCALAR,0,PETSC_COMM_WORLD);
	}
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode Polylines::writeInAsciiFormat
(
	const std::string &name,
	PetscInt t
) const
{
	PetscInt rank; MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
	for (PetscInt p=0; p<pls_.size(); ++p)
	{
		char pname[1000];
		sprintf(pname,"%s%02d_rank%02d",name.c_str(),p,rank);
		pls_[p].writeInAsciiFormat(pname,t);
	}
	return(0);
}


PetscErrorCode Polylines::writeInAsciiFormat
(
	PetscInt t
) const
{
	this->writeInAsciiFormat(name_,t);
	return(0);
}


PetscErrorCode Polylines::writeFromSourceInASCIIFormat
(
	const std::string &name,
	PetscInt t
) const
{
	PetscInt rank; MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
	PetscInt source = 0;
	if (rank == source)
	{
		for (PetscInt p=0; p<pls_.size(); ++p)
		{
			char pname[1000];
			sprintf(pname,"%s%02d",name.c_str(),p);
			pls_[p].writeInAsciiFormat(pname,t);
		}
	}
	return(0);
}


PetscErrorCode Polylines::writeFromSourceInASCIIFormat
(
	PetscInt t
) const
{
	this->writeFromSourceInASCIIFormat(name_,t);
	return(0);
}


PetscErrorCode Polylines::clearAll()
{
	areSmooth_ = false;
	for (auto& pl:pls_)
	{
		pl.clear();
	}
	pls_.clear();
	return(0);
}


PetscErrorCode Polylines::removeEmptyPolylines()
{
	pls_.erase
	(
		std::remove_if
		(
			pls_.begin(),
			pls_.end(),
			[](const Polyline& pl){return (pl().size() == 0);}
		),
		pls_.end()
	);
	return(0);
}


PetscErrorCode Polylines::cleanAll
(
	PetscScalar epsilon
)
{
	for (auto& pl:pls_)
	{
		epsilon = std::min
		(
			epsilon,
			pl.length()/10.
		);
		pl.clean(epsilon);
	}
	this->removeEmptyPolylines();
	return(0);
}


PetscErrorCode Polylines::smoothAll
(
	PetscScalar epsilon,
	PetscScalar s
)
{
	areSmooth_ = false;
	if (pls_.size() == 0)
	{
		areSmooth_ = true;
	}
	else
	{
		EigenWrapper::EigenVecB flags(pls_.size());
		for (PetscInt p=0; p<pls_.size(); ++p)
		{
			pls_[p].smooth(epsilon,s);
			flags(p) = pls_[p].isSmooth();
		}
		if ( (flags.array() == true).all() )
		{
			areSmooth_ = true;
		}
	}
	return(0);
}


bool Polylines::concatenate
(
	PetscInt p
)
{
	if ( p > pls_.size()-1 )
	{
		PetscInt rank; MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
		std::cout << "  Error(rank " << rank << "): accessing polyline undefined in the cloud" << std::endl;
		PetscEnd();
	}
	EigenVecI connectivity(pls_.size());
	connectivity.setZero();
	for (PetscInt k=0; k<pls_.size(); ++k)
	{
		if ( pls_[p].isConnectedTo(pls_[k]) )
		{
			connectivity(k) = 1;
		}
	}
	if ( connectivity.sum()>1 )
	{
		for (PetscInt k=0; k<pls_.size(); ++k)
		{
			if ( k!=p && connectivity(k) )
			{
				pls_[p].concatenateWith(pls_[k]);
				pls_[k].clear();
			}
		}
		this->removeEmptyPolylines();
		return true;
	}
	return false;
}


PetscErrorCode Polylines::concatenateAll()
{
	PetscInt p = 0;
	PetscInt n = pls_.size();
	while (p<n)
	{
		while ( this->concatenate(p) )
		{
			p = 0;
		}
		p++;
		n = pls_.size();
	}
	return(0);
}


PetscErrorCode Polylines::extractFromTwoIntersectingMeshes
(
	const EigenMat  &Va,
	const EigenMatI &Fa,
	const EigenMat  &Vb,
	const EigenMatI &Fb
)
{
	//Reset cloud
	this->clearAll();

	if ( (Va.size() > 0) && (Vb.size() > 0) )
	{
		//Combine meshes
		std::vector<EigenMat> listVertices =
		{
			Va,
			Vb
		};
		std::vector<EigenMatI> listFaces =
		{
			Fa,
			Fb
		};
		EigenMat combinedVertices;
		EigenMatI combinedFaces;
		igl::combine(listVertices,listFaces,combinedVertices,combinedFaces);

		//Resolve intersections (copy/paste from mesh_boolean.cpp)
		EigenMat intersectedVertices;
		EigenMatI intersectedFaces,IF;
		EigenVecI J,IM;
		igl::copyleft::cgal::RemeshSelfIntersectionsParam params;
		params.stitch_all = false;
		igl::copyleft::cgal::remesh_self_intersections
		(
			combinedVertices,
			combinedFaces,
			params,
			intersectedVertices,
			intersectedFaces,
			IF,
			J,
			IM
		);

		//Check if there is an intersection
		if ( IF.size()>0 )
		{
			//Merge coinciding vertices into non-manifold vertices
			std::for_each(intersectedFaces.data(),intersectedFaces.data()+intersectedFaces.size(),[&IM](int &a){a=IM(a);});

			//Remove unreferenced vertices
			EigenMat unreferencedVertices;
			EigenMatI unreferencedFaces;
			EigenVecI UIM;
			igl::remove_unreferenced(intersectedVertices,intersectedFaces,unreferencedVertices,unreferencedFaces,UIM);

			//Compute edges of (F) --> (E,uE,EMAP,uE2E)
			EigenMatI E,uE;
			EigenVecI EMAP;
			std::vector<std::vector<size_t> > uE2E;
			igl::unique_edge_map(unreferencedFaces,E,uE,EMAP,uE2E);

			//Non manifold edges
			std::vector<std::vector<size_t> > curves;
			igl::extract_non_manifold_edge_curves(unreferencedFaces,EMAP,uE2E,curves);

			//Edge boundaries
			EigenVecI boundaryEdge;
			igl::is_boundary_edge(uE,unreferencedFaces,boundaryEdge);

			//Polylines
			for (PetscInt cj=0; cj<curves.size(); ++cj)
			{
				Polyline pl, segment(2);
				for (PetscInt ci=0; ci<curves[cj].size(); ++ci)
				{
					PetscInt index = curves[cj][ci];
					if ( !boundaryEdge(index) )
					{
						PetscInt A = uE.row(index)(0);
						PetscInt B = uE.row(index)(1);
						segment().row(0) << unreferencedVertices.row(A);
						segment().row(1) << unreferencedVertices.row(B);
						pl.concatenateWith(segment);
					}
					else
					{
						break;
					}
				}

				//Add polyline to cloud
				if ( pl().rows() > 0 )
				{
					pls_.push_back(pl);
				}
			}
		}
	}

	return(0);
}

PetscInt Polylines::globalSize() const
{
	PetscInt localSize = pls_.size(), globalSize;
	MPI_Allreduce
	(
		&localSize,
		&globalSize,
		1,
		MPI_INT,
		MPI_SUM,
		PETSC_COMM_WORLD
	);
	return globalSize;
}


PetscErrorCode Polylines::gatherAll()
{
	//Check whether cloud size > 0
	PetscInt globalSize = this->globalSize();
	if (globalSize == 0)
	{
		return(0);
	}

	//Check whether multiprocs are used
	PetscInt commSize; MPI_Comm_size(PETSC_COMM_WORLD,&commSize);
	if ( commSize > 1 )
	{
		//Rank
		PetscInt rank; MPI_Comm_rank(PETSC_COMM_WORLD,&rank);

		//Dimensions
		const PetscInt three = 3;

		//Local number of polylines
		PetscInt localN = pls_.size();

		//Array containing local number of rows (=number of points) in each polyline matrix
		EigenVecI localRows(localN);

		//Number of rows in the local stacked matrix (See below)
		PetscInt localStackRows = 0;

		//Assign localRows[0,..,localN-1] and localStackRows
		for (PetscInt p=0; p<localN; p++)
		{
			PetscInt rows   = pls_[p].m().rows();
			localRows(p)    = rows;
			localStackRows += rows;
		}

		//Stack all polylines into one single matrix stack of size (localStackRows x 3)
		EigenMat localStack(localStackRows,three);
		PetscInt start=0, rows;
		for (const auto &pl:pls_)
		{
			rows = pl().rows();
			localStack.middleRows(start,rows) = pl();
			start += rows;
		}

		//Gather the number of stacked points into an array of length COMM_SIZE
		EigenVecI gatheredStackRows(commSize);
		MPI_Allgather
		(
			&localStackRows,
			1,
			MPI_INT,
			gatheredStackRows.data(),
			1,
			MPI_INT,
			PETSC_COMM_WORLD
		);

		//Set the displacement (relative to recvbuf), at which to place the incoming data from process ranki
		EigenVecI displs(commSize);
		displs(0) = 0;
		for (PetscInt ranki=1; ranki<commSize; ++ranki)
		{
			displs(ranki) = displs(ranki-1) + gatheredStackRows(ranki-1);
		}

		//Create the global stack matrix of size (globalStackRows x 3)
		PetscInt globalStackRows = displs(commSize-1) + gatheredStackRows(commSize-1);
		EigenMat globalStack(globalStackRows,three);

		//Loop over columns
		for (PetscInt dim=0; dim<three; ++dim)
		{
			//X-th column of the local stack, X=0,1,2
			EigenVec localColX = localStack.col(dim);

			//Receiving buffer as X-th column of the global stack
			EigenVec globalColX(globalStackRows);

			//Gather local column data to global data
			MPI_Allgatherv
			(
				localColX.data(),
				localStackRows,
				MPIU_SCALAR,
				globalColX.data(),
				gatheredStackRows.data(),
				displs.data(),
				MPIU_SCALAR,
				PETSC_COMM_WORLD
			);

			//Fill the global stack matrix
			globalStack.col(dim) = globalColX;
		}

		//Gather the number of polylines into an array of length COMM_SIZE
		EigenVecI gatheredN(commSize);
		MPI_Allgather
		(
			&localN,
			1,
			MPI_INT,
			gatheredN.data(),
			1,
			MPI_INT,
			PETSC_COMM_WORLD
		);

		//Set the displacement
		displs(0) = 0;
		for (PetscInt ranki=1; ranki<commSize; ++ranki)
		{
			displs(ranki) = displs(ranki-1) + gatheredN(ranki-1);
		}

		//Receiving buffer as array of number of rows in each polyline
		PetscInt globalN = displs(commSize-1) + gatheredN(commSize-1);
		EigenVecI globalRows(globalN);

		//Gather local to global data
		MPI_Allgatherv
		(
			localRows.data(),
			localN,
			MPIU_INT,
			globalRows.data(),
			gatheredN.data(),
			displs.data(),
			MPIU_INT,
			PETSC_COMM_WORLD
		);

		//Resize cloud
		pls_.resize(globalN);

		//Extract polylines from the globalStack
		start = 0;
		rows = globalRows[0];
		pls_[0] = globalStack.middleRows(start,rows);
		for (PetscInt p=1; p<globalN; ++p)
		{
			start+= globalRows[p-1];
			rows  = globalRows[p  ];
			pls_[p] = globalStack.middleRows(start,rows);
		}
	}

	return(0);
}


PetscErrorCode Polylines::ungatherAll()
{
	PetscInt commSize;
	MPI_Comm_size(PETSC_COMM_WORLD,&commSize);
	if ( commSize > 1 )
	{
		PetscInt rank; MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
		for (PetscInt p=0; p<pls_.size(); ++p)
		{
			if ( p % commSize != rank )
			{
				pls_[p].clear();
			}
		}
	}
	this->removeEmptyPolylines();
	return (0);
}


bool Polylines::gatheredAreSmooth() const
{
	PetscInt globalSize = this->globalSize();
	if (globalSize == 0)
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Warning: gathering an empty cloud of polylines");
		return false;
	}
	PetscInt commSize;
	MPI_Comm_size(PETSC_COMM_WORLD,&commSize);
	EigenWrapper::EigenVecB flags(commSize);
// 	const bool smooth = areSmooth_;
	MPI_Allgather
	(
		&areSmooth_,
		1,
		MPI_CXX_BOOL,
		flags.data(),
		1,
		MPI_CXX_BOOL,
		PETSC_COMM_WORLD
	);
	if ( (flags.array() == true).all() )
	{
		return true;
	}
	return false;
}


PetscErrorCode TestPolyline()
{
	PetscInt rank, commSize;
	MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
	MPI_Comm_size(PETSC_COMM_WORLD,&commSize);
	Polylines cloud("cloud");
	if (rank==0)
	{
		cloud().resize(6);
		cloud(0) = Polyline(3); cloud(0).m() <<  0,0,0,  1,0,0,  2,0,0;
		cloud(1) = Polyline(5); cloud(1).m() <<  2,0,0,  3,0,0,  4,0,0,  5,0,0,  6,0,0;
		cloud(2) = Polyline(4); cloud(2).m() <<  6,0,0,  7,0,0,  8,0,0,  9,0,0;

		cloud(3) = Polyline(4); cloud(3).m() << 10,0,0, 11,0,0, 12,0,0, 13,0,0;
		cloud(4) = Polyline(4); cloud(4).m() << 13,0,0, 14,0,0, 15,0,0, 16,0,0;
		cloud(5) = Polyline(4); cloud(5).m() << 16,0,0, 17,0,0, 18,0,0, 19,0,0;
	}
	if (rank==1)
	{
		cloud().resize(3);
		cloud(0) = Polyline(5); cloud(0).m() << 20,0,0, 21,0,0, 22,0,0, 23,0,0, 24,0,0;
		cloud(1) = Polyline(3); cloud(1).m() << 24,0,0, 25,0,0, 26,0,0;
		cloud(2) = Polyline(5); cloud(2).m() << 26,0,0, 27,0,0, 28,0,0, 29,0,0, 30,0,0;
	}
	if (rank==2)
	{
		cloud().resize(2);
		cloud(0) = Polyline(5); cloud(0).m() << 30,0,0, 31,0,0, 32,0,0, 33,0,0, 34,0,0;
		cloud(1) = Polyline(3); cloud(1).m() << 35,0,0, 36,0,0, 37,0,0;
	}
	if (rank==3)
	{
		cloud().resize(1);
		cloud(0) = Polyline(2); cloud(0).m() << 37,0,0, 38,0,0;
	}
	cloud.gatherAll();
	cloud.concatenateAll();
	cloud.writeInBinaryFormat("cloud",-1);
	return(0);
}
