#ifndef POLYLINE_H
#define POLYLINE_H

#include <iostream>
#include <fstream>
#include <vector>
#include "PetscWrapper.h"
#include "EigenWrapper.h"
#include "FitPackWrapper.h"
#include "igl/combine.h"
#include "igl/copyleft/cgal/remesh_self_intersections.h"
#include "igl/remove_unreferenced.h"
#include "igl/unique_edge_map.h"
#include "igl/extract_non_manifold_edge_curves.h"
#include "igl/is_boundary_edge.h"
// #include "igl/writeSTL.h"

/** \class Polyline
 *  \brief A polyline is an Eigen matrix of the form [X0, X1, X2, ...] with Xi = [xi,yi,zi]^T.
 *         It consists of consecutive line segments connecting the vertices Xi */
class Polyline
{
private:

	/** \brief Name */
	std::string name_;

	/** \brief Flag telling if polyline has been closed */
	bool isClosed_ = false;

	/** \brief Flag telling if polyline has been smoothed */
	bool isSmooth_ = false;

	/** \brief Points in form of an Eigen matrix [X0, X1, ...] with Xi = [xi,yi,zi] */
	EigenWrapper::EigenMat m_;

public:

	/** \brief Constructor */
	Polyline();

	/** \brief Constructor
	 *  \param[in] name name
	 *  \param[in] rows number of points */
	Polyline(const std::string &name,PetscInt rows);

	/** \brief Constructor
	  * \param[in] rows number of points */
	Polyline(PetscInt rows);

	/** \brief Constructor
	 *  \param[in] name name
	 *  \param[in] m    points in form of an Eigen matrix */
	Polyline(const std::string &name,const EigenWrapper::EigenMat &m);

	/** \brief Constructor
	 *  \param[in] m points in form of an Eigen matrix */
	Polyline(const EigenWrapper::EigenMat &m);

	/** \brief Constructor
	 *  \param[in] name name */
	Polyline(const std::string &name);

	/** \brief Copy constructor */
	Polyline(const Polyline &rhs);

	/** \brief Destructor */
	~Polyline();

	/** \brief Assignment operator */
	Polyline& operator=(const Polyline &rhs);

	/** \brief Assignment operator */
	Polyline& operator=(const EigenWrapper::EigenMat &rhs);

// 	/** \brief Move constructor */
// 	Polyline(Polyline&& rhs);

// 	/** \brief move assignment */
// 	Polyline& operator=(Polyline&& rhs);

	/** \brief Write polyline. The polyline name remains unchanged
	 *  \param[in] name file name
	 *  \param[in] t    time */
	PetscErrorCode writeInBinaryFormat(const std::string &name,PetscInt t) const;

	/** \brief Write polyline. The polyline name remains unchanged
	 *  \param[in] t time */
	PetscErrorCode writeInBinaryFormat(PetscInt t) const;

	/** \brief Write polyline. The polyline name remains unchanged
	 *  \param[in] name file name
	 *  \param[in] t    time */
	PetscErrorCode writeInAsciiFormat(const std::string &name,PetscInt t) const;

	/** \brief Write polyline
	 *  \param[in] t    time */
	PetscErrorCode writeInAsciiFormat(PetscInt t) const;

	/** \brief Read polyline
	 *  \param[in] name file name
	 *  \param[in] t    time */
	PetscErrorCode readInBinaryFormat(const std::string &name,PetscInt t);

	/** \brief Read polyline
	 *  \param[in] t time */
	PetscErrorCode readInBinaryFormat(PetscInt t);

	/** \brief Clear matrix and string */
	PetscErrorCode clear();

	/** \brief Clean polyline by removing points from matrix. First and last points are kept if polyline is open
	 *  \param[in] epsilon minimum distance between two consecutive points */
	PetscErrorCode clean(PetscScalar epsilon);

	/** \brief Smooth polyline
	 *  \note The two extremity nodes unchanged is polyline is open
	 *  \param[in] epsilon minimum distance distinguishing between closed and open polyline
	 *  \param[in] s       smoothing factor */
	PetscErrorCode smooth(PetscScalar epsilon, PetscScalar s);

	/** \brief Can two polylines be concatenated ?
	 *  \param[in] rhs second polyline */
	bool isConnectedTo(const Polyline &rhs) const;

	/** \brief Concatenation of two polylines into a single one
	 *  \param[in] rhs second polyline */
	PetscErrorCode concatenateWith(const Polyline &rhs);

	/** \brief Length of a polyline */
	PetscScalar length() const;

	/** \brief Force along polyline
	 *  \param[in] dir force direction */
	EigenWrapper::EigenVec3 force(const EigenWrapper::EigenVec3 &dir) const;

	/** \brief Centre of mass */
	EigenWrapper::EigenVec3 barycentre() const;

// 	/** \brief Number of points in the matrix m */
// 	PetscInt rows() const {return m_.rows();}
//
// 	/** \brief Space dimension. It is three in 3D */
// 	PetscInt cols() const {return m_.cols();}

	/** \brief Reference access to name */
	std::string &name(){return name_;}

	/** \brief Const reference access to name */
	const std::string &name() const {return name_;}

	/** \brief Reference to close flag */
	bool& isClosed(){return isClosed_;}

	/** \brief Const reference to close flag */
	const bool& isClosed() const {return isClosed_;}

	/** \brief Reference to smooth flag */
	bool& isSmooth(){return isSmooth_;}

	/** \brief Const reference to smooth flag */
	const bool& isSmooth() const {return isSmooth_;}

	/** \brief Reference access to matrix m */
	EigenWrapper::EigenMat &m(){return m_;}

	/** \brief Const reference access to matrix m */
	const EigenWrapper::EigenMat &m() const {return m_;}

	/** \brief Reference access to matrix m */
	EigenWrapper::EigenMat &operator()(){return m_;}

	/** \brief Const reference access to matrix m */
	const EigenWrapper::EigenMat &operator()() const {return m_;}

};

/** \class Polylines
  * \brief Array of polylines */
class Polylines
{
private:

	/** \brief Name */
	std::string name_;

	/** \brief Smooth flag */
	bool areSmooth_ = false;

	/** \brief Array of polylines */
	std::vector<Polyline> pls_;

public:

	/** \brief Constructor */
	Polylines();

	/** \brief Constructor */
	Polylines(const std::string &name);

	/** \brief Constructor */
	Polylines(const std::vector<Polyline> &pls,const std::string &name);

	/** \brief Constructor */
	Polylines(const std::vector<Polyline> &pls);

	/** \brief Constructor */
	Polylines(const Polylines &rhs);

	/** \brief Create polylines by intersecting two triangulated surfaces
	 *  \note https://github.com/libigl/libigl/issues/1248
	 *  \note The environment variable IGL_NUM_THREADS should be set to one
	 *  \param[in] Va libigl matrix storing vertex coordinates of first surface
	 *  \param[in] Fa libigl matrix storing triangle connectivity of first surface
	 *  \param[in] Vb libigl matrix storing vertex coordinates of second surface
	 *  \param[in] Fb libigl matrix storing triangle connectivity of second surface */
	Polylines
	(
		const EigenWrapper::EigenMat  &Va,
		const EigenWrapper::EigenMatI &Fa,
		const EigenWrapper::EigenMat  &Vb,
		const EigenWrapper::EigenMatI &Fb
	);

	/** \brief Destructor */
	~Polylines();

	/** \brief Assigment operator */
	Polylines& operator=(const Polylines &rhs);

	/** \brief Assigment operator */
	Polylines& operator=(const std::vector<Polyline> &rhs);

// 	/** \brief Move constructor */
// 	Polylines(Polylines&& rhs);

// 	/** \brief move assignment */
// 	Polylines& operator=(Polylines&& rhs);

	/** \brief Write polylines. Cloud name remains unchanged
	 *  \param[in] name file name
	 *  \param[in] t    time */
	PetscErrorCode writeInBinaryFormat(const std::string &name,PetscInt t) const;

	/** \brief Write polylines
	 *  \param[in] t time */
	PetscErrorCode writeInBinaryFormat(PetscInt t) const;

	/** \brief Write polylines. Cloud name remains unchanged
	 *  \param[in] name file name
	 *  \param[in] t    time */
	PetscErrorCode writeInAsciiFormat(const std::string &name,PetscInt t) const;

	/** \brief Write polylines
	 *  \param[in] t time */
	PetscErrorCode writeInAsciiFormat(PetscInt t) const;

	/** \brief Write polylines from source (rank=0). Cloud name remains unchanged
	 *  \param[in] name file name
	 *  \param[in] t    time */
	PetscErrorCode writeFromSourceInASCIIFormat(const std::string &name,PetscInt t) const;

	/** \brief Write polylines from source (rank=0)
	 *  \param[in] t time */
	PetscErrorCode writeFromSourceInASCIIFormat(PetscInt t) const;

	/** \brief Read polylines
	 *  \param[in] t time */
	PetscErrorCode readInBinaryFormat(PetscInt t);

	/** \brief Read polylines
	 *  \param[in] name name
	 *  \param[in] t    time */
	PetscErrorCode readInBinaryFormat(const std::string &name,PetscInt t);

	/** \brief Clear all polylines and empty cloud */
	PetscErrorCode clearAll();

	/** \brief Remove empty polylines from cloud */
	PetscErrorCode removeEmptyPolylines();

	/** \brief Clean all polylines and remove from could if empty.
	 *  \note If unspecified or too large, epsilon is set to 10% of the polyline length
	 *  \param[in] epsilon minimum distance between two consecutive points */
	PetscErrorCode cleanAll(PetscScalar epsilon=-1);

	/** \brief Smooth all polylines in the cloud
	 *  \param[in]  epsilon minimum distance distinguishing between closed and open polyline
	 *  \param[in]  s       smoothing factor */
	PetscErrorCode smoothAll(PetscScalar epsilon,PetscScalar s);

	/** \brief Concatenate p-th polylines with others in the cloud
	 *  \param[in] p p-th polyline index */
	bool concatenate(PetscInt p);

	/** \brief Search and concatenate all polylines, which can be concatenated */
	PetscErrorCode concatenateAll();

	/** \copydoc Polylines(const EigenWrapper::EigenMat&,const EigenWrapper::EigenMatI&,const EigenWrapper::EigenMat&,const EigenWrapper::EigenMatI&) */
	PetscErrorCode extractFromTwoIntersectingMeshes
	(
		const EigenWrapper::EigenMat  &Va,
		const EigenWrapper::EigenMatI &Fa,
		const EigenWrapper::EigenMat  &Vb,
		const EigenWrapper::EigenMatI &Fb
	);

	/** \brief Return global number of polylines */
	PetscInt globalSize() const;

	/** \brief Gather polylines to all procs
	 *  \note https://www.mpi-forum.org/docs/mpi-1.1/mpi-11-html/node70.html */
	PetscErrorCode gatherAll();

	/** \brief Ungather polylines
	 *  \note polylines must have been gathered first! */
	PetscErrorCode ungatherAll();

	/** \brief Tells whether polylines in all processors are smooth */
	bool gatheredAreSmooth() const;

	/** \brief Tells whether polylines exist in the cloud */
	bool areEmpty() const {return (pls_.size() == 0) ? true : false;}

	/** \brief Cloud size */
	PetscInt size() const {return (PetscInt) pls_.size();}

	/** \brief Reference access to name */
	std::string &name(){return name_;}

	/** \brief Const reference access to name */
	const std::string &name() const {return name_;}

	/** \brief Reference to smooth flag */
	bool& areSmooth(){return areSmooth_;}

	/** \brief Const reference to smooth flag */
	const bool& areSmooth() const {return areSmooth_;}

	/** \brief Reference access to cloud */
	std::vector<Polyline> &pls(){return pls_;}

	/** \brief Const reference access to cloud */
	const std::vector<Polyline> &pls() const {return pls_;}

	/** \brief Reference access to cloud */
	std::vector<Polyline> &operator()(){return pls_;}

	/** \brief Const reference access to cloud */
	const std::vector<Polyline> &operator()() const {return pls_;}

	/** \brief Reference access to p-th polyline */
	Polyline &pl(PetscInt p){return pls_[p];}

	/** \brief Const reference access to p-th polyline */
	const Polyline &pl(PetscInt p) const {return pls_[p];}

	/** \brief Reference access to p-th polyline */
	Polyline &operator()(PetscInt p){return pls_[p];}

	/** \brief Const reference access to p-th polyline */
	const Polyline &operator()(PetscInt p) const {return pls_[p];}

// 	/** \brief Reference access to p-th polyline */
// 	Polyline &operator[](PetscInt p){return pls_[p];}

// 	/** \brief Const reference access to p-th polyline */
// 	const Polyline &operator[](PetscInt p) const {return pls_[p];}

	/** \brief Reference access to first polyline */
	Polyline &front() {return pls_.front();}

	/** \brief Const reference access to first polyline */
	const Polyline &front() const {return pls_.front();}

	/** \brief Reference access to last polyline */
	Polyline &back() {return pls_.back();}

	/** \brief Const reference access to last polyline */
	const Polyline &back() const {return pls_.back();}

};

PetscErrorCode TestPolyline();

#endif
