#ifndef SNESWRAPPER_H
#define SNESWRAPPER_H


#include "petsc.h"


/** \namespace SNESWrapper
*   \brief Wrapper to Petsc Scalable Nonlinear Equations Solvers (SNES) */
namespace SNESWrapper
{


	/** \brief Structure holding elements of the Petsc Scalable Nonlinear Equations Solver (SNES) */
	class PetscNonLinearSolver
	{
	public:
		std::string name; /** <\brief Solver name */
		SNES snes; /** <\brief SNES object */
		Mat jac; /** <\brief Jacobian matrix */
		Vec solution; /** <\brief Solution stored as global Petsc vector */
		Vec residual; /** <\brief Residual stored as global Petsc vector */
	};


	/** \brief Create %PetscNonLinearSolver
	 *  \param[in] da     DMDA distributed array
	 *  \param[in] solver %PetscNonLinearSolver
	 *  \param[in] name   Solver name */
	PetscErrorCode CreatePetscNonLinearSolver
	(
		DM da,
		PetscNonLinearSolver &solver,
		const std::string& name
	);


	/** \brief Destroy Nonlinear Petsc solver
	  * \param[in] solver Nonlinear Petsc SNES solver */
	PetscErrorCode DestroyPetscNonLinearSolver
	(
		PetscNonLinearSolver &solver
	);


	/** \brief Run Nonlinear Petsc solver
	 *  \param[in] solver          Nonlinear Petsc SNES solver
	 *  \param[in] SNESFunctionPtr Pointer to evaluation function
	 *  \param[in] SNESJacobianPtr Pointer to Jacobian matrix
	 *  \param[in] ctxPtr          Pointer to solution context */
	PetscErrorCode RunPetscNonLinearSolver
	(
		PetscNonLinearSolver &solver,
		PetscErrorCode(*SNESFunctionPtr)(SNES,Vec,Vec,void*),
		PetscErrorCode(*SNESJacobianPtr)(SNES,Vec,Mat,Mat,void*),
		void *ctxPtr
	);


	/** \brief Set solution
	 *  \param[in] solver Nonlinear Petsc SNES solver
	 *  \param[in] sol    Global Petsc vector */
	PetscErrorCode SetNonLinearSolverSolution
	(
		PetscNonLinearSolver &solver,
		Vec sol
	);


	/** \brief Get solution
	 *  \param[in]  solver Nonlinear Petsc SNES solver
	 *  \param[out] sol    Global Petsc vector */
	PetscErrorCode GetNonLinearSolverSolution
	(
		PetscNonLinearSolver &solver,
		Vec sol
	);


}


#endif
