#include "SolidBody.h"

//~~~~~~~~~~~~~~~~~~
//~~~ SOLID BODY ~~~
//~~~~~~~~~~~~~~~~~~

SolidBody::SolidBody
(
	const std::string& name,
	const EigenWrapper::EigenVec3 &centre,
	const EigenWrapper::EigenQuat &quaternion,
	const EigenWrapper::EigenMat &vertices,
	const EigenWrapper::EigenMatI &faces,
	const EigenWrapper::EigenMat &topVertices,
	const EigenWrapper::EigenMatI &topFaces
):
	name_{name},
	centre_{centre},
	quaternion_{quaternion},
	vertices_{vertices},
	faces_{faces},
	topVertices_{topVertices},
	topFaces_{topFaces}
{}


SolidBody::SolidBody
(
	const EigenWrapper::EigenVec3 &centre,
	const EigenWrapper::EigenQuat &quaternion,
	const EigenWrapper::EigenMat &vertices,
	const EigenWrapper::EigenMatI &faces,
	const EigenWrapper::EigenMat &topVertices,
	const EigenWrapper::EigenMatI &topFaces
):
	centre_{centre},
	quaternion_{quaternion},
	vertices_{vertices},
	faces_{faces},
	topVertices_{topVertices},
	topFaces_{topFaces}
{}


SolidBody::SolidBody():
	centre_{EigenWrapper::EigenVec3::Zero()},
	quaternion_{EigenWrapper::EigenQuat::Identity()}
{}


SolidBody::SolidBody
(
	const SolidBody &rhs
):
	name_{rhs.name_},
	centre_{rhs.centre_},
	quaternion_{rhs.quaternion_},
	vertices_{rhs.vertices_},
	faces_{rhs.faces_},
	topVertices_{rhs.topVertices_},
	topFaces_{rhs.topFaces_}
{}


SolidBody& SolidBody::operator=
(
	const SolidBody &rhs
)
{
	if (this != &rhs)
	{

		name_ = rhs.name_;
		centre_ = rhs.centre_;
		quaternion_ = rhs.quaternion_;
		vertices_ = rhs.vertices_;
		faces_ = rhs.faces_;
		topVertices_ = rhs.topVertices_;
		topFaces_ = rhs.topFaces_;
	}
	return *this;
}


SolidBody::~SolidBody()
{}


PetscErrorCode SolidBody::create
(
	DM da
)
{
	if (da)
	{
		PetscObjectReference((PetscObject)da);
		da_ = da;
		phi_.create(da_,name_);
		if (flattenedGrid_.size() == 0)
		{
			IglWrapper::FlattenCoordinates(da_,flattenedGrid_);
			flattenedSignedDistance_.resize(flattenedGrid_.rows());
		}
	}
	else
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Error: dynamic array is undefined\n");
		PetscEnd();
	}
	return(0);
}


PetscErrorCode SolidBody::create
(
	DM da,
	const std::string& name
)
{
	name_ = name;
	this->create(da);
	return(0);
}


PetscErrorCode SolidBody::destroy()
{
	phi_.destroy();
	return(0);
}


PetscErrorCode SolidBody::writeInVtsFormat
(
	PetscInt t
) const
{
	phi_.writeInVtsFormat(t);
	return(0);
}


PetscErrorCode SolidBody::writeInStlFormat
(
	PetscInt rank,
	PetscInt t
) const
{
	PetscInt r;
	MPI_Comm_rank(PETSC_COMM_WORLD,&r);
	if (rank == r)
	{
		char fname[1000];
		if (t<0)
		{
			sprintf(fname,"%s.stl",name_.c_str());
		}
		else
		{
			sprintf(fname,"%s_t%05d.stl",name_.c_str(),t);
		}
		std::cout << "  Write " << fname << std::endl;
		IglWrapper::WriteSTL(fname,vertices_,faces_);
	}
	return(0);
}


PetscErrorCode SolidBody::writeInStlFormat
(
	PetscInt t
) const
{
	PetscInt source = 0;
	this->writeInStlFormat(source,t);
	return(0);
}


PetscErrorCode SolidBody::writeInObjFormat
(
	PetscInt rank,
	PetscInt t
) const
{
	PetscInt rk;
	MPI_Comm_rank(PETSC_COMM_WORLD,&rk);
	if (rk == rank)
	{
		char fname[1000];
		if (t<0)
		{
			sprintf(fname,"%s.obj",name_.c_str());
		}
		else
		{
			sprintf(fname,"%s_t%05d.obj",name_.c_str(),t);
		}
		std::cout << "  Write " << fname << std::endl;
		IglWrapper::WriteOBJ(fname,vertices_,faces_);
	}
	return(0);
}


PetscErrorCode SolidBody::writeInObjFormat
(
	PetscInt t
) const
{
	PetscInt source = 0;
	this->writeInObjFormat(source,t);
	return(0);
}


PetscErrorCode SolidBody::setPhi
(
	const PetscScalar& xi
)
{
	PetscScalar cutOff = 10.*xi;
	this->setSignedDistance();
	phi_.getArray();
	ForAllKJI(da_,
	{
		PetscInt index = (i-i0) + iw * ( (j-j0) + jw * (k-k0) );
		PetscScalar d = flattenedSignedDistance_(index);
		phi_(k,j,i) =
		(
			d >  cutOff ? 0 :
			d < -cutOff ? 1.:
			0.5 - 0.5 * PetscTanhScalar( d / xi )
		);
	})
	phi_.restoreArray();
// 	PetscScalar cutOff = 10.*xi;
// 	for (PetscInt p=0; p<flattenedPhi_.size(); ++p)
// 	{
// 		PetscScalar d = signedDistance(p);
// 		flattenedPhi_(p) =
// 		(
// 			d >  cutOff ? 0 :
// 			d < -cutOff ? 1.:
// 			0.5 - 0.5 * PetscTanhScalar( d / xi )
// 		);
// 	}
	return(0);
}

// #include "igl/signed_distance.h" //TO REMOVE
PetscErrorCode SolidBody::setSignedDistance()
{
// 	EigenWrapper::EigenVec signedDistance;
	IglWrapper::FlattenedSignedDistance
	(
		vertices_,
		faces_,
		flattenedGrid_,
		flattenedSignedDistance_
	); //ERROR

//TO REMOVE
// 	igl::SignedDistanceType signedDistanceType = igl::SignedDistanceType::SIGNED_DISTANCE_TYPE_PSEUDONORMAL;
// 	//EigenWrapper::EigenVec signedDistance; //list of signed distances
// 	EigenWrapper::EigenVecI faceIndice;    //list of face indices with smallest signed distances
// 	EigenWrapper::EigenMat closestPoint;   //list of closest points
// 	EigenWrapper::EigenMat closestNormal;  //list of closest normal vectors
// 	igl::signed_distance
// 	(
// 		flattenedGrid_,
// 		vertices_,
// 		faces_,
// 		signedDistanceType,
// 		-10000,
// 		+10000,
// 		flattenedSignedDistance_,
// 		faceIndice,
// 		closestPoint,
// 		closestNormal
// 	);
//TO REMOVE
	return(0);
}


//~~~~~~~~~~~~~~~~~~~~~~~~
//~~~ TRIANGULAR PRISM ~~~
//~~~~~~~~~~~~~~~~~~~~~~~~


TriangularPrism::TriangularPrism()
:
	SolidBody
	{
		EigenWrapper::EigenVec3::Zero(),
		EigenWrapper::EigenQuat::Identity(),
		EigenWrapper::EigenMat (6,3),
		EigenWrapper::EigenMatI(8,3),
		EigenWrapper::EigenMat (1,3),
		EigenWrapper::EigenMatI(1,3)
	},
	m_{EigenWrapper::EigenMat(3,3)},
	h_{0},
	n_{EigenWrapper::EigenVec3::Zero()}
{
	m_.setZero();
}


TriangularPrism::TriangularPrism
(
	const std::string& name,
	const EigenWrapper::EigenMat3 &m,
	const PetscScalar& height
):
	SolidBody
	{
		name,
		EigenWrapper::EigenVec3::Zero(),
		EigenWrapper::EigenQuat::Identity(),
		EigenWrapper::EigenMat (6,3),
		EigenWrapper::EigenMatI(8,3),
		EigenWrapper::EigenMat (1,3),
		EigenWrapper::EigenMatI(1,3)
	},
	m_{m},
	h_{0},
	n_{EigenWrapper::EigenVec3::Zero()}
{
	this->updateMatrix();
}


TriangularPrism::TriangularPrism
(
	const TriangularPrism &rhs
):
	SolidBody{rhs},
	m_{rhs.m_},
	h_{rhs.h_},
	n_{rhs.n_}
{}


TriangularPrism& TriangularPrism::operator=
(
	const TriangularPrism &rhs
)
{
	if (this != &rhs)
	{
		SolidBody::operator=(rhs);
		m_ = rhs.m_;
		h_ = rhs.h_;
		n_ = rhs.n_;
	}
	return *this;
}


TriangularPrism::~TriangularPrism()
{}


PetscErrorCode TriangularPrism::updateMatrix()
{
	EigenWrapper::EigenVec3 A = m_.row(0);
	EigenWrapper::EigenVec3 B = m_.row(1);
	EigenWrapper::EigenVec3 C = m_.row(2);
	n_ = ( (B-A).cross(C-A) ).normalized();
	centre_ = (A+B+C)/3.;
	m_.row(0) = A-centre_;
	m_.row(1) = B-centre_;
	m_.row(2) = C-centre_;
// 	m_ = m_.rowwise() - centre_.transpose();
	return(0);
}


PetscErrorCode TriangularPrism::updateMatrix
(
	const EigenWrapper::EigenMat3 &m
)
{
	m_ = m;
	this->updateMatrix();
	return(0);
}


PetscErrorCode TriangularPrism::writeInBinaryFormat
(
	PetscInt t
) const
{
	EigenWrapper::EigenVec v(20);
	v <<
		Eigen::Map<const EigenWrapper::EigenVec>(m_.data(),m_.size()),//9
		h_,                  //1
		n_,                  //3
		centre_,             //3
		quaternion_.coeffs();//4
	PetscWrapper::WriteScalarInBinaryFormat(v.data(),name_,t,v.size());
	return(0);
}


PetscErrorCode TriangularPrism::writeInAsciiFormat
(
	PetscInt t
) const
{
	EigenWrapper::EigenVec v(20);
	v <<
		Eigen::Map<const EigenWrapper::EigenVec>(m_.data(),m_.size()),//9
		h_,             //1
		n_,                  //3
		centre_,             //3
		quaternion_.coeffs();//4
	PetscWrapper::WriteScalarInAsciiFormat(v.data(),name_,t,v.size());
	return(0);
}


PetscErrorCode TriangularPrism::readInBinaryFormat
(
	PetscInt t
)
{
	EigenWrapper::EigenVec v(20);
	PetscWrapper::ReadScalarInBinaryFormat(v.data(),name_,t,v.size());
	EigenWrapper::EigenVec m = v.segment(0,m_.size());
	m_                   = Eigen::Map<const EigenWrapper::EigenMat3>(m.data());//9
	h_              = v(m.size());             //1
	n_                   = v.segment(m.size()+1,3); //3
	centre_              = v.segment(m.size()+4,3); //3
	quaternion_.coeffs() = v.segment(m.size()+7,4); //4
	return(0);
}


PetscErrorCode TriangularPrism::triangulate()
{
	//  Example
	//  6 vertices V = {V0,V1,...V5}
	//  top         bottom      back    left    right
	//  0-------1   3-------4   1---0   0---2   2---1
	//   \     /     \     /    |\  |   |\  |   |\  |
	//    \   /       \   /     | \ |   | \ |   | \ |
	//     \ /         \ /      |  \|   |  \|   |  \|
	//      2           5       4---3   3---5   5---4
	//
	//  8 triangles T = {T0,T1,...T7}
	//  A-------B   D-------E   B---A   A---C   C---B
	//   \  0  /     \  1  /    |\ 2|   |\ 3|   |\ 4|
	//    \   /       \   /     | \ |   | \ |   | \ |
	//     \ /         \ /      |5 \|   |6 \|   |7 \|
	//      C           F       E---D   D---F   F---E
	// with vertex indices defined as
	// A=0, B=1, C=2
	// D=3, E=4, F=5
	//
	// top surface obtained by translating over the distance (height/2) along the vector T/|T|
	// bottom surface obtained by translating along -T/|T|
	// where T = (B-A)x(C-A)

	EigenWrapper::EigenVec3 delta = (h_/2.) * (n_);
	vertices_.row(0) = m_.row(0) + delta.transpose();
	vertices_.row(1) = m_.row(1) + delta.transpose();
	vertices_.row(2) = m_.row(2) + delta.transpose();
	vertices_.row(3) = m_.row(0) - delta.transpose();
	vertices_.row(4) = m_.row(1) - delta.transpose();
	vertices_.row(5) = m_.row(2) - delta.transpose();
	for (PetscInt i=0; i<vertices_.rows(); ++i)
	{
		EigenWrapper::EigenVec3 vi = vertices_.row(i);
		vi  = quaternion_.toRotationMatrix() * vi;
		vi += centre_;
		vertices_.row(i) = vi;
	}
	faces_ <<
		0,1,2, //T0 = {ABC}
		3,5,4, //T1 = [DFE}
		1,0,3, //T2 = {BAD}
		0,2,5, //T3 = {ACF}
		2,1,4, //T4 = {CBE}
		1,3,4, //T5 = {BDE}
		0,5,3, //T6 = {AFD}
		2,4,5; //T7 = {CEF}
	topVertices_ = vertices_.topRows(3);
	topFaces_ = faces_.topRows(1);
	return(0);
}


PetscErrorCode TriangularPrism::triangulateAndSetPhi
(
	const PetscScalar& xi
)
{
	this->triangulate();
	this->setPhi(xi);
	return (0);
}


EigenWrapper::EigenMat3 TriangularPrism::Tm() const
{
	EigenWrapper::EigenMat3 m;
	for (PetscInt r=0; r<m.rows(); ++r)
	{
		EigenWrapper::EigenVec3 pt = m_.row(r).transpose();
		pt = quaternion_.toRotationMatrix() * pt;
		pt += centre_;
		m.row(r) = pt;
	}
	return m;
}


EigenWrapper::EigenVec3 TriangularPrism::Tm
(
	PetscInt p
) const
{
	if ( p > m_.rows()-1 )
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Error: trying to access an undefined point in solid body");
		PetscEnd();
	}
	EigenWrapper::EigenVec3 pt = m_.row(p).transpose();
	pt = quaternion_.toRotationMatrix() * pt;
	pt += centre_;
	return pt;
}


EigenWrapper::EigenVec3 TriangularPrism::Tn() const
{
	return (quaternion_.toRotationMatrix()) * n_;
}


//~~~~~~~~~~~~~~~~~~~~~
//~~~ N-SIDED PRISM ~~~
//~~~~~~~~~~~~~~~~~~~~~

Prism::Prism()
:
	h_{0.},
	n_{EigenWrapper::EigenVec3::Zero()}
{}


Prism::Prism
(
	PetscInt n
):
	SolidBody
	{
		EigenWrapper::EigenVec3::Zero(),
		EigenWrapper::EigenQuat::Identity(),
		EigenWrapper::EigenMat (2*n+2,3),
		EigenWrapper::EigenMatI(4*n  ,3),
		EigenWrapper::EigenMat (  n+1,3),
		EigenWrapper::EigenMatI(  n  ,3)
	},
	m_{EigenWrapper::EigenMat(n,3)},
	h_{0.},
	n_{EigenWrapper::EigenVec3::Zero()}
{}


Prism::Prism
(
	const std::string& name,
	const EigenWrapper::EigenMat &m,
	const PetscScalar &height
):
	SolidBody
	{
		name,
		EigenWrapper::EigenVec3::Zero(),
		EigenWrapper::EigenQuat::Identity(),
		EigenWrapper::EigenMat (2*m.rows()+2,m.cols()),
		EigenWrapper::EigenMatI(4*m.rows()  ,m.cols()),
		EigenWrapper::EigenMat (  m.rows()+1,m.cols()),
		EigenWrapper::EigenMatI(  m.rows()  ,m.cols())
	},
	m_{m},
	h_{height},
	n_{EigenWrapper::EigenVec3::Zero()}
{
	this->updateMatrix();
}


Prism::Prism
(
	const Prism &rhs
):
	SolidBody{rhs},
	m_{rhs.m_},
	h_{rhs.h_},
	n_{rhs.n_}
{}


Prism& Prism::operator=
(
	const Prism &rhs
)
{
	if (this != &rhs)
	{
		SolidBody::operator=(rhs);
		m_ = rhs.m_;
		h_ = rhs.h_;
		n_ = rhs.n_;
	}
	return *this;
}


Prism::~Prism()
{}


PetscErrorCode Prism::updateMatrix()
{
	EigenWrapper::EigenVec3 A = m_.row(0);
	EigenWrapper::EigenVec3 B = m_.row(1);
	EigenWrapper::EigenVec3 C = m_.row(2);
	n_ = ( (B-A).cross(C-A) ).normalized();
	centre_ = m_.colwise().mean().transpose();
	m_.rowwise() -= centre_.transpose();
// 	m_ = m_.rowwise() - centre_.transpose();
	return(0);
}


PetscErrorCode Prism::updateMatrix
(
	const EigenWrapper::EigenMatX3 &m
)
{
	vertices_    = EigenWrapper::EigenMat (2*m.rows()+2,m.cols());
	faces_       = EigenWrapper::EigenMatI(4*m.rows()  ,m.cols());
	topVertices_ = EigenWrapper::EigenMat (  m.rows()+1,m.cols());
	topFaces_    = EigenWrapper::EigenMatI(  m.rows()  ,m.cols());
	m_ = m;
	this->updateMatrix();
	return(0);
}


PetscErrorCode Prism::writeInBinaryFormat
(
	PetscInt t
) const
{
	char fname[1000];
	if (t<0)
	{
		sprintf(fname,"%s.dat",name_.c_str());
	}
	else
	{
		sprintf(fname,"%s_t%05d.dat",name_.c_str(),t);
	}
	PetscInt mdim[2] =
	{
		static_cast<PetscInt>(m_.rows()),
		static_cast<PetscInt>(m_.cols())
	};
	EigenWrapper::EigenVec v(11);
	v <<
		h_,                  //1
		n_,                  //3
		centre_,             //3
		quaternion_.coeffs();//4
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
	PetscViewerBinaryWrite(viewer,mdim,2,PETSC_INT);
	PetscViewerBinaryWrite(viewer,m_.data(),m_.size(),PETSC_SCALAR);
	PetscViewerBinaryWrite(viewer,v.data(),v.size(),PETSC_SCALAR);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode Prism::writeInAsciiFormat
(
	PetscInt t
) const
{
	EigenWrapper::EigenVec v(m_.size()+11);
	v <<
		Eigen::Map<const EigenWrapper::EigenVec>(m_.data(),m_.size()),
		h_,                  //1
		n_,                  //3
		centre_,             //3
		quaternion_.coeffs();//4
	PetscWrapper::WriteScalarInAsciiFormat(v.data(),name_,t,v.size());
	return(0);
}


PetscErrorCode Prism::readInBinaryFormat
(
	PetscInt t
)
{
	char fname[1000];
	if (t<0)
	{
		sprintf(fname,"%s.dat",name_.c_str());
	}
	else
	{
		sprintf(fname,"%s_t%05d.dat",name_.c_str(),t);
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_READ,&viewer);
	PetscInt mdim[2];
	PetscViewerBinaryRead(viewer,mdim,2,NULL,PETSC_INT);
	MPI_Bcast(mdim,2,MPIU_INT,0,PETSC_COMM_WORLD);
	m_ = EigenWrapper::EigenMat
	(
		static_cast<int>(mdim[0]),
		static_cast<int>(mdim[1])
	);
	PetscViewerBinaryRead(viewer,m_.data(),m_.size(),NULL,PETSC_SCALAR);
	MPI_Bcast(m_.data(),m_.size(),MPIU_SCALAR,0,PETSC_COMM_WORLD);
	EigenWrapper::EigenVec v(11);
	PetscViewerBinaryRead(viewer,v.data(),v.size(),NULL,PETSC_SCALAR);
	MPI_Bcast(v.data(),v.size(),MPIU_SCALAR,0,PETSC_COMM_WORLD);
	h_                   = v        (0);    //1
	n_                   = v.segment(1,3);  //3
	centre_              = v.segment(4,3);  //3
	quaternion_.coeffs() = v.segment(7,4);  //4
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode Prism::triangulate()
{
	//  Example: Prism obtained with n = 4
	//
	//  (2n+2) = 10 vertices {V0,V1,...V9}
	//  top      bottom   side0    side1    side2    side3
	//  1-----4  6-----9  4-----1  1-----2  2-----3  3-----4
	//  |\   /|  |\   /|  |\    |  |\    |  |\    |  |\    |
	//  | \ / |  | \ / |  | \   |  | \   |  | \   |  | \   |
	//  |  0  |  |  5  |  |  \  |  |  \  |  |  \  |  |  \  |
	//  | / \ |  | / \ |  |   \ |  |   \ |  |   \ |  |   \ |
	//  |/   \|  |/   \|  |    \|  |    \|  |    \|  |    \|
	//  2-----3  7-----8  9-----6  6-----7  7-----8  8-----9
	//
	//  (4n) = 16 triangles {T0, T1,...T15}
	//  top       bottom   side0    side1    side2    side3
	//  A-----D   E-----H  D-----A  A-----B  B-----C  C-----D
	//  |\ 0 /|   |\ 4 /|  |\    |  |\    |  |\    |  |\    |
	//  | \ / |   | \ / |  | \ 8 |  | \ 9 |  | \ 10|  | \ 11|
	//  |1 X 3|   |5 Y 7|  |  \  |  |  \  |  |  \  |  |  \  |
	//  | / \ |   | / \ |  |12 \ |  |13 \ |  |14 \ |  |15 \ |
	//  |/ 2 \|   |/ 6 \|  |    \|  |    \|  |    \|  |    \|
	//  B-----C   F-----G  H-----E  E-----F  F-----G  G-----H
	//
	//  with vertex indices defined as
	//  X=0  A=1  B=2  C=3  D=4
	//  Y=5  E=6  F=7  G=8  H=9
	//
	// top surface obtained by translating over the distance (height/2) along the vector n = T/|T|
	// bottom surface obtained by translating along -n
	// where T=(B-A)x(C-A)

	const PetscInt n = m_.rows();
	const EigenWrapper::EigenVec3 delta = (h_/2.) * (n_);
	vertices_.row(0  ) =   delta.transpose();//V0
	vertices_.row(1+n) = - delta.transpose();//V5
	for (PetscInt r=0; r<n; ++r)
	{
		vertices_.row(1  +r) = m_.row(r) + delta.transpose(); //V1, V2, V3, V4
		vertices_.row(2+n+r) = m_.row(r) - delta.transpose(); //V6, V7, V8, V9
	}
	for (PetscInt i=0; i<vertices_.rows(); ++i)
	{
		EigenWrapper::EigenVec3 vi = vertices_.row(i);
		vi  = quaternion_.toRotationMatrix() * vi;
		vi += centre_;
		vertices_.row(i) = vi;
	}
	faces_.row(0) << (0  ),(n  ),(1    );//T0 = {041}
	faces_.row(n) << (1+n),(2+n),(1+2*n);//T4 = {569}
	for (PetscInt r=0; r<n-1; ++r)
	{
		faces_.row(1  +r) << (0  ),(1  +r),(2  +r);//T1 = {012}, T2 = {023}, T3 = {034}
		faces_.row(1+n+r) << (1+n),(3+n+r),(2+n+r);//T5 = {576}, T6 = {587}, T7 = {598}
	}
	faces_.row(2*n) << (n),(2+  n),(1  ); //T8 = {461}
	faces_.row(3*n) << (n),(1+2*n),(2+n); //T12= {496}
	for (PetscInt r=0; r<n-1; ++r)
	{
		faces_.row(1+2*n+r) << (1+r),(3+n+r),(2  +r); //T8  = {172}, T9  = {283}, T10 = {394}
		faces_.row(1+3*n+r) << (1+r),(2+n+r),(3+n+r); //T13 = {167}, T14 = {278}, T15 = {389}
	}
	topVertices_ = vertices_.topRows(n+1);
	topFaces_ = faces_.topRows(n);
	return(0);
}


EigenWrapper::EigenMatX3 Prism::Tm() const
{
	EigenWrapper::EigenMatX3 m(m_.rows(),m_.cols());
	for (PetscInt r=0; r<m.rows(); ++r)
	{
		EigenWrapper::EigenVec3 pt = m_.row(r).transpose();
		pt = quaternion_.toRotationMatrix() * pt;
		pt += centre_;
		m.row(r) = pt;
	}
	return m;
}


PetscErrorCode Prism::triangulateAndSetPhi
(
	const PetscScalar& xi
)
{
	this->triangulate();
	this->setPhi(xi);
	return (0);
}


EigenWrapper::EigenVec3 Prism::Tm
(
	PetscInt p
) const
{
	if ( p > m_.rows()-1 )
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Error: trying to access an undefined point in solid body");
		PetscEnd();
	}
	EigenWrapper::EigenVec3 pt = m_.row(p).transpose();
	pt = quaternion_.toRotationMatrix() * pt;
	pt += centre_;
	return pt;
}


EigenWrapper::EigenVec3 Prism::Tn() const
{
	return (quaternion_.toRotationMatrix()) * n_;
}


//--- BOX ---


// Box::Box
// (
// 	const std::string& name,
// 	const EigenWrapper::EigenVec3& size
// ):
// 	SolidBody
// 	{
// 		name,
// 		EigenWrapper::EigenVec3::Zero(),
// 		EigenWrapper::EigenQuat::Identity(),
// 		EigenWrapper::EigenMat ( 8,3),
// 		EigenWrapper::EigenMatI(12,3),
// 		EigenWrapper::EigenMat ( 4,3),
// 		EigenWrapper::EigenMatI( 2,3)
// 	},
// 	size_{size}
// {}
//
//
// Box::Box()
// :
// 	SolidBody
// 	{
// 		EigenWrapper::EigenVec3::Zero(),
// 		EigenWrapper::EigenQuat::Identity(),
// 		EigenWrapper::EigenMat ( 8,3),
// 		EigenWrapper::EigenMatI(12,3),
// 		EigenWrapper::EigenMat ( 4,3),
// 		EigenWrapper::EigenMatI( 2,3)
// 	},
// 	size_{EigenWrapper::EigenVec3::Ones()}
// {}
//
//
// Box::Box
// (
// 	const Box &rhs
// ):
// 	SolidBody{rhs},
// 	size_{rhs.size_}
// {}
//
//
// Box& Box::operator=(const Box &rhs)
// {
// 	if (this != &rhs)
// 	{
// 		SolidBody::operator=(rhs);
// 		size_ = rhs.size_;
// 	}
// 	return *this;
// }
//
//
// Box::~Box()
// {}


// PetscErrorCode Box::set
// (
// 	const EigenWrapper::EigenVec3 &centre,
// 	const EigenWrapper::EigenVec3 &size
// )
// {
// 	centre_ = centre;
// 	size_ = size;
// 	return(0);
// }
//
//
// PetscErrorCode Box::writeInBinaryFormat
// (
// 	PetscInt t
// ) const
// {
// 	EigenWrapper::EigenVec v(10);
// 	v <<
// 		size_,                //3
// 		centre_,              //3
// 		quaternion_.coeffs(); //4
// 	PetscWrapper::WriteScalarInBinaryFormat(v.data(),name_,t,v.size());
// 	return(0);
// }


// PetscErrorCode Box::writeInAsciiFormat
// (
// 	PetscInt t
// ) const
// {
// 	EigenWrapper::EigenVec v(10);
// 	v <<
// 		size_,                //3
// 		centre_,              //3
// 		quaternion_.coeffs(); //4
// 	PetscWrapper::WriteScalarInAsciiFormat(v.data(),name_,t,v.size());
// 	return(0);
// }
//
//
// PetscErrorCode Box::readInBinaryFormat
// (
// 	PetscInt t
// )
// {
// 	EigenWrapper::EigenVec v(10);
// 	PetscWrapper::ReadScalarInBinaryFormat(v.data(),name_,t,v.size());
// 	size_                << v.segment(0,3);
// 	centre_              << v.segment(3,3);
// 	quaternion_.coeffs() << v.segment(6,4);
// 	return(0);
// }


// PetscErrorCode Box::triangulate()
// {
// // 	 Example
// // 	 8 vertices V = {V0,V1,...V7}
// // 	 top     bottom  side0   side1   side2   side3
// // 	 0---1   4---5   3---2   2---1   1---0   0---3
// // 	 |\  |   |\  |   |\  |   |\  |   |\  |   |\  |
// // 	 | \ |   | \ |   | \ |   | \ |   | \ |   | \ |
// // 	 |  \|   |  \|   |  \|   |  \|   |  \|   |  \|
// // 	 3---2   7---6   7---6   6---5   5---4   4---7
// //
// // 	 12 triangles T = {T0,T1,...T11}
// // 	 A---B   E---F   D---C   C---B   B---A   A---D
// // 	 |\ 0|   |\ 2|   |\ 4|   |\ 5|   |\ 6|   |\ 7|
// // 	 | \ |   | \ |   | \ |   | \ |   | \ |   | \ |
// // 	 |1 \|   |3 \|   |8 \|   |9 \|   |10\|   |11\|
// // 	 D---C   H---G   H---G   G---F   F---E   E---H
// // 	with vertex indices defined as
// // 	A=0, B=1, C=2, D=3
// // 	E=4, F=5, G=6, H=7
//
// 	vertices_.row(0) = EigenWrapper::EigenVec3(0,1,0);
// 	vertices_.row(4) = EigenWrapper::EigenVec3(0,0,0);
// 	vertices_.row(1) = EigenWrapper::EigenVec3(1,1,0);
// 	vertices_.row(5) = EigenWrapper::EigenVec3(1,0,0);
// 	vertices_.row(2) = EigenWrapper::EigenVec3(1,1,1);
// 	vertices_.row(6) = EigenWrapper::EigenVec3(1,0,1);
// 	vertices_.row(3) = EigenWrapper::EigenVec3(0,1,1);
// 	vertices_.row(7) = EigenWrapper::EigenVec3(0,0,1);
// 	vertices_.array() -= 0.5;
// 	for (PetscInt i=0; i<vertices_.rows(); ++i) //auto &vi : rows ??
// 	{
// 		EigenWrapper::EigenVec3 vi = vertices_.row(i);
// 		vi  = vi.cwiseProduct(size_);
// 		vi  = quaternion_.toRotationMatrix() * vi;
// 		vi += centre_;
// 		vertices_.row(i) = vi;
// 	}
// 	faces_ <<
// 		0,2,1, //T0 = {012}
// 		0,3,2, //T1
// 		4,5,6, //T2
// 		4,6,7, //T3
// 		3,6,2, //T4
// 		2,5,1, //T5
// 		1,4,0, //T6
// 		0,7,3, //T7
// 		3,7,6, //T8
// 		2,6,5, //T9
// 		1,5,4, //T10
// 		0,4,7; //T11
// 	topVertices_ = vertices_.topRows(4);
// 	topFaces_ = faces_.topRows(2);
// 	return(0);
// };


// PetscErrorCode Box::flattenPhi
// (
// 	const PetscScalar& xi
// )
// {
// 	//https://www.youtube.com/watch?v=62-pRVZuS5c
// 	EigenWrapper::EigenVec signedDistance(flattenedPhi_.size());
// 	PetscScalar cutOff = 10.*xi;
// 	for (PetscInt p=0; p<flattenedPhi_.size(); ++p)
// 	{
// 		EigenWrapper::EigenVec3 d = flattenedGrid_.row(p);
// 		d = d - centre_;
// 		d = quaternion_.toRotationMatrix().transpose() * d;
// 		d = d.cwiseAbs() - size_/2; //[|x|-Lx/2,|y|-Ly/2,|z|-Lz/2] = [|x_i|-L_i/2]
// 		bool inside = (d.array() < 0.).all(); //any()
// 		if ( inside )
// 		{
// 			signedDistance(p) = d.maxCoeff();
// 		}
// 		else
// 		{
// 			d = d.cwiseMax(EigenWrapper::EigenVec3::Zero());//[max(|x_i|-Li/2,0)]
// 			signedDistance(p) = d.norm();
// 		}
// 		flattenedPhi_(p) =
// 		(
// 			signedDistance(p) >  cutOff ? 0 :
// 			signedDistance(p) < -cutOff ? 1.:
// 			0.5 - 0.5 * PetscTanhScalar( signedDistance(p) / xi ) //sqrt(2) <- double check later
// 		);
// 	}
// 	return(0);
// }
