#ifndef SOLIDBODY_H
#define SOLIDBODY_H
#include "EigenWrapper.h"
#include "IglWrapper.h"
#include "Field3D.h"
#include "Utils.h"


class SolidBody
{
protected:
	std::string name_;
	EigenWrapper::EigenVec3 centre_;
	EigenWrapper::EigenQuat quaternion_;
	EigenWrapper::EigenMat vertices_;
	EigenWrapper::EigenMatI faces_;
	EigenWrapper::EigenMat topVertices_;
	EigenWrapper::EigenMatI topFaces_;
	DM da_;
	ScalarField phi_;
	EigenWrapper::EigenMat flattenedGrid_; //better: use a const reference
	EigenWrapper::EigenVec flattenedSignedDistance_;
	virtual PetscErrorCode setSignedDistance();
public:
	SolidBody
	(
		const std::string& name,
		const EigenWrapper::EigenVec3 &centre,
		const EigenWrapper::EigenQuat &quaternion,
		const EigenWrapper::EigenMat  &vertices,
		const EigenWrapper::EigenMatI &faces,
		const EigenWrapper::EigenMat  &topVertices,
		const EigenWrapper::EigenMatI &topFaces
	);
	SolidBody
	(
		const EigenWrapper::EigenVec3 &centre,
		const EigenWrapper::EigenQuat &quaternion,
		const EigenWrapper::EigenMat  &vertices,
		const EigenWrapper::EigenMatI &faces,
		const EigenWrapper::EigenMat  &topVertices,
		const EigenWrapper::EigenMatI &topFaces
	);
	SolidBody();
	SolidBody(const SolidBody& rhs);
	SolidBody& operator=(const SolidBody& rhs);
	~SolidBody();
	PetscErrorCode create(DM da);
	PetscErrorCode create(DM da,const std::string& name);
	PetscErrorCode destroy();
	PetscErrorCode writeInVtsFormat(PetscInt t) const;
	PetscErrorCode writeInStlFormat(PetscInt rank,PetscInt t) const;
	PetscErrorCode writeInStlFormat(PetscInt t) const;
	PetscErrorCode writeInObjFormat(PetscInt rank,PetscInt t) const;
	PetscErrorCode writeInObjFormat(PetscInt t) const;
	PetscErrorCode setPhi(const PetscScalar& xi);

	//Reference access
	std::string &name(){return name_;}
	EigenWrapper::EigenVec3 &centre(){return centre_;}
	EigenWrapper::EigenQuat &quaternion(){return quaternion_;}
	EigenWrapper::EigenMat &vertices(){return vertices_;}
	EigenWrapper::EigenMatI &faces(){return faces_;}
	EigenWrapper::EigenMat &topVertices(){return topVertices_;}
	EigenWrapper::EigenMatI &topFaces(){return topFaces_;}
	//DM &da(){return da_;}
	ScalarField &phi(){return phi_;}
	EigenWrapper::EigenMat &flattenedGrid(){return flattenedGrid_;}
	EigenWrapper::EigenVec &flattenedSignedDistance(){return flattenedSignedDistance_;}

	//Const access
	const std::string &name() const {return name_;}
	const EigenWrapper::EigenVec3 &centre() const {return centre_;}
	const EigenWrapper::EigenQuat &quaternion() const {return quaternion_;}
	const EigenWrapper::EigenMat &vertices() const {return vertices_;}
	const EigenWrapper::EigenMatI &faces() const {return faces_;}
	const EigenWrapper::EigenMat &topVertices() const {return topVertices_;}
	const EigenWrapper::EigenMatI &topFaces() const {return topFaces_;}
	//const DM &da() const {return da_;}
	const ScalarField &phi() const {return phi_;}
	const EigenWrapper::EigenMat &flattenedGrid() const {return flattenedGrid_;}
	const EigenWrapper::EigenVec &flattenedSignedDistance() const {return flattenedSignedDistance_;}
};


class TriangularPrism: public SolidBody
{
private:
	/** \brief Template points stored as untransformed matrix
	 * \note The barycenter is set to (0 0 0)^T */
	EigenWrapper::EigenMat3 m_;

	/** \brief Thickness of the template */
	PetscScalar h_;

	/** \brief Unit normal vector, untransformed */
	EigenWrapper::EigenVec3 n_;

public:
	TriangularPrism();
	TriangularPrism
	(
		const std::string& name,
		const EigenWrapper::EigenMat3 &m,
		const PetscScalar& height
	);
	TriangularPrism(const TriangularPrism& rhs);
	TriangularPrism& operator=(const TriangularPrism& rhs);
	~TriangularPrism();
	PetscErrorCode updateMatrix(); /**< \brief Set matrix and other properties deriving from m_ */
	PetscErrorCode updateMatrix(const EigenWrapper::EigenMat3 &m); /**< \brief Set matrix and other properties deriving from m_ */
	PetscErrorCode writeInBinaryFormat(PetscInt t) const;
	PetscErrorCode writeInAsciiFormat(PetscInt t) const;
	PetscErrorCode readInBinaryFormat(PetscInt t);
	PetscErrorCode triangulate();
	PetscErrorCode triangulateAndSetPhi(const PetscScalar& xi);

	/** \brief Template points in body frame */
	EigenWrapper::EigenVec3 m(PetscInt p) const {return m_.row(p).transpose();}

	/** \brief Template points transformed to inertial frame */
	EigenWrapper::EigenMat3 Tm() const;

	/** \brief Template point transformed to inertial frame */
	EigenWrapper::EigenVec3 Tm(PetscInt) const;

	/** \brief Normal vector transformed to inertial frame */
	EigenWrapper::EigenVec3 Tn() const;

	//Reference access
	//EigenWrapper::EigenMat3 &m(){return m_;}
	PetscScalar &height(){return h_;}
	//EigenWrapper::EigenVec3 &n(){return n_;}

	//Const access
	const EigenWrapper::EigenMat3 &m() const {return m_;}
	const PetscScalar &height() const {return h_;}
	const EigenWrapper::EigenVec3 &n() const {return n_;}
};


class Prism: public SolidBody
{
private:
	/** \brief Template points stored as untransformed matrix
	 * \note The barycenter is set to (0 0 0)^T */
	EigenWrapper::EigenMatX3 m_;

	/** \brief Thickness of the template */
	PetscScalar h_;

	/** \brief Unit normal vector, untransformed */
	EigenWrapper::EigenVec3 n_;

public:
	Prism
	(
		const std::string& name,
		const EigenWrapper::EigenMat &m,
		const PetscScalar &height
	);
	Prism();
	Prism(PetscInt n);
	Prism(const Prism& rhs);
	Prism& operator=(const Prism& rhs);
	~Prism();
	PetscErrorCode updateMatrix(); /**< \brief Set matrix and other properties deriving from m_ */
	PetscErrorCode updateMatrix(const EigenWrapper::EigenMatX3 &m); /**< \brief Set matrix and other properties deriving from m_ */
	PetscErrorCode writeInBinaryFormat(PetscInt t) const;
	PetscErrorCode writeInAsciiFormat(PetscInt t) const;
	PetscErrorCode readInBinaryFormat(PetscInt t);
	PetscErrorCode triangulate();
	PetscErrorCode triangulateAndSetPhi(const PetscScalar& xi);
	EigenWrapper::EigenVec3 m(PetscInt p) const {return m_.row(p).transpose();}
	EigenWrapper::EigenMatX3 Tm() const;
	EigenWrapper::EigenVec3 Tm(PetscInt) const;
	EigenWrapper::EigenVec3 Tn() const;

	//Reference access
	//EigenWrapper::EigenMatX3 &m(){return m_;}
	PetscScalar &height(){return h_;}
	//EigenWrapper::EigenVec3 &n(){return n_;}

	//Const access
	const EigenWrapper::EigenMatX3 &m() const {return m_;}
	const PetscScalar &height() const {return h_;}
	const EigenWrapper::EigenVec3 &n() const {return n_;}
};


// class Box: public SolidBody
// {
// private:
// 	EigenWrapper::EigenVec3 size_;
// 	PetscErrorCode set(const EigenWrapper::EigenVec3 &centre,const EigenWrapper::EigenVec3 &size);
// 	PetscErrorCode flattenPhi(const PetscScalar& xi);
// public:
// 	Box
// 	(
// 		const std::string& name,
// 		const EigenWrapper::EigenVec3& size
// 	);
// 	Box();
// 	Box(const Box& rhs);
// 	Box& operator=(const Box& rhs);
// 	~Box();
// 	PetscErrorCode writeInBinaryFormat(PetscInt t) const;
// 	PetscErrorCode writeInAsciiFormat(PetscInt t) const;
// 	PetscErrorCode readInBinaryFormat(PetscInt t);
// 	PetscErrorCode triangulate();
// 	EigenWrapper::EigenVec3 &size(){return size_;}
// 	const EigenWrapper::EigenVec3 &size() const {return size_;}
// };


#endif



